
#include "Screen.h"
#include "Camera.h"

#pragma once

class Object3d;
class Light;
class Player;
class VoxelWorld;
class Sparks;
class UIButton;
class FSEffect;

typedef enum { UPGRADE_ARMOR=0, UPGRADE_TRACTOR=1, UPGRADE_ANTIGRAV=2, UPGRADE_THRUSTERS=3, UPGRADE_AMMO=4, UPGRADE_DAMAGE=5, UPGRADE_MAX=6 };
typedef enum { TUTORIAL_INTRO=0, TUTORIAL_CONTROLS=1, 
			   TUTORIAL_UI_ARMOR=2, TUTORIAL_UI_AMMO=3, TUTORIAL_UI_PRESSURE=4, TUTORIAL_UI_MINERALS=5, TUTORIAL_VEINS=6, TUTORIAL_REPAIR=7, TUTORIAL_UI_LIVES=8, 
			   TUTORIAL_SURVEY=9, TUTORIAL_ENEMIES=10, TUTORIAL_LEAVING=11,
			   TUTORIAL_CURRENTS=12, TUTORIAL_XP_MODE=13,
			   TUTORIAL_MAX=14 };
typedef enum { ARROW_NONE, ARROW_2D, ARROW_3D } ArrowType;

struct LevelDef
{
	bool	_menu_mode;
	int		_level;
	int		_xp;
	int		_n_lives;
	bool	_loaded;
	int		_next_seed;
	int		_available_upgrades;
	float	_gravity;
	float	_density;
	float	_height;
	Color	_atmosphere_color[2];
	int		_rock_count;
	int		_mineral_count;
	int		_current_count;
	int		_jellyfish_count;
	int		_pirate_count;
	int		_upgrades[UPGRADE_MAX];
	bool	_tutorials[TUTORIAL_MAX];
};

#define NEXT_SCREEN_EXIT		1
#define NEXT_SCREEN_NEWGAME		2
#define NEXT_SCREEN_CONTINUE	3
#define NEXT_SCREEN_MENU		4
#define NEXT_SCREEN_GAMEOVER	5

class GameScreen : public Screen
{
protected:
	typedef std::vector<Drawable*>	Objects2d;
	typedef std::vector<Object3d*>	Objects3d;
	typedef std::vector<Light*>		Lights;
	struct Text
	{
		std::string		_text;
		Vec3f			_pos;
		float			_life;
		float			_max_life;
		Color			_color;
		float			_value;
	};
	typedef std::list<Text>	Texts;
	typedef std::vector<UIButton*>	UIButtons;
public:
	GameScreen();
	virtual ~GameScreen();

			void init(LevelDef* level_def);
	virtual	void anim(const float t);
	virtual	void render();

	inline	float		get_gravity() { return _def->_gravity; }
	inline	float		get_density() { return _def->_density; }
			bool		check_collision(const Vec3f& p,Vec3f& collision_point);
			void		spawn_sparks(const Vec3f& pos,const float speed,const int n_particles,const float max_life,const unsigned long color1,const unsigned long color2);
			void		spawn_explosion(const Vec3f& pos,const float radius,Object3d* source);
			void		remove_object3d(Object3d* obj);
			float		get_pressure_factor(const float y);
			Vec3f		get_mouse_pos();
	inline	float		get_cell_y_size() { return _def->_height/256.0f; }
	inline	float		get_top() { return _def->_height; }
			bool		cube_colliding(const int x,const int y,const int z,const int r);
			void		spawn_combat_text(const Vec3f& pos,const std::string& str,const Color& c,const float life);
			Vec3f		convert_3d_to_2d(const Vec3f& pos);
			void		add_xp(const int xp,const Vec3f& pos);
	inline	float		get_xp_factor() { return _xp_factor; }
	inline	bool		has_surveyed() { return _surveyed; }
	inline	LevelDef*	get_level_def() { return _def; }
	inline	Player*		get_player() { return _player; }
			bool		has_los(const Vec3f& p1,const Vec3f& p2);

protected:
	void	render2d();
	void	render3d();
	void	draw_pass(const PassType pass);
	void	actual_remove_object2d(Drawable* obj);
	void	actual_remove_object3d(Object3d* obj);
	void	render_icon(const float x,const float y,const float z,const char* img_name,const Rect& uv,const float p);
	void	spawn_rock(const int xx,const int yy,const int sx,const int sy);
	void	trigger_tutorials();
	void	trigger_tutorial(const Vec3f& tutorial_pos,const std::string& text,const bool focus_on,const Vec3f& focus_point,const ArrowType arrow_type,const Vec3f& point_to);
	void	render_arrow(const Vec3f& sp1,const Vec3f& sp2,const Color& color);
	void	anim_camera(const float t);

protected:
	Camera			_camera;
	Lights			_lights;
	Objects3d		_objects3d;
	Objects3d		_to_add_objects3d;
	Objects3d		_to_del_objects3d;
	Objects2d		_objects2d;
	Objects2d		_to_add_objects2d;
	Objects2d		_to_del_objects2d;
	Player*			_player;
	VoxelWorld*		_voxels;
	Sparks*			_sparks;
	int				_landing_icon_anim_index;
	float			_landing_icon_anim_time;
	bool			_landing_icon_was_visible;
	float			_camera_wander_time;
	Vec3f			_camera_wander_inc;
	Vec3f			_camera_target_pos;
	bool			_camera_start;
	Texts			_texts;
	UIButtons		_upgrades;
	float			_time_to_respawn;
	bool			_gameover;
	unsigned long	_xp_sound;
	bool			_paused_anim;
	std::string		_tutorial_text;
	ArrowType		_tutorial_arrow;
	Vec3f			_tutorial_arrow_target;
	bool			_tutorial_focus;
	Vec3f			_tutorial_focus_point;
	Vec3f			_tutorial_pos;
	unsigned long	_time_of_last_tutorial;
	FSEffect*		_effect;
	// Level parameters
	LevelDef*	_def;
	int			_total_xp;
	float		_xp_factor;
	bool		_surveyed;
	float		_exit_warning_time;
	bool		_xp_mode;
	float		_xp_time;
protected:
	static const char* g_tooltips[UPGRADE_MAX];
};
