#include "Jellyfish.h"
#include "GameScreen.h"
#include "ModelManager.h"
#include "Material.h"
#include "RndGenerator.h"

Jellyfish::Jellyfish(void) : GameEnemy("jellyfish",NULL,MODEL_MANAGER->find_by_name("jelly01"))
{ 
	_anim_index=rand()%3;
	_anim_base_time=g_rnd.get_random_f(250.0f,750.0f);
	_anim_time=_anim_base_time;

	make_material_copy();
	_material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);
	_material->enable_pass(PASS_ALPHA);
	_material->disable_pass(PASS_SOLID);
	_material->set_material_color(1.0f,1.0f,1.0f,0.4f);

	_speed=4.0f;
	_angle=0.0f;
	_radius=2.0f;
	_wander_time=0.0f;
	_wander_inc.set(0.0f,0.0f,0.0f);
}

Jellyfish::~Jellyfish(void)
{
}

void Jellyfish::anim(const float t)
{
	GameObject::anim(t);

	_anim_time-=t;
	if (_anim_time<0)
	{
		_anim_time=_anim_base_time;
		_anim_index=(_anim_index+1)%4;
	}

	switch (_anim_index)
	{
		case 0: _model=MODEL_MANAGER->find_by_name("jelly02"); break;
		case 1: _model=MODEL_MANAGER->find_by_name("jelly01"); break;
		case 2: _model=MODEL_MANAGER->find_by_name("jelly03"); break;
		case 3: _model=MODEL_MANAGER->find_by_name("jelly01"); break;
	}

	_angle+=(t*0.1f); if (_angle>360.0f) _angle-=360.0f;

	_ori.rotation_quaternion(0.0f,1.0f,0.0f,_angle);

	// Wandering
	if (_wander_time>0.0f)
	{
		_wander_time-=t;
		
		Vec3f old_pos=_pos;
		_pos=_pos+_wander_inc*t;

		float sy=((GameScreen*)SCREEN)->get_cell_y_size();
		int px=(int)((_pos.x+128.0f+1.0f)/2.0f); 
		int py=(int)((_pos.y+sy*0.5f)/sy); 

		if (((GameScreen*)SCREEN)->cube_colliding(px,py,2,2))
		{
			_wander_time=0.0f;
			_wander_inc.set(0.0f,0.0f,0.0f);
			_pos=old_pos;
		}
	}
	else
	{
		Vec3f target_pos=_pos+Vec3f(g_rnd.get_random_f(-20.0f,20.0f),g_rnd.get_random_f(-20.0f,20.0f),0.0f);
		if ((target_pos.x>=2.0f) && (target_pos.x<=125) && (target_pos.y>=10.0f) && (target_pos.y<((GameScreen*)SCREEN)->get_top()-10.0f))
		{
			_wander_inc=target_pos-_pos;
			float dist=_wander_inc.normalize(); _wander_inc=_wander_inc*0.001f*_speed;
			_wander_time=(dist*1000.0f)/_speed;
		}
	}
}
