#include "Minerals.h"
#include "GameScreen.h"

#include "RndGenerator.h"
#include "BlitManager3d.h"
#include "Gfx.h"

Minerals::Vector	Minerals::g_minerals;

Minerals::Minerals(void) : Object3d("minerals",NULL)
{
	set_minerals(1000.0f,false);

	g_minerals.push_back(this);
}

Minerals::~Minerals(void)
{
	set_minerals(0.0f,false);

	auto it=g_minerals.begin();
	while (it!=g_minerals.end())
	{
		if ((*it)==this)
		{
			g_minerals.erase(it);
			break;
		}
		++it;
	}
}

void Minerals::anim(const float t)
{
	auto it=_elements.begin();
	while (it!=_elements.end())
	{
		(*it)->_angle+=((*it)->_speed/100.0f)*t;
		if ((*it)->_angle>360.0f) (*it)->_angle-=360.0f;
		else if ((*it)->_angle<0.0f) (*it)->_angle+=360.0f;

		++it;
	}
}

void Minerals::draw()
{
	if (GFX->get_pass()!=PASS_ALPHA) return;

	auto it=_elements.begin();
	while (it!=_elements.end())
	{
		Vec3f last_pos((*it)->_base_pos.x+(*it)->_radius*cosf(deg_to_rad((*it)->_angle)),(*it)->_base_pos.y+(*it)->_radius*sinf(deg_to_rad((*it)->_angle)),0.0f);
		for (size_t i=1; i<5; i++)
		{
			float a1=((float)(i-1))/4.0f;
			float a2=((float)i)/4.0f;
			Vec3f new_pos((*it)->_base_pos.x+(*it)->_radius*cosf(deg_to_rad((*it)->_angle+(*it)->_speed*i)),(*it)->_base_pos.y+(*it)->_radius*sinf(deg_to_rad((*it)->_angle+(*it)->_speed*i)),0.0f);
			BLIT3D->add_line(last_pos+_pos,new_pos+_pos,0.05f,Color(0.5f,1.0f,1.0f,a1),Color(0.5f,1.0f,1.0f,a2),BLIT_SRCALPHA_ONE);
			last_pos=new_pos;
		}

		++it;
	}
}

void Minerals::set_minerals(const float f,const bool sparks)
{
	_minerals=f;

	size_t n_minerals=(size_t)(ceilf((_minerals/100.0f)));

	for (size_t i=_elements.size(); i<n_minerals; i++)
	{
		Element* element=new Element;
		element->_base_pos.set(g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f),0.0f);
		element->_angle=g_rnd.get_random_f(0.0f,360.0f);
		element->_speed=g_rnd.get_random_f(10.0f,20.0f); element->_speed=(g_rnd.get_random_i(0,100)<50)?(-element->_speed):(element->_speed);
		element->_radius=g_rnd.get_random_f(0.8f,1.5f);

		_elements.push_back(element);
	}

	while (_elements.size()>n_minerals)
	{
		auto it=_elements.begin();

		if (sparks) 
		{
			((GameScreen*)SCREEN)->spawn_sparks((*it)->_base_pos+_pos,0.000025f,20,500.0f,0xFF00FFFF,0x00000022);
			((GameScreen*)SCREEN)->add_xp(10,_pos);
		}
		delete (*it);
		_elements.erase(it);
	}
}
