#include "Pirate.h"
#include "GameScreen.h"
#include "Player.h"
#include "ModelManager.h"
#include "Material.h"
#include "RndGenerator.h"

Pirate::Pirate(const float speed) : GameEnemy("pirate",NULL,MODEL_MANAGER->find_by_name("pirate"))
{ 
	make_material_copy();

	_speed=speed;
	_radius=2.0f;
	_wander_time=0.0f;
	_wander_inc.set(0.0f,0.0f,0.0f);
	_tracking=false;
}

Pirate::~Pirate(void)
{
}

void Pirate::anim(const float t)
{
	GameObject::anim(t);

	if (_tracking)
	{
		// Check if LOS at a decent distance
		GameScreen*	screen=(GameScreen*)SCREEN;
		Player*		player=screen->get_player();
		if (player)
		{
			if (player->get_health()>0.0f)
			{
				float dist=(player->_pos-_pos).norm_squared();
				if (dist>Sqr(100.0f))
				{
					_tracking=false;
				}
				else
				{
					if (!screen->has_los(_pos,player->_pos))
					{
						_tracking=false;
						_wander_time=0.0f;
					}
				}
			}
		}
		if (_tracking)
		{
			Vec3f dir=player->_pos-_pos;
			float dist=dir.normalize();
			_pos=_pos+dir*t*_speed/1000.0f;
		}
	}
	else
	{
		// Check if LOS at a decent distance
		GameScreen*	screen=(GameScreen*)SCREEN;
		Player*		player=screen->get_player();
		if (player)
		{
			if (player->get_health()>0.0f)
			{
				float dist=(player->_pos-_pos).norm_squared();
				if (dist<Sqr(40.0f))
				{
					if (screen->has_los(_pos,player->_pos))
					{
						_tracking=true;
						_wander_time=0.0f;
					}
				}
			}
		}

		// Wandering
		if (_wander_time>0.0f)
		{
			_wander_time-=t;
		
			Vec3f old_pos=_pos;
			_pos=_pos+_wander_inc*t;

			float sy=screen->get_cell_y_size();
			int px=(int)((_pos.x+128.0f+1.0f)/2.0f); 
			int py=(int)((_pos.y+sy*0.5f)/sy); 

			if (screen->cube_colliding(px,py,2,2))
			{
				_wander_time=0.0f;
				_wander_inc.set(0.0f,0.0f,0.0f);
				_pos=old_pos;
			}
		}
		else
		{
			Vec3f target_pos=_pos+Vec3f(g_rnd.get_random_f(-20.0f,20.0f),g_rnd.get_random_f(-20.0f,20.0f),0.0f);
			if ((target_pos.x>=2.0f) && (target_pos.x<=125) && (target_pos.y>=10.0f) && (target_pos.y<screen->get_top()-10.0f))
			{
				_wander_inc=target_pos-_pos;
				float dist=_wander_inc.normalize(); _wander_inc=_wander_inc*0.001f*_speed;
				_wander_time=(dist*1000.0f)/_speed;
			}
		}
	}

	Quaternion ori; ori.rotation_quaternion(0.0f,1.0f,0.0f,t*0.5f);
	_ori=_ori*ori;
}
