#include "Player.h"
#include "GameScreen.h"
#include "Minerals.h"
#include "Currents.h"
#include "GameEnemy.h"
#include "Weapon_MachineGun.h"

#include "ModelManager.h"
#include "Keyboard.h"
#include "Launcher.h"
#include "Sound.h"
#include "Mouse.h"
#include "BlitManager3d.h"
#include "ParticleSystem3d.h"
#include "RndGenerator.h"

Player::Player(const float ammo) : GameObject("player","player",MODEL_MANAGER->find_by_name("miner"))
{
	_velocity.set(0.0f,0.0f,0.0f);

	_collision_points[0].set(-2.5f,0.0f,0.0f);
	_collision_points[1].set( 2.5f,0.0f,0.0f);
	_collision_points[2].set(-2.7f,2.9f,0.0f);
	_collision_points[3].set( 2.7f,2.9f,0.0f);
	_collision_points[4].set( 0.0f,3.6f,0.0f);
	_collision_point_count=5;

	_minerals=0.0f;
	_tractor_alpha=0.0f;
	_tractor_alpha_inc=0.0f;
	_landed=false;
	_radius=3.0f;
	_max_ammo=ammo;
	_ammo=_max_ammo;
	_input=true;

	_weapon=new Weapon_MachineGun(1000.0f,0.05f,0.0f,0.0f);

	if (options_get_sound())
	{
		_thruster_vol=0.0f;
		_thruster_snd=SOUND->play("thruster",0.0f);
	}
	else _thruster_snd=-1;

	if (options_get_sound())
	{
		_tractor_snd=SOUND->play("tractor",0.0f);
	}

	_ps=new ParticleSystem3d("thruster");
	_ps->set_life(FLT_MAX);
	_ps->set_rate(0.0f,0.0f);
	_ps->set_particle_life(250.0f,100.0f);
	_ps->set_particle_speed(5.0f,0.05f);
	_ps->set_particle_size(0.05f,1.0f);
	_ps->set_particle_color(0.0f,1.0f,1.0f,1.0f,0.5f);
	_ps->set_particle_color(1.0f,1.0f,1.0f,1.0f,0.0f);
	_ps->set_emitter(EMITTER3D_CONE);
	_ps->set_blend_mode(BLEND_ALPHA);
}

Player::~Player(void)
{
	if (_thruster_snd!=-1)
	{
		SOUND->stop(_thruster_snd);
	}
	if (_tractor_snd!=-1)
	{
		SOUND->stop(_tractor_snd);
	}
}

void Player::anim(const float t)
{
	GameObject::anim(t);

	if (_tractor_alpha_inc!=0.0f)
	{
		_tractor_alpha+=_tractor_alpha_inc*t;
		if ((_tractor_alpha_inc>0.0f) && (_tractor_alpha>=1.0f)) { _tractor_alpha=1.0f; _tractor_alpha_inc=0.0f; }
		else if ((_tractor_alpha_inc<0.0f) && (_tractor_alpha<=0.0f)) { _tractor_alpha=0.0f; _tractor_alpha_inc=0.0f; }
	}
	_weapon->weapon_anim(t);
	_ps->anim(t);

	if (_health<=0.0f) 
	{
		if (_thruster_snd!=-1)
		{
			float inc_snd=-1.0f;
			_thruster_vol=Clamp(_thruster_vol+inc_snd*t/100.0f,0.0f,1.0f);
			SOUND->set_volume(_thruster_snd,_thruster_vol*options_get_fx_volume());
		}

		return;
	}
	if (!_input) return;

	Vec3f old_pos=_pos;

	GameScreen* screen=((GameScreen*)SCREEN);
	LevelDef*	def=screen->get_level_def();
	// Use thrusters
	Vec3f thruster(0.0f,0.0f,0.0f);
	if ((KEY->is_pressed(DIK_W)) || (KEY->is_pressed(DIK_UP))) thruster.y=0.25f+def->_upgrades[UPGRADE_ANTIGRAV]*0.1f;
	if ((KEY->is_pressed(DIK_S)) || (KEY->is_pressed(DIK_DOWN))) thruster.y=-0.25f+def->_upgrades[UPGRADE_ANTIGRAV]*0.1f;
	if ((KEY->is_pressed(DIK_A)) || (KEY->is_pressed(DIK_LEFT))) thruster.x=-1.0f+def->_upgrades[UPGRADE_THRUSTERS]*0.1f;
	if ((KEY->is_pressed(DIK_D)) || (KEY->is_pressed(DIK_RIGHT))) thruster.x=1.0f+def->_upgrades[UPGRADE_THRUSTERS]*0.1f;
	
	if (_landed) thruster.set(0.0f,Max(0.0f,thruster.y),0.0f);

	_velocity=_velocity+thruster*0.025f;

	if (_thruster_snd!=-1)
	{
		float inc_snd=-1.0f;
		if (!thruster.is_null()) inc_snd=1.0f;
		_thruster_vol=Clamp(_thruster_vol+inc_snd*t/100.0f,0.0f,1.0f);
		SOUND->set_volume(_thruster_snd,_thruster_vol*options_get_fx_volume());
	}
	if (!thruster.is_null())
	{
		// Get orientation
		_ps->_pos=_pos+Vec3f(0.0f,2.0f,0.0f);

		float s=_velocity.norm();

		_ps->set_particle_speed(5.0f+s*60.0f,2.0f+s*30.0f);

		if (thruster.x<0.0f) _ps->_ori.rotation_quaternion(0.0f,0.0f,1.0f,-90.0f);
		else if (thruster.x>0.0f) _ps->_ori.rotation_quaternion(0.0f,0.0f,1.0f,90.0f);
		_ps->new_particles(10);

		if (thruster.y<0.0f) _ps->_ori.rotation_quaternion(0.0f,0.0f,1.0f,0.0f);
		else if (thruster.y>0.0f) _ps->_ori.rotation_quaternion(0.0f,0.0f,1.0f,180.0f);
		_ps->new_particles(10);
	}

	// Horizontal drag
	float density=screen->get_density();
	_velocity.x=_velocity.x*Min((0.98f*(1.0f/density)),1.0f);

	// Use gravity
	float g=((GameScreen*)SCREEN)->get_gravity();
	Vec3f accel(0.0f,-0.001f*g,0.0f);

	_velocity=_velocity+(accel*0.001f)*t*t;

	// Use the currents
	auto	current_it=Currents::g_currents.begin();
	Vec3f	current_force;
	while (current_it!=Currents::g_currents.end())
	{
		if ((*current_it)->get_force(_pos,current_force))
		{
			current_force=current_force*(t*t/1000.0f);
			_velocity=_velocity+current_force;
		}

		++current_it;
	}

	// Terminal velocity
	if (!_velocity.is_null())
	{
		float d=_velocity.normalize();
		if (d>1.0f) d=1.0f;
		_velocity=_velocity*d;
	}

	// Move
	_pos=_pos+_velocity;

	// Boundaries
	if ((_pos.x<-126.0f) && (_velocity.x<0.0f)) { _pos.x=-126.0f; }
	if ((_pos.x>126.0f) && (_velocity.x>0.0f)) { _pos.x=126.0f; }
	if ((_pos.y>(screen->get_top()-10.0f)) && (_velocity.y>0.0f)) { _velocity.y=0.0f; _pos.y=screen->get_top()-10.0f; }

	// Adjust orientation (tilt miner)
	float ang=-_velocity.x*30.0f;
	_ori.rotation_quaternion(0.0f,0.0f,1.0f,ang);

	// Check for landing
	_landed=false;
	Vec3f	col_pt;
	float	speed=_velocity.norm();
	if (speed<0.06f)
	{
		if (check_collision(0,col_pt))
		{
			if (check_collision(1,col_pt))
			{
				_landed=true;
				_velocity.set(0.0f,0.0f,0.0f);
			}
		}
	}

	float	total_damage=0.0f;
	if (!_landed)
	{
		// Check collisions
		int		col_pt_index;
		if (check_collision(col_pt,col_pt_index))
		{
			float d=_velocity.norm();
			float damage=d*1500.0f;

			int		sel=0;
			Vec3f	delta;
			if (fabs(_velocity.x)>fabs(_velocity.y)) { delta.set(-_velocity.x,0.0f,0.0f); sel=0; }
			else { delta.set(0.0f,-_velocity.y,0.0f); sel=1; }
			_velocity=_velocity+delta*1.5f;
			if (!_velocity.is_null())
			{
				float d=_velocity.normalize();
				d=Max(0.05f,d);
				_velocity=_velocity*d;
			}
			Vec3f new_pos=_pos+_velocity*t;
			Vec3f old_pos=_pos;
			_pos=new_pos;
			if (check_collision(col_pt,col_pt_index))
			{
				if (sel==0) delta.set(0.0f,-_velocity.y,0.0f);
				else delta.set(-_velocity.x,0.0f,0.0f);

				_velocity=_velocity+delta*1.5f;
				if (!_velocity.is_null())
				{
					float d=_velocity.normalize();
					d=Max(0.05f,d);
					_velocity=_velocity*d;
				}
			}
			_pos=old_pos;

			total_damage=damage;

			screen->spawn_sparks(col_pt,0.0001f,50,1000.0f,0xFFFFFF00,0x00FF0000);
			if (options_get_sound()) SOUND->play("hit",options_get_fx_volume());
		}
	}
	else
	{
		// Repair!
		if (_minerals>0.0f)
		{
			// Repair
			if (_health<_max_health)
			{
				float used=Min(_minerals,Min(t*0.25f,_max_health-_health));
				_minerals-=used;
				_health+=used;
			}
			if ((_health>=_max_health*0.95f) && (_ammo<_max_ammo))
			{
				// Reload
				float used=Min(_minerals,10.0f);
				if (used>=10.0f)
				{
					_minerals-=10.0f;
					_ammo++;
				}
			}
		}
	}

	// Deal damage based on pressure
	float pressure_factor=screen->get_pressure_factor(_pos.y)-0.05f*(def->_upgrades[UPGRADE_ARMOR]);
	total_damage+=pressure_factor*t/100.0f;

	// Check enemies
	auto enemy_it=GameEnemy::g_enemies.begin();
	while (enemy_it!=GameEnemy::g_enemies.end())
	{
		Vec3f delta=_pos-(*enemy_it)->_pos;
		float dist=delta.normalize();
		if (dist<(_radius+(*enemy_it)->get_radius()))
		{
			_velocity=_velocity+delta*0.25f;
			if (!_velocity.is_null())
			{
				float d=_velocity.normalize();
				d=Max(0.05f,d);
				_velocity=_velocity*d;

				float damage=(*enemy_it)->eval_contact_damage(_pos);
				total_damage+=damage;

				(*enemy_it)->deal_damage(damage*1.1f);

				Vec3f hit_pos=((*enemy_it)->_pos+_pos)*0.5f;
				screen->spawn_sparks(hit_pos,0.0001f,50,1000.0f,0xFFFFFF00,0x00FF0000);
			}
		}

		++enemy_it;
	}

	// Evaluate damage
	if (total_damage>0)
	{
		_health=_health-total_damage;
		if (_health<=0.0f)
		{
			// BOOM
			if (options_get_sound()) SOUND->play("explosion01",options_get_fx_volume());
			screen->spawn_explosion(_pos,5.0f,this);
			screen->spawn_explosion(_pos,5.0f,this);
			set_draw(true);
			_inc_alpha=-0.01f;
		}
	}

	// Check if there's a mineral deposit below the ship
	_tractor_alpha_inc=-0.01f;
	auto mineral_it=Minerals::g_minerals.begin();
	while (mineral_it!=Minerals::g_minerals.end())
	{
		float minerals=(*mineral_it)->get_minerals();
		if (minerals>0.0f)
		{
			Vec3f delta=(*mineral_it)->_pos-(_pos+Vec3f(0.0f,1.5f,0.0f));
			float dist=delta.normalize();
			if ((dist>0.1f) && (dist<5.0f))
			{
				if ((fabs(delta.x)<cosf(deg_to_rad(45.0f))) && (delta.y<0.0f))
				{
					_tractor_alpha_inc=0.01f;
					float grab=Min(minerals,t*0.1f*(1+def->_upgrades[UPGRADE_TRACTOR]));

					minerals-=grab;
					_minerals+=grab;

					(*mineral_it)->set_minerals(minerals,true);
				}
			}
		}
		else
		{
			screen->remove_object3d(*mineral_it);
		}

		++mineral_it;
	}

	if (_tractor_snd!=-1)
	{
		SOUND->set_volume(_tractor_snd,_tractor_alpha*options_get_fx_volume()*0.5f);
	}

	// Shooting
	if ((_ammo>0) && (MOUSE->button_pressed(0)))
	{
		Vec3f	mp=screen->get_mouse_pos();
		_weapon->shoot(_pos+Vec3f(0.0f,1.0f,0.0f),mp,this);
		_ammo--;
	}
}

void Player::draw()
{
	GameObject::draw();

	if (GFX->get_pass()!=PASS_ALPHA) return;

	_ps->draw();
	_weapon->weapon_draw();

	if (_tractor_alpha==0.0f) return;
	
	Vec3f p1(_pos.x-1.0f,_pos.y+1.5f,0.0f);
	Vec3f p2(_pos.x-2.0f,_pos.y-3.5f,0.0f);
	Vec3f p3(_pos.x+2.0f,_pos.y-3.5f,0.0f);
	Vec3f p4(_pos.x+1.0f,_pos.y+1.5f,0.0f);
	Color c1(0.0f,1.0f,0.0f,0.6f*_tractor_alpha*g_rnd.get_random_f(0.8f,1.0f));
	Color c2(0.0f,0.2f,0.0f,0.0f);
	BLIT3D->add_quad(p1,p2,p3,p4,c1,c2,c2,c1,Vec3f(0.0f,1.0f,0.0f),BLIT_SRCALPHA_ONE);
}

bool Player::is_landing_speed()
{
	float	speed=_velocity.norm();
	if (speed<0.05f) return true;
	
	return false;
}

bool Player::close_to_landing()
{
	Vec3f	col_pt1,col_pt2;
	Vec3f	old_pos=_pos;
	for (int i=0; i<3; i++)
	{
		_pos=_pos+Vec3f(0.0f,-1.0f,0.0f);
		if (check_collision(0,col_pt1))
		{
			if (check_collision(1,col_pt2))
			{
				_pos=old_pos;
				if (col_pt1.y==col_pt2.y) return true;
			}
			_pos=old_pos;
			return false;
		}
	}
	_pos=old_pos;
	return false;
}

void Player::stop_thruster()
{
	if (_thruster_snd!=-1)
	{
		float inc_snd=-1.0f;
		_thruster_vol=0.0f;
		SOUND->set_volume(_thruster_snd,_thruster_vol*options_get_fx_volume());
	}
	_velocity.set(0.0f,0.0f,0.0f);
}
