#include "Weapon_MachineGun.h"
#include "GameObject.h"
#include "GameScreen.h"
#include "GameEnemy.h"
#include "Launcher.h"
#include "Sound.h"
#include "RndGenerator.h"

Weapon_MachineGun::Weapon_MachineGun(const float life,const float speed,const float heating,const float heating_recover)
{
	_life=life;
	_speed=speed;
	_heat_factor=heating;
	_heat_drain=heating_recover;
	_heat=0;
	_overheat=false;

	Material* mat=new Material;
	mat->set_color_vertex(true,false);
	mat->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);
	mat->set_shader("basic");

	_mesh=new Mesh<VertexPC>(true);
	_mesh->set_primitive(D3DPT_LINELIST);
	_mesh->alloc_vert(2*1000);
	_mesh->set_material(mat,true);

	_last_shot=0;
	_last_hit=0;
	_shot_count=0;
}

Weapon_MachineGun::~Weapon_MachineGun(void)
{
	if (_mesh)
	{
		delete _mesh;
		_mesh=NULL;
	}
}

void Weapon_MachineGun::weapon_anim(const float t)
{
	VertexPC* dest=_mesh->get_write_pointer_v();

	_shot_count=0;
	auto it=_shots.begin();
	while (it!=_shots.end())
	{
		(*it)._life-=t;
		if ((*it)._life<0.0f)
		{
			it=_shots.erase(it);
		}
		else
		{
			// Check if it hit the target
			(*it)._old_pos=(*it)._pos;
			(*it)._pos=(*it)._pos+(*it)._inc*t;

			Vec3f	pt;
			bool	hit_wall=((GameScreen*)SCREEN)->check_collision((*it)._pos,pt);
			if (hit_wall)
			{
				((GameScreen*)SCREEN)->spawn_sparks((*it)._pos,0.0001f,50,250.0f,0xFFFFFF00,0x00FF0000);
				
				if (timeGetTime()-_last_hit>50)
				{
					if (options_get_sound()) SOUND->play("hit",options_get_fx_volume());
					_last_hit=timeGetTime();
				}

				it=_shots.erase(it);
			}
			else
			{
				float t=(*it)._life/_life;

				bool		hit_enemy=false;
				GameEnemy*	enemy=NULL;
				// Check enemies
				auto enemy_it=GameEnemy::g_enemies.begin();
				while (enemy_it!=GameEnemy::g_enemies.end())
				{
					float dist=((*enemy_it)->_pos-(*it)._pos).norm_squared();
					if (dist<Sqr((*enemy_it)->get_radius()))
					{
						enemy=(*enemy_it);
						break;
					}
					++enemy_it;
				}

				if (enemy)
				{
					enemy->deal_damage(15.0f+2.5f*((GameScreen*)SCREEN)->get_level_def()->_upgrades[UPGRADE_DAMAGE]);

					((GameScreen*)SCREEN)->spawn_sparks((*it)._pos,0.0001f,50,250.0f,0xFFFFFF00,0x00FF0000);

					if (timeGetTime()-_last_hit>50)
					{
						if (options_get_sound()) SOUND->play("hit",options_get_fx_volume());
						_last_hit=timeGetTime();
					}

					it=_shots.erase(it);
				}
				else
				{
					float a=1.0f;
					if (t<0.25f) a=t*4.0f;

					dest->pos=(*it)._pos;
					dest->color=float_to_colorref(1.0f,1.0f,1.0f,a);
					dest++;

					dest->pos=(*it)._old_pos;
					dest->color=float_to_colorref(1.0f,0.0f,0.0f,a);
					dest++;

					++it;
					_shot_count++;
				}
			}
		}
	}
	_mesh->set_render_verts(_shot_count*2);
	_mesh->release_write_pointer_v();
}

void Weapon_MachineGun::weapon_draw()
{
	if (_mesh)
	{
		if (_mesh->get_vertex_count()>0)
		{
			GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0.0f,0.0f,0.0f,1.0f));

			_mesh->draw();
		}
	}
}

bool Weapon_MachineGun::is_available()
{
	if (_shot_count>490) return false;
	if (_overheat) return false;

	return true;
}

void Weapon_MachineGun::shoot(const Vec3f& source_point,const Vec3f& target_point,GameObject* owner)
{
	if (_shot_count>490) return;

	if (timeGetTime()-_last_shot>50)
	{
		if (options_get_sound()) SOUND->play("laser",options_get_fx_volume());
		_last_shot=timeGetTime();
	}
	Shot shot;
	shot._old_pos=shot._pos=source_point+Vec3f(g_rnd.get_random_f(-0.5f,0.5f),g_rnd.get_random_f(-0.5f,0.5f),0.0f);
	shot._inc=(target_point-shot._old_pos); shot._inc.normalize(); shot._inc=shot._inc*_speed;
	shot._life=_life;

	_shots.push_back(shot);
}
