// BlitManager.cpp: implementation of the BlitManager class.
//
//////////////////////////////////////////////////////////////////////

#include "BlitManager.h"
#include "TextureManager.h"
#include "Material.h"
#include "Mesh.h"
#include "Utils.h"
#include "Gfx.h"
#include <algorithm>

BlitManager* BLIT=NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BlitManager::BlitManager()
{ 
	_material=new Material;
	if (!BLIT) BLIT=this;
}

BlitManager::~BlitManager()
{
	if (BLIT==this) BLIT=NULL;
	delete _material; _material=NULL;
	clear_cache();
}

void BlitManager::clear()
{
	BlitGroups::iterator i;
	for (i=_blits.begin(); i!=_blits.end(); ++i)
	{
		(*i)->prim_count=0;
		(*i)->vcount=0;
		(*i)->avg_z=0.0f;
	}
}

void BlitManager::clear_cache()
{
	BlitGroups::iterator i;
	for (i=_blits.begin(); i!=_blits.end(); ++i)
	{
		(*i)->prim_count=0;
		(*i)->vcount=0;
		(*i)->avg_z=0.0f;
		if ((*i)->vptr) (*i)->vertex_buffer->release_write_pointer();
		if ((*i)->iptr) (*i)->index_buffer->release_write_pointer();
		(*i)->vallocated=0;
		(*i)->iallocated=0;
		delete (*i)->vertex_buffer; (*i)->vertex_buffer=NULL;
		delete (*i)->index_buffer; (*i)->index_buffer=NULL;
	}
	_blits.clear();
}

#define SET_VERTEX(vertex,px,py,pz,pc,pu,pv) vertex.p.x=px; vertex.p.y=py; vertex.p.z=pz; vertex.color=pc; vertex.u=pu; vertex.v=pv;

void BlitManager::add(long texture,float x,float y,float z,float sx,float sy,float r,float g,float b,float a,float u1,float v1,float u2,float v2,D3DTEXTUREFILTERTYPE filter)
{
	BlitGroup *grp=get_blit_group(texture,filter,4,6);

	// Setup extra stuff
	D3DCOLOR	color=D3DCOLOR_COLORVALUE(r,g,b,a);
	float		sizex=(sx==0.0f)?(TEXTURES->get_size_x(texture)*(u2-u1)):(sx);
	float		sizey=(sy==0.0f)?(TEXTURES->get_size_y(texture)*(v2-v1)):(sy);
	// Add the blit proper
	Quad	quad;

	SET_VERTEX(quad.vertex[0],x,y,z,color,u1,v1);
	SET_VERTEX(quad.vertex[1],x+sizex,y,z,color,u2,v1);
	SET_VERTEX(quad.vertex[2],x,y+sizey,z,color,u1,v2);
	SET_VERTEX(quad.vertex[3],x+sizex,y+sizey,z,color,u2,v2);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->index; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+2; grp->iptr++;
	grp->index=grp->index+4;

	grp->prim_count+=2;
	grp->vcount+=4;
	grp->avg_z+=z*2.0f;
}

void BlitManager::add(long texture,Vec3f& p,Vec2f& hotspot,float angle,float sx,float sy,float r,float g,float b,float a,float u1,float v1,float u2,float v2,D3DTEXTUREFILTERTYPE filter)
{
	BlitGroup *grp=get_blit_group(texture,filter,4,6);

	// Setup extra stuff
	D3DCOLOR	color=D3DCOLOR_COLORVALUE(r,g,b,a);
	float		sizex=(sx==0.0f)?(TEXTURES->get_size_x(texture)):(sx);
	float		sizey=(sy==0.0f)?(TEXTURES->get_size_y(texture)):(sy);
	// Add the blit proper
	Quad	quad;

	SET_VERTEX(quad.vertex[0],-hotspot.x,-hotspot.y,p.z,color,u1,v1);
	SET_VERTEX(quad.vertex[1],sizex-hotspot.x,-hotspot.y,p.z,color,u2,v1);
	SET_VERTEX(quad.vertex[2],-hotspot.x,sizey-hotspot.y,p.z,color,u1,v2);
	SET_VERTEX(quad.vertex[3],sizex-hotspot.x,sizey-hotspot.y,p.z,color,u2,v2);

	Vec3f	t(p.x,p.y,0);
	for (unsigned long i=0; i<4; i++)
	{
		quad.vertex[i].p=rotate_point2d(quad.vertex[i].p,angle);

		quad.vertex[i].p=quad.vertex[i].p+t;
	}

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->index; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+2; grp->iptr++;
	grp->index=grp->index+4;

	grp->prim_count+=2;
	grp->vcount+=4;
	grp->avg_z+=p.z*2.0f;
}

void BlitManager::add(float x,float y,float z,float sx,float sy,const Color& c)
{
	BlitGroup *grp=get_blit_group(-1,D3DTEXF_LINEAR,4,6);

	// Setup extra stuff
	D3DCOLOR	color=D3DCOLOR_COLORVALUE(c.r,c.g,c.b,c.a);
	float		sizex=sx;
	float		sizey=sy;
	// Add the blit proper
	Quad	quad;

	SET_VERTEX(quad.vertex[0],0.0f,0.0f,z,color,0.0f,0.0f);
	SET_VERTEX(quad.vertex[1],sizex,0.0f,z,color,0.0f,0.0f);
	SET_VERTEX(quad.vertex[2],0.0f,sizey,z,color,0.0f,0.0f);
	SET_VERTEX(quad.vertex[3],sizex,sizey,z,color,0.0f,0.0f);

	Vec3f	t(x,y,0);
	for (unsigned long i=0; i<4; i++)
	{
		quad.vertex[i].p=quad.vertex[i].p+t;
	}

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->index; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+2; grp->iptr++;
	grp->index=grp->index+4;

	grp->prim_count+=2;
	grp->vcount+=4;
	grp->avg_z+=z*2.0f;
}

void BlitManager::add_quad(const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Vec3f& p4,const Color& c)
{
	BlitGroup *grp=get_blit_group(-1,D3DTEXF_LINEAR,4,6);

	// Setup extra stuff
	D3DCOLOR	color=D3DCOLOR_COLORVALUE(c.r,c.g,c.b,c.a);
	// Add the blit proper
	Quad	quad;

	SET_VERTEX(quad.vertex[0],p1.x,p1.y,p1.z,color,0.0f,0.0f);
	SET_VERTEX(quad.vertex[1],p2.x,p2.y,p2.z,color,0.0f,0.0f);
	SET_VERTEX(quad.vertex[2],p3.x,p3.y,p3.z,color,0.0f,0.0f);
	SET_VERTEX(quad.vertex[3],p4.x,p4.y,p4.z,color,0.0f,0.0f);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->index; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+2; grp->iptr++;
	grp->index=grp->index+4;

	grp->prim_count+=2;
	grp->vcount+=4;
	grp->avg_z+=(p1.z+p2.z+p3.z)/3.0f+(p2.z+p3.z+p4.z)/3.0f;
}

void BlitManager::add_tri(const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Color& c)
{
	BlitGroup *grp=get_blit_group(-1,D3DTEXF_LINEAR,3,3);

	// Setup extra stuff
	D3DCOLOR	color=D3DCOLOR_COLORVALUE(c.r,c.g,c.b,c.a);
	// Add the blit proper
	Tri			tri;

	SET_VERTEX(tri.vertex[0],p1.x,p1.y,p1.z,color,0.0f,0.0f);
	SET_VERTEX(tri.vertex[1],p2.x,p2.y,p2.z,color,0.0f,0.0f);
	SET_VERTEX(tri.vertex[2],p3.x,p3.y,p3.z,color,0.0f,0.0f);

	memcpy(grp->vptr,&tri,sizeof(Tri));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Tri);

	*(grp->iptr)=(WORD)grp->index; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->index+2; grp->iptr++;
	grp->index=grp->index+3;

	grp->prim_count++;
	grp->vcount+=3;
	grp->avg_z+=p1.z;
}

class BlitManager::BlitSortFunction
{
public:
	bool operator()(const BlitGroup* o1,const BlitGroup* o2)
	{
		return o1->avg_z>o2->avg_z;
	}
};

void BlitManager::flush()
{
	draw();
	clear();
}

void BlitManager::draw()
{
	// Sort grps by z
	for (size_t i=0; i<_blits.size(); i++)
	{
		if (_blits[i]->prim_count==0) _blits[i]->avg_z=1.0f;
		else _blits[i]->avg_z=_blits[i]->avg_z/(float)_blits[i]->prim_count;
	}
	std::sort(_blits.begin(),_blits.end(),BlitSortFunction());

	// Render states
	_material->set_color_vertex(true,false);
	_material->set_cull_mode(D3DCULL_NONE);
	_material->set_alpha_test(true,10);
	_material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);

	// Render
	auto it=_blits.begin();
	while (it!=_blits.end())
	{
		if (*it)
		{
			if ((*it)->prim_count>0)
			{
				if ((*it)->texture_id==-1) { _material->set_texture(0,-1); _material->set_shader("basic"); }
				else
				{
					_material->set_texture(0,(*it)->texture_id);
					_material->set_shader("basic");
					_material->set_texture_sampling(0,(*it)->filter);
				}

				// Unlock vertex buffers, etc	
				if ((*it)->vptr) { (*it)->vertex_buffer->release_write_pointer(); (*it)->vptr=NULL; }
				if ((*it)->iptr) { (*it)->index_buffer->release_write_pointer(); (*it)->iptr=NULL; }

				// Draw!
				_material->set(D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1);

				GFX->render(D3DPT_TRIANGLELIST,(*it)->vertex_buffer,(*it)->vcount,(*it)->index_buffer,(*it)->prim_count*3);
			}
		}
		++it;
	}
}

VertexBuffer* BlitManager::create_vertex_buffer(long vcount)
{
	VertexBuffer* tmp=new VertexBuffer;

	if (tmp->create(vcount,sizeof(Vertex),D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1,true))
	{
		return tmp;
	}

	delete tmp;

	ERROR_MSG("Can't create vertex buffer!");

	return NULL;
}

IndexBuffer* BlitManager::create_index_buffer(long size)
{
	IndexBuffer* tmp=new IndexBuffer;

	if (tmp->create(size*6,true))
	{
		return tmp;
	}

	delete tmp;

	ERROR_MSG("Can't create index buffer!");

	return NULL;
}

BlitManager::BlitGroup *BlitManager::get_blit_group(long texture,D3DTEXTUREFILTERTYPE filter,const int vrequest,const int irequest)
{
	BlitGroup *grp=NULL;

	// Seek proper blit group
	BlitGroups::iterator i;
	for (i=_blits.begin(); i!=_blits.end(); ++i)
	{
		if (((*i)->texture_id==texture) && ((*i)->filter==filter))
		{
			grp=*i;
			break;
		}
	}
	if (!grp)
	{
		grp=new BlitGroup;
		grp->vallocated=0;
		grp->iallocated=0;
		grp->prim_count=0;
		grp->vcount=0;
		grp->avg_z=0.0f;
		grp->vertex_buffer=NULL;
		grp->index_buffer=NULL;
		grp->texture_id=texture;
		grp->filter=filter;
		grp->vptr=NULL;
		grp->iptr=NULL;
		grp->index=0;
		_blits.push_back(grp);
	}
	if (!grp->vertex_buffer)
	{
		grp->vertex_buffer=create_vertex_buffer(MIN_BUFFER_SIZE);
		grp->vallocated=MIN_BUFFER_SIZE;
		grp->vptr=grp->vertex_buffer->get_write_pointer();
	}
	else
	{
		if (grp->vallocated<(grp->vcount+vrequest))
		{
			// Resize
			VertexBuffer*	tmp=create_vertex_buffer(grp->vallocated*2);
			void*			ptr=tmp->get_write_pointer();

			// Copy old data
			grp->vptr=grp->vertex_buffer->get_write_pointer();
			memcpy(ptr,grp->vptr,sizeof(Vertex)*grp->vallocated);
			// Delete old vb
			grp->vertex_buffer->release_write_pointer();
			delete grp->vertex_buffer;
			grp->vertex_buffer=tmp;
			grp->vptr=((BYTE*)ptr)+(sizeof(Vertex)*grp->vallocated);
			grp->vallocated*=2;
		}
	}
	if (!grp->index_buffer)
	{
		grp->index_buffer=create_index_buffer(MIN_BUFFER_SIZE);
		grp->iallocated=MIN_BUFFER_SIZE;
		grp->iptr=grp->index_buffer->get_write_pointer();
		grp->index=0;
	}
	else
	{
		if (grp->iallocated<((grp->prim_count*3)+irequest))
		{
			// Resize
			IndexBuffer*	tmp=create_index_buffer(grp->iallocated*2);
			WORD*			ptr=tmp->get_write_pointer();

			// Copy old data
			grp->iptr=grp->index_buffer->get_write_pointer();
			memcpy(ptr,grp->iptr,sizeof(WORD)*3*grp->prim_count);
			// Delete old vb
			grp->index_buffer->release_write_pointer();
			delete grp->index_buffer;
			grp->index_buffer=tmp;
			grp->iptr=((WORD*)ptr)+(3*grp->prim_count);
			grp->iallocated*=2;
		}
	}
	if (!grp->vptr)
	{
		grp->vptr=grp->vertex_buffer->get_write_pointer();
		grp->index=0;
	}
	if (!grp->iptr)
	{
		grp->iptr=grp->index_buffer->get_write_pointer();
	}
	return grp;
}
