// BlitManager.h: interface for the BlitManager class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_BLITMANAGER_H__2AD7D1D1_ECF1_4779_B147_598AF5F61AAA__INCLUDED_)
#define AFX_BLITMANAGER_H__2AD7D1D1_ECF1_4779_B147_598AF5F61AAA__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Defs.h"
#include <vector>

#define MIN_BUFFER_SIZE	100

class VertexBuffer;
class IndexBuffer;
class Material;

class BlitManager  
{
protected:
	struct Vertex
	{
		Vec3f		p;
		DWORD		color;
		float		u,v;
	};
	struct Tri
	{
		Vertex	vertex[3];
	};
	struct Quad
	{
		Vertex	vertex[4];
	};
	struct BlitGroup
	{
		long					texture_id;
		D3DTEXTUREFILTERTYPE	filter;
		long					prim_count;
		long					vcount;
		long					index;
		VertexBuffer*			vertex_buffer;
		IndexBuffer*			index_buffer;
		long					vallocated;
		long					iallocated;
		void*					vptr;
		WORD*					iptr;
		float					avg_z;
	};
	typedef std::vector<BlitGroup*>	BlitGroups;
	class BlitSortFunction;
public:
	BlitManager();
	virtual ~BlitManager();

	void	add(long texture,float x,float y,float z,float sx,float sy,float r,float g,float b,float a,float u1,float v1,float u2,float v2,D3DTEXTUREFILTERTYPE filter=D3DTEXF_LINEAR);
	void	add(long texture,Vec3f& p,Vec2f& hotspot,float angle,float sx,float sy,float r,float g,float b,float a,float u1,float v1,float u2,float v2,D3DTEXTUREFILTERTYPE filter=D3DTEXF_LINEAR);
	void	add(float x,float y,float z,float sx,float sy,const Color& c);
	void	add_quad(const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Vec3f& p4,const Color& color);
	void	add_tri(const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Color& color);
	void	flush();
	void	draw();
	void	clear();
	void	clear_cache();

protected:
	VertexBuffer*	create_vertex_buffer(long size);
	IndexBuffer*	create_index_buffer(long size);
	BlitGroup*		get_blit_group(long texture,D3DTEXTUREFILTERTYPE filter,const int vrequest,const int irequest);

protected:
	BlitGroups			_blits;
	Material*			_material;
};

extern BlitManager* BLIT;

#endif // !defined(AFX_BLITMANAGER_H__2AD7D1D1_ECF1_4779_B147_598AF5F61AAA__INCLUDED_)
