// BlitManager3d.cpp: implementation of the BlitManager3d class.
//
//////////////////////////////////////////////////////////////////////

#include "BlitManager3d.h"
#include "TextureManager.h"
#include "Material.h"
#include "Mesh.h"
#include "Utils.h"
#include "Gfx.h"
#include "Camera.h"
#include <algorithm>

BlitManager3d* BLIT3D=NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BlitManager3d::BlitManager3d()
{
	if (!BLIT3D) BLIT3D=this;

	_material=new Material;
}

BlitManager3d::~BlitManager3d()
{
	if (BLIT3D==this) BLIT3D=NULL;
	clear_cache();

	delete _material; _material=NULL;
}

void BlitManager3d::soft_clear()
{
	BlitGroups::iterator i;
	for (i=_active_blits.begin(); i!=_active_blits.end(); ++i)
	{
		(*i)->active=false;
		(*i)->index_count=(*i)->vertex_count=0;
		if ((*i)->vptr) { (*i)->vertex_buffer->release_write_pointer(); (*i)->vptr=NULL; }
		if ((*i)->iptr) { (*i)->index_buffer->release_write_pointer(); (*i)->iptr=NULL; }
	}
	_active_blits.clear();
}

void BlitManager3d::clear_cache()
{
	BlitGroups::iterator i;
	for (i=_blits.begin(); i!=_blits.end(); ++i)
	{
		(*i)->active=false;
		(*i)->index_count=(*i)->vertex_count=0;
		if ((*i)->vptr) (*i)->vertex_buffer->release_write_pointer(); (*i)->vptr=NULL;
		if ((*i)->iptr) (*i)->index_buffer->release_write_pointer(); (*i)->iptr=NULL;
		(*i)->vallocated=0;
		(*i)->iallocated=0;
		delete (*i)->vertex_buffer; (*i)->vertex_buffer=NULL;
		delete (*i)->index_buffer; (*i)->index_buffer=NULL;
	}
	_blits.clear();
	_active_blits.clear();
}

#define SET_VERTEX(vertex,px,py,pz,nx,ny,nz,pc,pu,pv) vertex.p.x=px; vertex.p.y=py; vertex.p.z=pz; vertex.n.x=nx; vertex.n.y=ny; vertex.n.z=nz; vertex.color=pc; vertex.u=pu; vertex.v=pv;

void BlitManager3d::add(const long texture,const Vec3f& p,const Vec2f& size,const Color& color,const Rect& uv,const BlitMode blit_mode)
{
	float	hx=size.x*0.5f;
	Vec3f	n=CAMERA3D->_y_plane_normal_axis;
	Vec3f	pos[4];
	
	pos[0]=p-CAMERA3D->_y_plane_right_axis*hx+Vec3f(0.0f,size.y,0.0f);
	pos[1]=p+CAMERA3D->_y_plane_right_axis*hx+Vec3f(0.0f,size.y,0.0f);
	pos[2]=p-CAMERA3D->_y_plane_right_axis*hx;
	pos[3]=p+CAMERA3D->_y_plane_right_axis*hx;

	Quad		quad;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);

	SET_VERTEX(quad.vertex[0],pos[0].x,pos[0].y,pos[0].z,n.x,n.y,n.z,c,uv.x1,uv.y1);
	SET_VERTEX(quad.vertex[1],pos[1].x,pos[1].y,pos[1].z,n.x,n.y,n.z,c,uv.x2,uv.y1);
	SET_VERTEX(quad.vertex[2],pos[2].x,pos[2].y,pos[2].z,n.x,n.y,n.z,c,uv.x1,uv.y2);
	SET_VERTEX(quad.vertex[3],pos[3].x,pos[3].y,pos[3].z,n.x,n.y,n.z,c,uv.x2,uv.y2);

	BlitGroup *grp=request_blit_group(texture,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add(const long texture,const Vec3f& p,const Rect& r,const Color& color,const Rect& uv,const BlitMode blit_mode)
{
	Vec3f	n=CAMERA3D->_y_plane_normal_axis;
	Vec3f	pos[4];

	pos[0]=CAMERA3D->_y_plane_right_axis*r.x1+Vec3f(0.0f,r.y1,0.0f);
	pos[1]=CAMERA3D->_y_plane_right_axis*r.x2+Vec3f(0.0f,r.y1,0.0f);
	pos[2]=CAMERA3D->_y_plane_right_axis*r.x1+Vec3f(0.0f,r.y2,0.0f);
	pos[3]=CAMERA3D->_y_plane_right_axis*r.x2+Vec3f(0.0f,r.y2,0.0f);

	pos[0]=p+pos[0];
	pos[1]=p+pos[1];
	pos[2]=p+pos[2];
	pos[3]=p+pos[3];

	Quad		quad;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);

	SET_VERTEX(quad.vertex[0],pos[0].x,pos[0].y,pos[0].z,n.x,n.y,n.z,c,uv.x1,uv.y1);
	SET_VERTEX(quad.vertex[1],pos[1].x,pos[1].y,pos[1].z,n.x,n.y,n.z,c,uv.x2,uv.y1);
	SET_VERTEX(quad.vertex[2],pos[2].x,pos[2].y,pos[2].z,n.x,n.y,n.z,c,uv.x1,uv.y2);
	SET_VERTEX(quad.vertex[3],pos[3].x,pos[3].y,pos[3].z,n.x,n.y,n.z,c,uv.x2,uv.y2);

	BlitGroup *grp=request_blit_group(texture,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add(const long texture,const Vec3f& p,const Rect& r,const Quaternion& rot,const Color& color,const Rect& uv,const BlitMode blit_mode)
{
	Vec3f	n=CAMERA3D->_y_plane_normal_axis;
	Vec3f	pos[4];

	pos[0]=CAMERA3D->_y_plane_right_axis*r.x1+Vec3f(0.0f,r.y1,0.0f);
	pos[1]=CAMERA3D->_y_plane_right_axis*r.x2+Vec3f(0.0f,r.y1,0.0f);
	pos[2]=CAMERA3D->_y_plane_right_axis*r.x1+Vec3f(0.0f,r.y2,0.0f);
	pos[3]=CAMERA3D->_y_plane_right_axis*r.x2+Vec3f(0.0f,r.y2,0.0f);

	for (int i=0; i<4; i++) 
	{
		rot.rotate_point(pos[i],pos[i]);
		pos[i]=p+pos[i];
	}

	Quad		quad;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);

	SET_VERTEX(quad.vertex[0],pos[0].x,pos[0].y,pos[0].z,n.x,n.y,n.z,c,uv.x1,uv.y1);
	SET_VERTEX(quad.vertex[1],pos[1].x,pos[1].y,pos[1].z,n.x,n.y,n.z,c,uv.x2,uv.y1);
	SET_VERTEX(quad.vertex[2],pos[2].x,pos[2].y,pos[2].z,n.x,n.y,n.z,c,uv.x1,uv.y2);
	SET_VERTEX(quad.vertex[3],pos[3].x,pos[3].y,pos[3].z,n.x,n.y,n.z,c,uv.x2,uv.y2);

	BlitGroup *grp=request_blit_group(texture,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add(const long texture,const Vec3f& p,const Vec3f& up,const Vec3f& right,const Vec3f& n,const Color& color,const Rect& uv,const BlitMode blit_mode)
{
	Vec3f	pos[4];

	pos[0]=p-right+up;
	pos[1]=p+right+up;
	pos[2]=p-right-up;
	pos[3]=p+right-up;

	Quad		quad;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);

	SET_VERTEX(quad.vertex[0],pos[0].x,pos[0].y,pos[0].z,n.x,n.y,n.z,c,uv.x1,uv.y1);
	SET_VERTEX(quad.vertex[1],pos[1].x,pos[1].y,pos[1].z,n.x,n.y,n.z,c,uv.x2,uv.y1);
	SET_VERTEX(quad.vertex[2],pos[2].x,pos[2].y,pos[2].z,n.x,n.y,n.z,c,uv.x1,uv.y2);
	SET_VERTEX(quad.vertex[3],pos[3].x,pos[3].y,pos[3].z,n.x,n.y,n.z,c,uv.x2,uv.y2);

	BlitGroup *grp=request_blit_group(texture,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add_billboard(const long texture,const Vec3f& p,const Vec2f& size,const Color& color,const Rect& uv,const BlitMode blit_mode)
{
	float	hx=size.x*0.5f;
	float	hy=size.y*0.5f;
	Vec3f	n=CAMERA3D->_normal_axis;
	Vec3f	pos[4];

	pos[0]=p-CAMERA3D->_right_axis*hx+CAMERA3D->_up_axis*hy;
	pos[1]=p+CAMERA3D->_right_axis*hx+CAMERA3D->_up_axis*hy;
	pos[2]=p-CAMERA3D->_right_axis*hx-CAMERA3D->_up_axis*hy;
	pos[3]=p+CAMERA3D->_right_axis*hx-CAMERA3D->_up_axis*hy;

	Quad		quad;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);

	SET_VERTEX(quad.vertex[0],pos[0].x,pos[0].y,pos[0].z,n.x,n.y,n.z,c,uv.x1,uv.y1);
	SET_VERTEX(quad.vertex[1],pos[1].x,pos[1].y,pos[1].z,n.x,n.y,n.z,c,uv.x2,uv.y1);
	SET_VERTEX(quad.vertex[2],pos[2].x,pos[2].y,pos[2].z,n.x,n.y,n.z,c,uv.x1,uv.y2);
	SET_VERTEX(quad.vertex[3],pos[3].x,pos[3].y,pos[3].z,n.x,n.y,n.z,c,uv.x2,uv.y2);

	BlitGroup *grp=request_blit_group(texture,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add_billboard(const long texture,const Vec3f& p,const Vec2f& size,const Quaternion& rot,const Color& color,const Rect& uv,const BlitMode blit_mode)
{
	float	hx=size.x*0.5f;
	float	hy=size.y*0.5f;
	Vec3f	n=CAMERA3D->_normal_axis;
	Vec3f	pos[4];

	pos[0]=-CAMERA3D->_right_axis*hx+CAMERA3D->_up_axis*hy;
	pos[1]= CAMERA3D->_right_axis*hx+CAMERA3D->_up_axis*hy;
	pos[2]=-CAMERA3D->_right_axis*hx-CAMERA3D->_up_axis*hy;
	pos[3]= CAMERA3D->_right_axis*hx-CAMERA3D->_up_axis*hy;

	rot.rotate_point(pos[0],pos[0]);
	rot.rotate_point(pos[1],pos[1]);
	rot.rotate_point(pos[2],pos[2]);
	rot.rotate_point(pos[3],pos[3]);

	pos[0]=p+pos[0];
	pos[1]=p+pos[1];
	pos[2]=p+pos[2];
	pos[3]=p+pos[3];

	Quad		quad;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);

	SET_VERTEX(quad.vertex[0],pos[0].x,pos[0].y,pos[0].z,n.x,n.y,n.z,c,uv.x1,uv.y1);
	SET_VERTEX(quad.vertex[1],pos[1].x,pos[1].y,pos[1].z,n.x,n.y,n.z,c,uv.x2,uv.y1);
	SET_VERTEX(quad.vertex[2],pos[2].x,pos[2].y,pos[2].z,n.x,n.y,n.z,c,uv.x1,uv.y2);
	SET_VERTEX(quad.vertex[3],pos[3].x,pos[3].y,pos[3].z,n.x,n.y,n.z,c,uv.x2,uv.y2);

	BlitGroup *grp=request_blit_group(texture,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add_ground(const long texture,const Vec3f& p,const Vec2f& size,const Color& color,const Rect& uv,const BlitMode blit_mode)
{
	float	hx=size.x*0.5f;
	float	hy=size.y*0.5f;
	Vec3f	n(0.0f,1.0f,0.0f);
	Vec3f	pos[4];

	pos[0]=p-CAMERA3D->_y_plane_right_axis*hx-CAMERA3D->_y_plane_normal_axis*hy;
	pos[1]=p+CAMERA3D->_y_plane_right_axis*hx-CAMERA3D->_y_plane_normal_axis*hy;
	pos[2]=p-CAMERA3D->_y_plane_right_axis*hx+CAMERA3D->_y_plane_normal_axis*hy;
	pos[3]=p+CAMERA3D->_y_plane_right_axis*hx+CAMERA3D->_y_plane_normal_axis*hy;

	Quad		quad;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);

	SET_VERTEX(quad.vertex[0],pos[0].x,pos[0].y,pos[0].z,n.x,n.y,n.z,c,uv.x1,uv.y1);
	SET_VERTEX(quad.vertex[1],pos[1].x,pos[1].y,pos[1].z,n.x,n.y,n.z,c,uv.x2,uv.y1);
	SET_VERTEX(quad.vertex[2],pos[2].x,pos[2].y,pos[2].z,n.x,n.y,n.z,c,uv.x1,uv.y2);
	SET_VERTEX(quad.vertex[3],pos[3].x,pos[3].y,pos[3].z,n.x,n.y,n.z,c,uv.x2,uv.y2);

	BlitGroup *grp=request_blit_group(texture,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add_ground(const long texture,const Vec3f& p,const Vec2f& size,const Quaternion& rot,const Color& color,const Rect& uv,const BlitMode blit_mode)
{
	float	hx=size.x*0.5f;
	float	hy=size.y*0.5f;
	Vec3f	n(0.0f,1.0f,0.0f);
	Vec3f	pos[4];

	pos[0]=-CAMERA3D->_y_plane_right_axis*hx-CAMERA3D->_y_plane_normal_axis*hy;
	pos[1]= CAMERA3D->_y_plane_right_axis*hx-CAMERA3D->_y_plane_normal_axis*hy;
	pos[2]=-CAMERA3D->_y_plane_right_axis*hx+CAMERA3D->_y_plane_normal_axis*hy;
	pos[3]= CAMERA3D->_y_plane_right_axis*hx+CAMERA3D->_y_plane_normal_axis*hy;

	for (int i=0; i<4; i++) 
	{
		rot.rotate_point(pos[i],pos[i]);
		pos[i]=p+pos[i];
	}

	Quad		quad;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);

	SET_VERTEX(quad.vertex[0],pos[0].x,pos[0].y,pos[0].z,n.x,n.y,n.z,c,uv.x1,uv.y1);
	SET_VERTEX(quad.vertex[1],pos[1].x,pos[1].y,pos[1].z,n.x,n.y,n.z,c,uv.x2,uv.y1);
	SET_VERTEX(quad.vertex[2],pos[2].x,pos[2].y,pos[2].z,n.x,n.y,n.z,c,uv.x1,uv.y2);
	SET_VERTEX(quad.vertex[3],pos[3].x,pos[3].y,pos[3].z,n.x,n.y,n.z,c,uv.x2,uv.y2);

	BlitGroup *grp=request_blit_group(texture,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add_perp(const long texture,const Vec3f& p,const Vec2f& size,const Color& color,const Rect& uv,const BlitMode blit_mode)
{
	float	hx=size.x*0.5f;
	Vec3f	n=CAMERA3D->_y_plane_normal_axis;
	Vec3f	pos[4];

	pos[0]=p-CAMERA3D->_y_plane_normal_axis*hx+Vec3f(0.0f,size.y,0.0f);
	pos[1]=p+CAMERA3D->_y_plane_normal_axis*hx+Vec3f(0.0f,size.y,0.0f);
	pos[2]=p-CAMERA3D->_y_plane_normal_axis*hx;
	pos[3]=p+CAMERA3D->_y_plane_normal_axis*hx;

	Quad		quad;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);

	SET_VERTEX(quad.vertex[0],pos[0].x,pos[0].y,pos[0].z,n.x,n.y,n.z,c,uv.x1,uv.y1);
	SET_VERTEX(quad.vertex[1],pos[1].x,pos[1].y,pos[1].z,n.x,n.y,n.z,c,uv.x2,uv.y1);
	SET_VERTEX(quad.vertex[2],pos[2].x,pos[2].y,pos[2].z,n.x,n.y,n.z,c,uv.x1,uv.y2);
	SET_VERTEX(quad.vertex[3],pos[3].x,pos[3].y,pos[3].z,n.x,n.y,n.z,c,uv.x2,uv.y2);

	BlitGroup *grp=request_blit_group(texture,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add_quad(const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Vec3f& p4,const Color& color,const BlitMode blit_mode)
{
	Quad		quad;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);
	Vec3f		n=cross_product(p2-p1,p3-p1); n.normalize();

	SET_VERTEX(quad.vertex[0],p1.x,p1.y,p1.z,n.x,n.y,n.z,c,0.0f,0.0f);
	SET_VERTEX(quad.vertex[1],p2.x,p2.y,p2.z,n.x,n.y,n.z,c,0.0f,0.0f);
	SET_VERTEX(quad.vertex[2],p3.x,p3.y,p3.z,n.x,n.y,n.z,c,0.0f,0.0f);
	SET_VERTEX(quad.vertex[3],p4.x,p4.y,p4.z,n.x,n.y,n.z,c,0.0f,0.0f);

	BlitGroup *grp=request_blit_group(-1,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add_quad(const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Vec3f& p4,const Color& c1,const Color& c2,const Color& c3,const Color& c4,const Vec3f& n,const BlitMode blit_mode)
{
	Quad		quad;

	SET_VERTEX(quad.vertex[0],p1.x,p1.y,p1.z,n.x,n.y,n.z,D3DCOLOR_COLORVALUE(c1.r,c1.g,c1.b,c1.a),0.0f,0.0f);
	SET_VERTEX(quad.vertex[1],p2.x,p2.y,p2.z,n.x,n.y,n.z,D3DCOLOR_COLORVALUE(c2.r,c2.g,c2.b,c2.a),0.0f,0.0f);
	SET_VERTEX(quad.vertex[2],p3.x,p3.y,p3.z,n.x,n.y,n.z,D3DCOLOR_COLORVALUE(c3.r,c3.g,c3.b,c3.a),0.0f,0.0f);
	SET_VERTEX(quad.vertex[3],p4.x,p4.y,p4.z,n.x,n.y,n.z,D3DCOLOR_COLORVALUE(c4.r,c4.g,c4.b,c4.a),0.0f,0.0f);

	BlitGroup *grp=request_blit_group(-1,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add_quad(const long texture,const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Vec3f& p4,const Color& color,const Rect& uv,const BlitMode blit_mode)
{
	Quad		quad;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);
	Vec3f		n=cross_product(p3-p1,p2-p1); n.normalize();

	SET_VERTEX(quad.vertex[0],p1.x,p1.y,p1.z,n.x,n.y,n.z,c,uv.x1,uv.y2);
	SET_VERTEX(quad.vertex[1],p2.x,p2.y,p2.z,n.x,n.y,n.z,c,uv.x2,uv.y2);
	SET_VERTEX(quad.vertex[2],p3.x,p3.y,p3.z,n.x,n.y,n.z,c,uv.x2,uv.y1);
	SET_VERTEX(quad.vertex[3],p4.x,p4.y,p4.z,n.x,n.y,n.z,c,uv.x1,uv.y1);

	BlitGroup *grp=request_blit_group(texture,D3DTEXF_POINT,4,6,blit_mode);

	memcpy(grp->vptr,&quad,sizeof(Quad));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Quad);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+3; grp->iptr++;

	grp->vertex_count+=4;
	grp->index_count+=6;
}

void BlitManager3d::add_line(const Vec3f& p1,const Vec3f& p2,const float width,const Color& color1,const Color& color2,const BlitMode blit_mode)
{
	Vec3f delta=p2-p1; delta.z=0.0f; delta.normalize(); delta.set(-delta.y*width,delta.x*width,0.0f);

	add_quad(p1-delta,p1+delta,p2+delta,p2-delta,color1,color1,color2,color2,Vec3f(0.0f,1.0f,0.0f),blit_mode);
}

void BlitManager3d::add_triangle(const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Color& color,const BlitMode blit_mode)
{
	Tri			tri;
	D3DCOLOR	c=D3DCOLOR_COLORVALUE(color.r,color.g,color.b,color.a);
	Vec3f		n=cross_product(p2-p1,p3-p1); n.normalize();

	SET_VERTEX(tri.vertex[0],p1.x,p1.y,p1.z,n.x,n.y,n.z,c,0.0f,0.0f);
	SET_VERTEX(tri.vertex[1],p2.x,p2.y,p2.z,n.x,n.y,n.z,c,0.0f,0.0f);
	SET_VERTEX(tri.vertex[2],p3.x,p3.y,p3.z,n.x,n.y,n.z,c,0.0f,0.0f);

	BlitGroup *grp=request_blit_group(-1,D3DTEXF_POINT,3,3,blit_mode);

	memcpy(grp->vptr,&tri,sizeof(Tri));
	grp->vptr=((BYTE*)grp->vptr)+sizeof(Tri);

	*(grp->iptr)=(WORD)grp->vertex_count; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+1; grp->iptr++;
	*(grp->iptr)=(WORD)grp->vertex_count+2; grp->iptr++;

	grp->vertex_count+=3;
	grp->index_count+=3;
}

void BlitManager3d::draw()
{
	GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0.0f,0.0f,0.0f,1.0f));

	// Render states
	_material->set_color_vertex(true,false);
	_material->set_cull_mode(D3DCULL_NONE);
	
	// Render
	BlitGroups::iterator i;
	for (i=_active_blits.begin(); i!=_active_blits.end(); ++i)
	{
		if (*i)
		{
			if (((*i)->vertex_count>0) && ((*i)->index_count>0))
			{
				// Set the texture
				if ((*i)->texture_id==-1) { _material->set_shader("basic"); _material->set_texture(0,-1); }
				else 
				{ 
					_material->set_shader("basic");
					_material->set_texture_sampling(0,(*i)->filter);
					_material->set_texture(0,(*i)->texture_id); 
				}

				switch ((*i)->blit_mode)
				{
					case BLIT_NORMAL:
						_material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);
						_material->set_z_write(true);
						_material->set_alpha_test(true,10);
						break;
					case BLIT_NORMAL_NOZWRITE:
						_material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);
						_material->set_z_write(false);
						_material->set_alpha_test(true,10);
						break;
					case BLIT_PREMULTALPHA:
						_material->set_alpha(true,D3DBLEND_ONE,D3DBLEND_INVSRCALPHA);
						_material->set_z_write(false);
						_material->set_alpha_test(true,2);
						break;
					case BLIT_ADD:
						_material->set_alpha(true,D3DBLEND_ONE,D3DBLEND_ONE);
						_material->set_z_write(true);
						_material->set_alpha_test(true,2);
						break;
					case BLIT_ADD_NOZWRITE:
						_material->set_alpha(true,D3DBLEND_ONE,D3DBLEND_ONE);
						_material->set_z_write(false);
						_material->set_alpha_test(true,2);
						break;
					case BLIT_SRCALPHA_ONE:
						_material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_ONE);
						_material->set_z_write(false);
						_material->set_alpha_test(true,2);
						break;
				}
				// Unlock vertex buffers, etc	
				if ((*i)->vptr) { (*i)->vertex_buffer->release_write_pointer(); (*i)->vptr=NULL; }
				if ((*i)->iptr) { (*i)->index_buffer->release_write_pointer(); (*i)->iptr=NULL; }

				_material->set(D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE | D3DFVF_TEX1);

				GFX->render(D3DPT_TRIANGLELIST,(*i)->vertex_buffer,(*i)->vertex_count,(*i)->index_buffer,(*i)->index_count);
			}
		}
	}
}

VertexBuffer* BlitManager3d::create_vertex_buffer(long size)
{
	VertexBuffer* tmp=new VertexBuffer;

	if (tmp->create(size,sizeof(Vertex),D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE | D3DFVF_TEX1,true))
	{
		return tmp;
	}

	delete tmp;

	ERROR_MSG("Can't create vertex buffer!");

	return NULL;
}

IndexBuffer* BlitManager3d::create_index_buffer(long size)
{
	IndexBuffer* tmp=new IndexBuffer;

	if (tmp->create(size,true))
	{
		return tmp;
	}

	delete tmp;

	ERROR_MSG("Can't create index buffer!");

	return NULL;
}

BlitManager3d::BlitGroup *BlitManager3d::request_blit_group(long texture,D3DTEXTUREFILTERTYPE filter,const long additional_verts,const long addition_indexes,const BlitMode blit_mode)
{
	BlitGroup *grp=NULL;

	// Seek proper blit group
	BlitGroups::iterator i;
	for (i=_blits.begin(); i!=_blits.end(); ++i)
	{
		if (((*i)->texture_id==texture) && ((*i)->filter==filter) && ((*i)->blit_mode==blit_mode))
		{
			grp=*i;
			break;
		}
	}
	if (!grp)
	{
		grp=new BlitGroup;
		grp->active=false;
		grp->vallocated=0;
		grp->iallocated=0;
		grp->vertex_count=grp->index_count=0;
		grp->vertex_buffer=NULL;
		grp->index_buffer=NULL;
		grp->texture_id=texture;
		grp->filter=filter;
		grp->vptr=NULL;
		grp->iptr=NULL;
		grp->blit_mode=blit_mode;
		_blits.push_back(grp);
	}

	if (!grp->active) 
	{
		_active_blits.push_back(grp);
		grp->active=true;
	}

	if (!grp->vertex_buffer)
	{
		grp->vertex_buffer=create_vertex_buffer(MIN_VERTEX_BUFFER_SIZE);
		grp->vallocated=MIN_VERTEX_BUFFER_SIZE;
		grp->vptr=grp->vertex_buffer->get_write_pointer();
	}
	else
	{
		if (grp->vallocated<(grp->vertex_count+additional_verts))
		{
			// Resize
			long					new_size=Max(grp->vallocated*10,(grp->vertex_count+additional_verts));
			VertexBuffer*			tmp=create_vertex_buffer(new_size);
			void*					ptr=tmp->get_write_pointer();

			// Copy old data
			grp->vptr=grp->vertex_buffer->get_write_pointer();
			memcpy(ptr,grp->vptr,sizeof(Vertex)*grp->vertex_count);
			// Delete old vb
			grp->vertex_buffer->release_write_pointer();
			delete grp->vertex_buffer;
			grp->vertex_buffer=tmp;
			grp->vptr=((BYTE*)ptr)+(sizeof(Vertex)*grp->vertex_count);
			grp->vallocated=new_size;
		}
	}
	if (!grp->index_buffer)
	{
		grp->index_buffer=create_index_buffer(MIN_INDEX_BUFFER_SIZE);
		grp->iallocated=MIN_INDEX_BUFFER_SIZE;
		grp->iptr=grp->index_buffer->get_write_pointer();
	}
	else
	{
		if (grp->iallocated<(grp->index_count+addition_indexes))
		{
			// Resize
			long					new_size=Max(grp->iallocated*10,(grp->index_count+addition_indexes));
			IndexBuffer*			tmp=create_index_buffer(new_size);
			WORD*					ptr=tmp->get_write_pointer();

			// Copy old data
			grp->iptr=grp->index_buffer->get_write_pointer();
			memcpy(ptr,grp->iptr,sizeof(WORD)*grp->index_count);
			// Delete old vb
			grp->index_buffer->release_write_pointer();
			delete grp->index_buffer;
			grp->index_buffer=tmp;
			grp->iptr=((WORD*)ptr)+(grp->index_count);
			grp->iallocated=new_size;
		}
	}
	if (!grp->vptr)
	{
		grp->vptr=grp->vertex_buffer->get_write_pointer();
	}
	if (!grp->iptr)
	{
		grp->iptr=grp->index_buffer->get_write_pointer();
	}
	return grp;
}

void BlitManager3d::flush()
{
	draw();
	soft_clear();
}
