// BlitManager3d.h: interface for the BlitManager3d class.
//
//////////////////////////////////////////////////////////////////////

#ifndef __BLITMANAGER3D_H
#define __BLITMANAGER3D_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Defs.h"
#include <vector>

#define MIN_VERTEX_BUFFER_SIZE	100
#define MIN_INDEX_BUFFER_SIZE	100

class Camera;
class Material;
class VertexBuffer;
class IndexBuffer;

class BlitManager3d  
{
protected:
	struct Vertex
	{
		Vec3f		p;
		Vec3f		n;
		DWORD		color;
		float		u,v;
	};
	struct Quad
	{
		Vertex	vertex[4];
	};
	struct Tri
	{
		Vertex	vertex[3];
	};
	struct BlitGroup
	{
		bool					active;
		long					texture_id;
		D3DTEXTUREFILTERTYPE	filter;
		long					index_count;
		long					vertex_count;
		VertexBuffer*			vertex_buffer;
		IndexBuffer*			index_buffer;
		long					vallocated;
		long					iallocated;
		void*					vptr;
		WORD*					iptr;
		BlitMode				blit_mode;
	};
	typedef std::vector<BlitGroup*>	BlitGroups;
	class BlitSortFunction;
public:
	BlitManager3d();
	virtual ~BlitManager3d();

	void	add(const long texture,const Vec3f& p,const Vec2f& size,const Color& color,const Rect& uv,const BlitMode blit_mode=BLIT_NORMAL);
	void	add(const long texture,const Vec3f& p,const Rect& r,const Color& color,const Rect& uv,const BlitMode blit_mode=BLIT_NORMAL);
	void	add(const long texture,const Vec3f& p,const Rect& r,const Quaternion& rot,const Color& color,const Rect& uv,const BlitMode blit_mode=BLIT_NORMAL);
	void	add(const long texture,const Vec3f& p,const Vec3f& up,const Vec3f& right,const Vec3f& normal,const Color& color,const Rect& uv,const BlitMode blit_mode=BLIT_NORMAL);
	void	add_ground(const long texture,const Vec3f& p,const Vec2f& size,const Color& color,const Rect& uv,const BlitMode blit_mode=BLIT_NORMAL);
	void	add_ground(const long texture,const Vec3f& p,const Vec2f& size,const Quaternion& rot,const Color& color,const Rect& uv,const BlitMode blit_mode=BLIT_NORMAL);
	void	add_perp(const long texture,const Vec3f& p,const Vec2f& size,const Color& color,const Rect& uv,const BlitMode blit_mode=BLIT_NORMAL);
	void	add_billboard(const long texture,const Vec3f& p,const Vec2f& size,const Color& color,const Rect& uv,const BlitMode blit_mode=BLIT_NORMAL);
	void	add_billboard(const long texture,const Vec3f& p,const Vec2f& size,const Quaternion& rot,const Color& color,const Rect& uv,const BlitMode blit_mode=BLIT_NORMAL);
	void	add_triangle(const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Color& color,const BlitMode blit_mode=BLIT_NORMAL);
	void	add_quad(const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Vec3f& p4,const Color& color,const BlitMode blit_mode=BLIT_NORMAL);
	void	add_quad(const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Vec3f& p4,const Color& c1,const Color& c2,const Color& c3,const Color& c4,const Vec3f& n,const BlitMode blit_mode=BLIT_NORMAL);
	void	add_quad(const long texture,const Vec3f& p1,const Vec3f& p2,const Vec3f& p3,const Vec3f& p4,const Color& color,const Rect& uv,const BlitMode blit_mode=BLIT_NORMAL);
	void	add_line(const Vec3f& p1,const Vec3f& p2,const float width,const Color& color1,const Color& color2,const BlitMode blit_mode=BLIT_NORMAL);

	void	flush();
	void	draw();
	void	soft_clear();
	void	clear_cache();

protected:
	VertexBuffer*			create_vertex_buffer(long size);
	IndexBuffer*			create_index_buffer(long size);
	BlitGroup*				request_blit_group(long texture,D3DTEXTUREFILTERTYPE filter,const long additional_verts,const long addition_indexes,const BlitMode blit_mode);

protected:
	BlitGroups			_blits;
	BlitGroups			_active_blits;
	Material*			_material;
};

extern BlitManager3d* BLIT3D;

#endif // !defined(AFX_BLITMANAGER_H__2AD7D1D1_ECF1_4779_B147_598AF5F61AAA__INCLUDED_)
