#include "Explosion.h"
#include "BlitManager3d.h"
#include "RndGenerator.h"
#include "TextureManager.h"
#include "Camera.h"
#include "Screen.h"
#include "Sprite3d.h"
#include "Utils.h"

Explosion::Explosion(const Vec3f& pos,const Vec3f& direction,const float radius,const bool show_shadow,const Vec3f& shadow_pos,Object3d* object) : Object3d("explosion",NULL)
{
	_pos=pos;
	_direction=direction;
	_shadow_pos=shadow_pos;
	_show_shadow=show_shadow;
	_object=object;
	_t=0;
	_radius=radius;
	_delete_effect_on_end=false;
	if (object)
	{
		_object=object;
		_object->set_draw(false);
	}
	else _object=NULL;

	_finished=false;
	build_explosion_objects();
}

Explosion::~Explosion(void)
{
	_smoke.clear();
}

void Explosion::draw()
{
	if (!_draw) return;

	if (GFX->get_pass()==PASS_SOLID) return;

	if (_show_shadow)
	{
		float r=_radius*5.0f;
		Quaternion	o; o.rotation_quaternion(0.0f,1.0f,0.0f,_shadow_angle);

		if (_t<200.0f) 
		{
			float alpha=Clamp(_t/200.0f,0.0f,1.0f);
			r=r*alpha;
			BLIT3D->add_ground(_shadow_img,_shadow_pos,Vec2f(r,r),o,Color(0.0f,0.0f,0.0f,alpha),Rect(0.0f,0.0f,1.0f,1.0f),BLIT_NORMAL_NOZWRITE);
		}
		else if ((_t>=5000.0f) && (_delete_effect_on_end))
		{
			float alpha=1.0f-Clamp((_t-5000.0f)/500.0f,0.0f,1.0f);
			BLIT3D->add_ground(_shadow_img,_shadow_pos,Vec2f(r,r),o,Color(0.0f,0.0f,0.0f,alpha),Rect(0.0f,0.0f,1.0f,1.0f),BLIT_NORMAL_NOZWRITE);
		}
		else
		{
			BLIT3D->add_ground(_shadow_img,_shadow_pos,Vec2f(r,r),o,Color(0.0f,0.0f,0.0f,1.0f),Rect(0.0f,0.0f,1.0f,1.0f),BLIT_NORMAL_NOZWRITE);
		}
	}

	auto it=_flash.begin();
	while (it!=_flash.end())
	{
		BLIT3D->add_billboard(it->_img,_pos+it->_pos,it->_size,it->_color,it->_uv,BLIT_ADD_NOZWRITE);

		++it;
	}

	Quaternion	ori;

	it=_smoke.begin();
	while (it!=_smoke.end())
	{
		ori.rotation_quaternion(CAMERA3D->_normal_axis,it->_angle);
		BLIT3D->add_billboard(it->_img,_pos+it->_pos,it->_size,ori,it->_color,it->_uv,BLIT_PREMULTALPHA);

		++it;
	}

	it=_sparks.begin();
	while (it!=_sparks.end())
	{
		Vec3f dir=it->_pos-it->_old_pos; 
		float len=dir.normalize();
		if (len>0.0f)
		{
			Vec3f up=CAMERA3D->_normal_axis;
			Vec3f right=cross_product(dir,up);
			up=cross_product(right,dir);

			BLIT3D->add(it->_img,_pos+it->_pos+dir*len*0.5f,dir*len*0.5f,right*it->_size.x*0.5f,up,it->_color,it->_uv,BLIT_SRCALPHA_ONE);
		}

		++it;
	}

	it=_smoke_trails.begin();
	while (it!=_smoke_trails.end())
	{
		Vec3f dir=it->_pos-it->_old_pos; 
		float len=it->_size.y;
		if (len>0.0f)
		{
			Vec3f up=CAMERA3D->_normal_axis;
			Vec3f right=cross_product(dir,up);
			up=cross_product(right,dir);

			BLIT3D->add(it->_img,_pos+it->_pos+dir*len*0.5f,dir*len*0.5f,right*it->_size.x*0.5f,up,it->_color,it->_uv,BLIT_SRCALPHA_ONE);
		}

		++it;
	}

	it=_debris.begin();
	while (it!=_debris.end())
	{
		ori.rotation_quaternion(CAMERA3D->_normal_axis,it->_angle);
		BLIT3D->add_billboard(it->_img,_pos+it->_pos,it->_size,ori,it->_color,it->_uv,BLIT_NORMAL_NOZWRITE);

		++it;
	}
}

void Explosion::anim(const float in_t)
{
	float t=in_t;
	_t+=t;

	float	tt;
	auto	it=_smoke.begin();
	while (it!=_smoke.end())
	{
		if (it->_time_to_wake>t)
		{
			it->_time_to_wake-=t;
			++it;
		}
		else
		{
			it->_time_to_wake=0;
			it->_current_life-=t;
			if (it->_current_life<0.0f)
			{
				it=_smoke.erase(it);
			}
			else
			{
				tt=it->_current_life/it->_total_life;

				it->_old_pos=it->_pos;
				it->_pos=it->_pos+it->_pos_inc*t;
				it->_color=it->_color+it->_color_inc*t;
				if (tt<0.25f) it->_color.a=tt*4.0f;
				else it->_color.a=1.0f;
				it->_size=it->_size+it->_size_inc*t;
				it->_angle+=it->_angle_inc*t;

				++it;
			}
		}
	}

	it=_flash.begin();
	while (it!=_flash.end())
	{
		if (it->_time_to_wake>t)
		{
			it->_time_to_wake-=t;
			++it;
		}
		else
		{
			it->_current_life-=t;
			if (it->_current_life<0.0f)
			{
				it=_flash.erase(it);
			}
			else
			{
				tt=it->_current_life/it->_total_life;

				if (tt<0.5f) it->_size=it->_size-it->_size_inc*t*0.5f;
				else it->_size=it->_size+it->_size_inc*t*0.5f;

				++it;
			}
		}
	}

	it=_sparks.begin();
	while (it!=_sparks.end())
	{
		if (it->_time_to_wake>t)
		{
			it->_time_to_wake-=t;
			++it;
		}
		else
		{
			it->_time_to_wake=0;
			it->_current_life-=t;
			if (it->_current_life<0.0f)
			{
				it=_sparks.erase(it);
			}
			else
			{
				tt=it->_current_life/it->_total_life;

				if ((it->_old_pos-it->_pos).norm_squared()>Sqr(it->_size.y))
				{
					it->_old_pos=it->_pos+(it->_old_pos-it->_pos)*it->_size.y;
				}
				it->_pos=it->_pos+it->_pos_inc*t;
				if (tt<0.25f) it->_color.a=tt*4.0f;
				else it->_color.a=1.0f;

				++it;
			}
		}
	}

	it=_smoke_trails.begin();
	while (it!=_smoke_trails.end())
	{
		if (it->_time_to_wake>t)
		{
			it->_time_to_wake-=t;
			++it;
		}
		else
		{
			it->_time_to_wake=0;
			it->_current_life-=t;
			if (it->_current_life<0.0f)
			{
				it=_smoke_trails.erase(it);
			}
			else
			{
				tt=it->_current_life/it->_total_life;

				it->_pos=it->_pos+it->_pos_inc*t;
				it->_size=it->_size+it->_size_inc*t;
				if (tt<0.2f) it->_color.a=tt*5.0f;
				else it->_color.a=1.0f;

				++it;
			}
		}
	}

	it=_debris.begin();
	while (it!=_debris.end())
	{
		if (it->_time_to_wake>t)
		{
			it->_time_to_wake-=t;
			++it;
		}
		else
		{
			it->_time_to_wake=0;
			it->_current_life-=t;
			if (it->_current_life<0.0f)
			{
				it=_debris.erase(it);
			}
			else
			{
				tt=it->_current_life/it->_total_life;

				it->_pos=it->_pos+it->_pos_inc*t;
				float speed=it->_pos_inc.norm();
				it->_pos_inc=it->_pos_inc+it->_accel*t; it->_pos_inc.normalize(); it->_pos_inc=it->_pos_inc*speed;
				it->_size=it->_size+it->_size_inc*t;
				if (tt>0.5f) it->_color=it->_color+it->_color_inc*t;
				if (tt<0.2f) it->_color.a=tt*5.0f;
				else it->_color.a=1.0f;
				it->_angle+=it->_angle_inc*t;

				++it;
			}
		}
	}
	
	if ((_smoke.empty()) && (_flash.empty()) && (_sparks.empty()) && (_smoke_trails.empty()) && (_debris.empty()) && (_t>6000.0f))
	{ 
		_finished=true;
		if ((_delete_effect_on_end) || (!_show_shadow))
		{
			remove_object();
			return;
		}
	}
}

void Explosion::build_explosion_objects()
{
	char buffer[512];

	// Smoke sprites
	for (int i=0; i<g_n_smoke; i++)
	{
		float ang=deg_to_rad(g_rnd.get_random_f(0.0f,360.0f));
		float r=g_rnd.get_random_f(0.0f,_radius);

		sprintf_s((char*)&buffer,512,"exp_cloud%02i",g_rnd.get_random_i(1,4));

		Particle smoke;
		smoke._time_to_wake=g_rnd.get_random_f(0.0f,200.0f);
		smoke._total_life=smoke._current_life=g_rnd.get_random_f(500.0f,700.0f);
		smoke._pos.set(r*cosf(ang),g_rnd.get_random_f(-_radius*0.5f,_radius*0.5f),r*sinf(ang));
		smoke._img=TEXTURES->find_by_name(buffer);
		smoke._color.set(0.6f,g_rnd.get_random_f(0.2f,0.4f),0.1f,1.0f);
		smoke._color_inc=(Color(0.0f,0.0f,0.0f,1.0f)-smoke._color)/smoke._total_life;
		smoke._size.set(1.5f,1.5f);
		smoke._size_inc=Vec2f(2.5f,2.0f)/smoke._total_life;
		smoke._pos_inc.set(0.0f,0.0025f,0.0f);
		smoke._angle=g_rnd.get_random_f(0.0,360.0f);
		smoke._angle_inc=g_rnd.get_random_f(-45.0f,45.0f)/1000.0f;
		smoke._uv.set(0.0f,0.0f,1.0f,1.0f);
		smoke._accel.set(0.0f,0.0f,0.0f);

		_smoke.push_back(smoke);
	}//*/

	// Flash sprites
	for (int i=0; i<g_n_flash; i++)
	{
		float ang=deg_to_rad(g_rnd.get_random_f(0.0f,360.0f));
		float r=g_rnd.get_random_f(0.0f,_radius*0.5f);

		sprintf_s((char*)&buffer,512,"exp_flash%02i",g_rnd.get_random_i(1,4));

		Particle flash;
		flash._time_to_wake=g_rnd.get_random_f(0.0f,150.0f);
		flash._total_life=flash._current_life=g_rnd.get_random_f(150.0f,250.0f);
		flash._pos.set(r*cosf(ang),0.0f,r*sinf(ang));
		flash._img=TEXTURES->find_by_name(buffer);
		flash._color.set(0.20f,0.15f,0.10f,1.0f);
		flash._size.set(_radius,_radius);
		float si=4.0f*_radius/(flash._total_life*0.5f);
		flash._size_inc.set(si,si);
		flash._uv.set(0.0f,0.0f,1.0f,1.0f);
		flash._accel.set(0.0f,0.0f,0.0f);
		flash._angle=g_rnd.get_random_f(0.0,360.0f);
		flash._angle_inc=g_rnd.get_random_f(-95.0f,95.0f)/1000.0f;

		_flash.push_back(flash);
	}//*/

	// Sparks sprites
	long spark_tex=TEXTURES->find_by_name("exp_sparks");
	for (int i=0; i<g_n_sparks; i++)
	{
		float r=g_rnd.get_random_f(0.0f,_radius*0.15f);

		Particle spark;
		spark._time_to_wake=0.0f; 
		spark._total_life=spark._current_life=g_rnd.get_random_f(400.0f,600.0f);
		spark._pos_inc.set(g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f)); spark._pos_inc.normalize(); 
		if (!_direction.is_null())
		{
			while (dot_product(_direction,spark._pos_inc)<0.1f)
			{
				spark._pos_inc.set(g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f)); spark._pos_inc.normalize(); 
			}
		}
		spark._pos=spark._old_pos=spark._pos_inc*r; 
		spark._pos_inc=spark._pos_inc*0.008f;
		spark._img=spark_tex;
		spark._color.set(0.80f,0.3f,0.10f,1.0f);
		spark._color_inc.set(0.0f,0.0f,0.0f,0.0f);
		spark._size.set(_radius*0.25f,_radius*4.0f);
		spark._size_inc.set(0.0f,0.0f);
		spark._uv.set(0.0f,0.0f,1.0f,1.0f);
		spark._accel.set(0.0f,0.0f,0.0f);

		_sparks.push_back(spark);
	}//*/

	// Smoke trails sprites
	for (int i=0; i<g_n_smoke_trails; i++)
	{
		float r=g_rnd.get_random_f(0.0f,_radius*0.15f);

		sprintf_s((char*)&buffer,512,"exp_smoketrails%02i",g_rnd.get_random_i(1,3));

		Particle trail;
		trail._time_to_wake=g_rnd.get_random_f(0.0f,150.0f);
		trail._total_life=trail._current_life=g_rnd.get_random_f(600.0f,800.0f);
		trail._pos_inc.set(g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f)); trail._pos_inc.normalize(); 
		if (!_direction.is_null())
		{
			while (dot_product(_direction,trail._pos_inc)<0.1f)
			{
				trail._pos_inc.set(g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f)); trail._pos_inc.normalize(); 
			}
		}
		trail._pos=trail._old_pos=trail._pos_inc*r; 
		trail._pos_inc=trail._pos_inc*0.005f; trail._pos=trail._pos+trail._pos_inc*100.0f;
		trail._pos_inc.set(0.0f,0.0f,0.0f);
		trail._img=TEXTURES->find_by_name(buffer);
		trail._color.set(1.0f,0.6f,0.1f,1.0f);
		trail._color_inc.set(0.0f,0.0f,0.0f,0.0f);
		trail._size.set(_radius*0.25f,0.0f);
		trail._size_inc=(Vec2f(_radius*0.5f,_radius*6.0f)-trail._size)/trail._total_life;
		trail._uv.set(0.0f,0.0f,1.0f,1.0f);
		trail._accel.set(0.0f,0.0f,0.0f);

		_smoke_trails.push_back(trail);
	}//*/

	Sprite3d* sprite=dynamic_cast<Sprite3d*>(_object);
	if (sprite)
	{
		float	n=ceilf(sqrtf((float)g_n_debris));
		long	debris_tex=sprite->get_image();
		float	sx=sprite->get_size().x;
		float	sy=sprite->get_size().y;
		Rect	base_uv=sprite->get_frame_uv();
		Vec3f	p_base=Vec3f(-sx*0.5f,sy,0.0f);
		sx=sx/n;
		sy=sy/n;
		p_base=p_base+Vec3f(sx*0.5f,sy*0.5f,0.0f);
		float	tu=(base_uv.x2-base_uv.x1)/n;
		float	tv=(base_uv.y2-base_uv.y1)/n;

		for (int yy=0; yy<(int)n; yy++)
		{
			for (int xx=0; xx<(int)n; xx++)
			{
				float r=g_rnd.get_random_f(0.0f,_radius*0.15f);

				Particle debris;
				debris._time_to_wake=0.0f; 
				debris._total_life=debris._current_life=g_rnd.get_random_f(700.0f,900.0f);
				debris._pos=p_base;
				debris._pos.x+=((float)xx)*sx;
				debris._pos.y-=((float)yy)*sy;
				debris._pos_inc=debris._pos; debris._pos_inc.z=g_rnd.get_random_f(-1.0f,1.0f); debris._pos_inc.normalize();
				debris._pos_inc.set(debris._pos_inc.x*0.02f,debris._pos_inc.y*0.015f,debris._pos_inc.z*0.02f);
				debris._img=debris_tex;
				debris._color.set(1.0f,0.8f,0.25f,1.0f);
				debris._color_inc=(Color(0.8f,0.8f,0.8f,1.0f)-debris._color)/(debris._total_life*0.5f);
				debris._size.set(0.5f,0.5f);
				debris._size_inc.set(0.0f,0.0f);
				debris._uv.set(0.0f,0.0f,1.0f,1.0f);
				debris._accel.set(0.0f,-0.00005f,0.0f);
				debris._angle=0.0f;
				debris._angle_inc=g_rnd.get_random_f(-145.0f,145.0f)/1000.0f;

				debris._uv.x1=base_uv.x1+tu*(float)xx;
				debris._uv.y1=base_uv.y1+tv*(float)yy;
				debris._uv.x2=base_uv.x1+tu*(float)(xx+1);
				debris._uv.y2=base_uv.y1+tv*(float)(yy+1);

				_debris.push_back(debris);
			}
		}
	}
	else
	{
		// Debris sprites
		long debris_tex=TEXTURES->find_by_name("exp_debris");
		for (int i=0; i<g_n_debris; i++)
		{
			float r=g_rnd.get_random_f(0.0f,_radius*0.15f);

			Particle debris;
			debris._time_to_wake=0.0f; 
			debris._total_life=debris._current_life=g_rnd.get_random_f(700.0f,900.0f);
			debris._pos_inc.set(g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f)); debris._pos_inc.normalize(); 
			if (!_direction.is_null())
			{
				while (dot_product(_direction,debris._pos_inc)<0.1f)
				{
					debris._pos_inc.set(g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f),g_rnd.get_random_f(-1.0f,1.0f)); debris._pos_inc.normalize(); 
				}
			}
			debris._pos=debris._old_pos=debris._pos_inc*r; 
			debris._pos_inc=debris._pos_inc*0.02f;
			debris._img=debris_tex;
			debris._color.set(1.0f,0.8f,0.1f,1.0f);
			debris._color_inc=(Color(0.8f,0.8f,0.8f,1.0f)-debris._color)/(debris._total_life*0.5f);
			debris._size.set(0.5f,0.5f);
			debris._size_inc.set(0.0f,0.0f);
			debris._uv.set(0.0f,0.0f,1.0f,1.0f);
			debris._accel.set(0.0f,-0.00005f,0.0f);
			debris._angle=g_rnd.get_random_f(0.0,360.0f);
			debris._angle_inc=g_rnd.get_random_f(-145.0f,145.0f)/1000.0f;

			int img=g_rnd.get_random_i(0,8);
			debris._uv.x1=(32.0f*(img%4))/128.0f;
			debris._uv.y1=(32.0f*(img/4))/128.0f;
			debris._uv.x2=((32.0f*((img%4)+1))-1)/128.0f;
			debris._uv.y2=((32.0f*((img/4)+1))-1)/128.0f;

			_debris.push_back(debris);
		}
	}//*/


	sprintf_s((char*)&buffer,512,"exp_flash%02i",g_rnd.get_random_i(1,4));
	
	_shadow_img=TEXTURES->find_by_name(buffer);
	_shadow_angle=g_rnd.get_random_f(0.0f,360.0f);
}

void Explosion::remove_object()
{

}
