#pragma once

#include "Object3d.h"
#include "Mesh.h"
#include "MathLib.h"

class Sprite3d;

class Explosion :
	public Object3d
{
protected:
	struct Particle
	{
		float		_time_to_wake;
		float		_current_life;
		float		_total_life;
		long		_img;
		Vec3f		_old_pos;
		Vec3f		_pos;
		Vec3f		_pos_inc;
		Vec3f		_accel;
		Vec2f		_size;
		Vec2f		_size_inc;
		Color		_color;
		Color		_color_inc;
		float		_angle;
		float		_angle_inc;
		Rect		_uv;
	};
	typedef std::list<Particle>	Particles;
public:
	Explosion(const Vec3f& pos,const Vec3f& direction,const float radius,const bool show_shadow,const Vec3f& shadow_pos,Object3d* object);
	virtual ~Explosion(void);

	virtual	void	draw();
	virtual void	anim(const float t);
	inline	void	set_delete_on_end(const bool b) { _delete_effect_on_end=b; }
	inline	bool	is_finished() { return _finished; }

protected:
			void	build_explosion_objects();
	virtual	void	remove_object();
protected:
	float			_t;
	float			_radius;
	Vec3f			_direction;
	Vec3f			_shadow_pos;
	Object3d*		_object;
	Particles		_smoke;
	Particles		_flash;
	Particles		_sparks;
	Particles		_smoke_trails;
	Particles		_debris;
	bool			_show_shadow;
	long			_shadow_img;
	float			_shadow_angle;
	bool			_delete_effect_on_end;
	bool			_finished;

protected:
	static const	int		g_n_smoke=30;
	static const	int		g_n_flash=10;
	static const	int		g_n_sparks=30;
	static const	int		g_n_smoke_trails=15;
	static const	int		g_n_debris=40;
};
