
#pragma once

#define USE_FREETYPE2

#ifdef USE_D3D9
# ifndef USE_FREETYPE2
#   include <d3d9.h>
#   include <d3dx9.h>
#	define USE_D3DFONT
#  else
#  endif
#else
# define USE_FREETYPE2
#endif

#ifdef USE_FREETYPE2
# include <ft2build.h>
# include FT_FREETYPE_H
# ifdef _DEBUG
#   pragma comment(lib, "freetype252MT_D.lib")
# else
#   pragma comment(lib, "freetype252MT.lib")
# endif
#endif

#include "MathLib.h"

#define TEXT_NORMAL			0
#define TEXT_CENTER			1
#define TEXT_BOTTOM			2
#define TEXT_SHADOW			4
#define TEXT_BACKGROUND		8
#define TEXT_OUTLINE		16

class Font
{
public:
	Font();
	virtual ~Font(void);

			bool	init(const int size,const char* name);
			void	write_text(const float x,const float y,const float z,const float r,const float g,const float b,const float a,const char *string,const unsigned long flags);
			Rect	get_text_rect(float x,float y,const char *string,const unsigned long flags);

	inline	void	set_outline(const int margin_x,const int margin_y,const Color& c) { _text_outline_margin_x=margin_x; _text_outline_margin_y=margin_y; _text_outline_color=c; }
	inline	void	set_background(const int margin_x,const int margin_y,const Color& c) { _text_background_margin_x=margin_x; _text_background_margin_y=margin_y; _text_background_color=c; }

			void	make_fixed_width();
protected:
#ifdef USE_FREETYPE2
			const char*	get_next_line(const char* start,int& n_chars);
			Vec2f		get_text_line_size(const char* start,const int n_chars,int& baseline_y);
			Vec2f		get_text_size(const char* start);
			bool		render_font_atlas(const int atlas_size,const int atlas_margin);
#endif
public:
	Color		_text_background_color;
	int			_text_background_margin_x;
	int			_text_background_margin_y;
	Color		_text_outline_color;
	int			_text_outline_margin_x;
	int			_text_outline_margin_y;

protected:
	float		_line_spacing;
	int			_shadow_offset;
	std::string	_font_name;
	long		_font_size;
#ifdef USE_D3DFONT
	LPD3DXFONT		_font;
	LPD3DXSPRITE	_sprite;
#endif
#ifdef USE_FREETYPE2
protected:
	FT_Face			_font; 
	Rect			_uv[256];
	int				_delta_x[256];
	int				_delta_y[256];
	int				_advance_x[256];
	int				_advance_y[256];
	int				_size_x[256];
	int				_size_y[256];
	long			_max_height;
	long			_max_top;
	long			_texture;
	bool			_texture_external;
protected:
	static FT_Library g_freetype2_lib;
	static int		  g_freetype2_lib_ref_count;
#endif
};
