// Gfx.h: interface for the Gfx class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_GFX_H__65AF6A21_DD51_4626_9C30_4141D1E621D0__INCLUDED_)
#define AFX_GFX_H__65AF6A21_DD51_4626_9C30_4141D1E621D0__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Defs.h"
#include "RenderEnv.h"

#ifdef USE_D3D9
#include "dx/dxutil.h"
#include "dx/d3denumeration.h"
#include "dx/d3dsettings.h"
#endif

class Light;
class VertexBuffer;
class IndexBuffer;
class VertexProgram;
class PixelShader;

class Gfx  
{
public:
	struct LightStruct
	{
		Vec4f	pos;
		Vec4f	dir;
		Vec4f	color;
		Vec4f	params;
	};
public:
	Gfx();
	virtual ~Gfx();

	// Init stuff and config stuff
			bool				init(HINSTANCE instance);
			void				shutdown();
	inline	void				set_window_mode(bool b) { _window_mode=b; }
	inline	bool				is_windowed() { return _window_mode; }
	inline	void				set_adapter(const int ad) { _adapter=ad; }
	inline	void				set_resolution(long rx,long ry,long bd,long rr) { _resolution.set(float(rx),float(ry)); _bitdepth=bd; _refresh_rate=rr; _vres=_resolution; }
	inline	void				set_virtual_resolution(long vx,long vy) { _vres.set((float)vx,(float)vy); }
	inline	Vec2f				get_real_resolution() { return _resolution; }
	inline	Vec2f				get_virtual_resolution() { return _vres; }
	inline	void				set_app_name(const std::string& str) { _app_name=str; }
	inline	void				set_window(HWND window) { _window=window; }
	inline	HWND				get_window() { return _window; }
	inline	unsigned long		get_window_style() { return _window_style; }

	// Real gfx stuff
			void		set_perspective(float fov,float aspect,float near_plane,float far_plane);
			void		set_2d_render();
			void		set_camera2d(float x,float y,float z);
			void		set_camera(float x,float y,float z,Quaternion rot);
			void		set_world_transformation(int idx,float x,float y,float z,Quaternion rot);
			void		set_world_transformation(int idx,float x,float y,float z,Quaternion rot,const float sx,const float sy,const float sz);
			void		unset_world_transformation(int idx);
			void		clear(DWORD clear_buffers,float r=0.0f,float g=0.0f,float b=0.0f,float a=1.0f,float zVal=1.0f,unsigned long sVal=0);
			void		begin_scene();
			void		end_scene();
			void		swap_buffers();
	inline	void		set_color_modulation(float r,float g,float b) { _modulate.x=r; _modulate.y=g; _modulate.z=b; }
			Vec3f		get_color_modulation() { return _modulate; }
			void		clear_lights();
			void		set_fog(const bool enable,const Color& color,const float start,const float end,const Vec3f& center);
	inline	Vec3f		get_fog_center() { return _fog_center; }
	inline	Vec2f		get_fog_range() { return _fog_range; }
	inline	Color		get_fog_color() { return _fog_color; }
			void		add_light(Light* lgt);
	inline	void		enable_z(const bool b)
	{
#ifdef USE_D3D9
		set_renderstate(D3DRS_ZENABLE,(b)?(1):(0));
#endif
	}
	inline	void		set_ambient_light(const Color& c)
	{
		_ambient_light=c;
#ifdef USE_D3D9
		set_renderstate(D3DRS_AMBIENT,float_to_colorref(c.r,c.g,c.b,1.0f));
#endif
	}
	inline	Color		get_ambient_light()
	{
		return _ambient_light;
	}
			void		set_pass(const PassType pass);
	inline	PassType	get_pass() { return _pass; }
	inline	RenderEnv*	get_render_env() { return &_env; }
			void		render(D3DPRIMITIVETYPE prim_type,VertexBuffer* vb,const UINT vcount,IndexBuffer* ib=NULL,const UINT icount=0);
			TextureID	load_texture(const std::string& filename,int& width,int& height);
			TextureID	new_texture(const int sx,const int sy,const D3DFORMAT format,const bool rt=false,ZBufferID* z_buffer=NULL);
			ZBufferID	create_z_buffer(const int sx,const int sy,const D3DFORMAT format);
			void		set_render_target(const int index,SurfaceID target,ZBufferID z_buffer);
			void		unset_render_target(const int index);
			SurfaceID	get_surface(TextureID texture);
			Vec3f		transform_vertex(const float fov,const float aspect,const float near_plane,const float far_plane,
										 const float cam_x,const float cam_y,const float cam_z,const Quaternion& rot,
										 const Vec3f& v);
			HRESULT		create_font(INT Height,UINT Width,UINT Weight,UINT MipLevels,BOOL Italic,DWORD CharSet,DWORD OutputPrecision,DWORD Quality,DWORD PitchAndFamily,LPCSTR pFaceName,LPD3DXFONT* ppFont);
	inline	Matrix*		get_clip_matrix()
	{
		if (_refresh_clip_matrix) refresh_clip_matrix();
		return &_clip_matrix;
	}
	inline	Matrix*		get_world_matrix(const int idx)
	{
		return &_world_matrix;
	}
	inline	LightStruct*	get_lights()
	{
		return &_lights_struct[0];
	}

protected:
			void		refresh_clip_matrix();
	inline	void		set_renderstate(D3DRENDERSTATETYPE rs,DWORD val)
	{
		if (_renderstate[rs]==val) return;
		_renderstate[rs]=val;
#ifdef USE_D3D9
		_device->SetRenderState(rs,val);
#elif defined(USE_OGL31)
		__asm int 3
		switch (rs)
		{
		case LDRS_LIGHTING: break;
		}
#else
#error "Undefined platform!"
#endif
	}
	inline	DWORD		get_renderstate(D3DRENDERSTATETYPE rs)
	{
		return _renderstate[rs];
	}
	inline	void		set_vertex_shader(VertexProgram* shader)
	{
		if (_vertex_shader==shader) return;
		_vertex_shader=NULL;
#ifdef USE_D3D9
		_device->SetVertexShader(NULL);
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
	}
	inline	void		set_fvf(DWORD fvf)
	{
		if (fvf==_fvf) return;
		_fvf=fvf;
#ifdef USE_D3D9
		_device->SetFVF(fvf);
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
	}
	inline	void		set_stream_source(UINT n,LPDIRECT3DVERTEXBUFFER9 buffer,UINT offset,UINT stride)
	{
		_device->SetStreamSource(n,buffer,offset,stride);
	}
	inline	void		set_indices(LPDIRECT3DINDEXBUFFER9 buffer)
	{
		_device->SetIndices(buffer);
	}
	inline	void		set_texture(UINT stage,TextureID tex)
	{
		if (_texture[stage]==tex) return;
		_texture[stage]=tex;
#ifdef USE_D3D9
		_device->SetTexture(stage,tex);
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
	}
	inline	void		set_texture_stage_state(UINT stage,D3DTEXTURESTAGESTATETYPE r,DWORD val)
	{
		if (_rss[stage][r]==val) return;
		_rss[stage][r]=val;
#ifdef USE_D3D9
		_device->SetTextureStageState(stage,r,val);
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
	}
	inline	void		set_sampler_state(UINT slot,D3DSAMPLERSTATETYPE state,DWORD val)
	{
		if (_sampler[slot][state]==val) return;
		_sampler[slot][state]=val;
#ifdef USE_D3D9
		_device->SetSamplerState(slot,state,val);
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
	}

	friend class Material;
	friend class VertexBuffer;
	friend class IndexBuffer;

public:
			void		set_cursor_position(long x,long y);
public:
	struct Resolution
	{
		int			_adapter_ordinal;
		std::string	_adapter;
		int			_height;
		int			_width;
		int			_bitdepth;
		int			_refresh_rate;
	};
	typedef std::vector<Resolution>	Resolutions;
	static	bool		get_available_resolutions(Resolutions* res);

protected:
#ifdef USE_D3D9
    static	bool		confirm_device_helper(D3DCAPS9* pCaps,VertexProcessingType vertexProcessingType,D3DFORMAT backBufferFormat,D3DFORMAT format);
			HRESULT		choose_initial_settings();
			HRESULT		init_d3d_environment();
			HRESULT		confirm_device(D3DCAPS9* pCaps,DWORD dwBehavior,D3DFORMAT Format);
			bool		find_best_fullscreen_mode(bool bRequireHAL,bool bRequireREF);
			bool		find_best_windowed_mode(bool bRequireHAL,bool bRequireREF);
			void		build_present_params_from_settings();
	static	int			get_bitcount(const D3DFORMAT format);
#endif

protected:
	bool							_window_mode;
	Vec2f							_resolution;
	Vec2f							_vres;
	int								_refresh_rate;
	int								_bitdepth;
	int								_adapter;
	Vec3f							_modulate;
	HWND							_window;	
	DWORD							_window_style;     // Saved window style for mode switches
	RECT							_window_bounds;    // Saved window bounds for mode switches
	RECT							_window_client;    // Saved client area size for mode switches
	DWORD							_create_flags;     // Indicate sw or hw vertex processing
	std::string						_app_name;
	DWORD							_renderstate[256];
	DWORD							_rss[MAX_TEX][256];
	DWORD							_sampler[MAX_TEX][256];
	bool							_in_scene;
	TextureID						_texture[MAX_TEX];
	VertexProgram*					_vertex_shader;
	PixelShader*					_pixel_shader;
	VertexDeclaration*				_vdecl;
	DWORD							_fvf;
	Light*							_lights[MAX_LIGHT];
	Color							_ambient_light;
	RenderEnv						_env;
	Color							_fog_color;
	Vec2f							_fog_range;
	Vec3f							_fog_center;
	bool							_refresh_clip_matrix;
	Matrix							_clip_matrix;
	Matrix							_projection_matrix;
	Matrix							_view_matrix;
	Matrix							_world_matrix;
	LightStruct						_lights_struct[MAX_LIGHT];
	PassType						_pass;
protected:
#ifdef USE_D3D9
    LPDIRECT3D9						_d3d9;          // The main D3D object
    LPDIRECT3DDEVICE9				_device;        // The D3D rendering device
    CD3DEnumeration					_d3d_enumeration;
	CD3DSettings					_d3d_settings;
	D3DPRESENT_PARAMETERS			_d3dpp;			   // Parameters for CreateDevice/Reset
	D3DCAPS9						_caps;			   // Caps for the device
	D3DSURFACE_DESC					_back_buffer_desc;   // Surface desc of the backbuffer
	LPDIRECT3DSURFACE9				_default_render_target;
	LPDIRECT3DSURFACE9				_default_depth_buffer;
	LPDIRECT3DSURFACE9				_set_depth_buffer;
#endif
#ifdef USE_OGL31
#endif
};

extern Gfx*		GFX;

#endif // !defined(AFX_GFX_H__65AF6A21_DD51_4626_9C30_4141D1E621D0__INCLUDED_)
