// Keyboard.cpp: implementation of the Keyboard class.
//
//////////////////////////////////////////////////////////////////////

#include "Keyboard.h"
#include "Utils.h"
#include "Gfx.h"

Keyboard*		KEY=NULL;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Keyboard::Keyboard()
{
	if (!KEY) KEY=this;
	_keyboard=NULL;
	_enable=true;

	memset((void*)&_last_state,0,sizeof(_last_state));
	memset((void*)&_virt_key,0,sizeof(_virt_key));
	memset((void*)&_shift_key,0,sizeof(_shift_key));

	_virt_key[DIK_1]='1'; _virt_key[DIK_2]='2'; _virt_key[DIK_3]='3'; _virt_key[DIK_4]='4'; 
	_virt_key[DIK_5]='5'; _virt_key[DIK_6]='6'; _virt_key[DIK_7]='7'; _virt_key[DIK_8]='8'; 
	_virt_key[DIK_9]='9'; _virt_key[DIK_0]='0'; _virt_key[DIK_MINUS]='-'; _virt_key[DIK_EQUALS]='='; 
	_virt_key[DIK_Q]='q'; _virt_key[DIK_W]='w'; _virt_key[DIK_E]='e'; _virt_key[DIK_R]='r'; 
	_virt_key[DIK_T]='t'; _virt_key[DIK_Y]='y'; _virt_key[DIK_U]='u'; _virt_key[DIK_I]='i'; 
	_virt_key[DIK_O]='o'; _virt_key[DIK_P]='p'; _virt_key[DIK_LBRACKET]='['; _virt_key[DIK_RBRACKET]=']'; 
	_virt_key[DIK_RETURN]=13; _virt_key[DIK_A]='a'; _virt_key[DIK_S]='s'; _virt_key[DIK_D]='d'; 
	_virt_key[DIK_F]='f'; _virt_key[DIK_G]='g'; _virt_key[DIK_H]='h'; _virt_key[DIK_J]='j'; 
	_virt_key[DIK_K]='k'; _virt_key[DIK_L]='l'; _virt_key[DIK_SEMICOLON]=';'; _virt_key[DIK_APOSTROPHE]='\''; 
	_virt_key[DIK_GRAVE]='`'; _virt_key[DIK_BACKSLASH]='\\'; _virt_key[DIK_Z]='z'; _virt_key[DIK_X]='x'; 
	_virt_key[DIK_C]='c'; _virt_key[DIK_V]='v'; _virt_key[DIK_B]='b'; _virt_key[DIK_N]='n'; 
	_virt_key[DIK_M]='m'; _virt_key[DIK_COMMA]=','; _virt_key[DIK_PERIOD]='.'; _virt_key[DIK_SLASH]='-'; 
	_virt_key[DIK_SPACE]=' ';
	_virt_key[DIK_ADD]='+'; _virt_key[DIK_SUBTRACT]='-';
	_virt_key[DIK_MINUS]='-';

	_shift_key[DIK_1]='!'; _shift_key[DIK_2]='"'; _shift_key[DIK_3]='#'; _shift_key[DIK_4]='$'; 
	_shift_key[DIK_5]='%'; _shift_key[DIK_6]='&'; _shift_key[DIK_7]='/'; _shift_key[DIK_8]='('; 
	_shift_key[DIK_9]=')'; _shift_key[DIK_0]='='; _shift_key[DIK_MINUS]='_'; _shift_key[DIK_EQUALS]='='; 
	_shift_key[DIK_Q]='Q'; _shift_key[DIK_W]='W'; _shift_key[DIK_E]='E'; _shift_key[DIK_R]='R'; 
	_shift_key[DIK_T]='T'; _shift_key[DIK_Y]='Y'; _shift_key[DIK_U]='U'; _shift_key[DIK_I]='I'; 
	_shift_key[DIK_O]='O'; _shift_key[DIK_P]='P'; _shift_key[DIK_LBRACKET]='['; _shift_key[DIK_RBRACKET]=']'; 
	_shift_key[DIK_RETURN]=13; _shift_key[DIK_A]='A'; _shift_key[DIK_S]='S'; _shift_key[DIK_D]='D'; 
	_shift_key[DIK_F]='F'; _shift_key[DIK_G]='G'; _shift_key[DIK_H]='H'; _shift_key[DIK_J]='J'; 
	_shift_key[DIK_K]='K'; _shift_key[DIK_L]='L'; _shift_key[DIK_SEMICOLON]=';'; _shift_key[DIK_APOSTROPHE]='\''; 
	_shift_key[DIK_GRAVE]='`'; _shift_key[DIK_BACKSLASH]='\\'; _shift_key[DIK_Z]='Z'; _shift_key[DIK_X]='X'; 
	_shift_key[DIK_C]='C'; _shift_key[DIK_V]='V'; _shift_key[DIK_B]='B'; _shift_key[DIK_N]='N'; 
	_shift_key[DIK_M]='M'; _shift_key[DIK_COMMA]=';'; _shift_key[DIK_PERIOD]=':'; _shift_key[DIK_SLASH]='-'; 
	_shift_key[DIK_SPACE]=' ';
	_shift_key[DIK_EQUALS]='+';
}

Keyboard::~Keyboard()
{
	if (KEY==this) KEY=NULL;
	if (_keyboard) _keyboard->Release();
	if (LPDI) 
	{
		if (LPDI->Release()==0) LPDI=NULL;
	}
}

bool Keyboard::init(HINSTANCE instance)
{
	if (!LPDI)
	{
		HRESULT hRet;

		hRet=DirectInput8Create(instance,DIRECTINPUT_VERSION,IID_IDirectInput8,(void**)&LPDI,NULL);
		if (hRet!=DI_OK)
		{
			ERROR_MSG("Couldn't initialize DirectInput!");
			return false;
		}
	}

	HRESULT hRet=LPDI->CreateDevice(GUID_SysKeyboard,&_keyboard,NULL);
	if (hRet!=DI_OK)
	{
		ERROR_MSG("Can't initialize keyboard!");
		if (LPDI) LPDI->Release(); LPDI=NULL;	
		return false;
	}
	
	hRet=_keyboard->SetDataFormat(&c_dfDIKeyboard);
	if (hRet!=DI_OK)
	{
		_keyboard->Release();
		if (LPDI) LPDI->Release(); LPDI=NULL;	
		ERROR_MSG("Couldn't set the data format for the keyboard!");
		return false;
	}

	hRet=_keyboard->SetCooperativeLevel(GFX->get_window(),DISCL_NONEXCLUSIVE | DISCL_BACKGROUND);

	if (hRet!=DI_OK)
	{
		ERROR_MSG("Couldn't set cooperative level for the keyboard!");
	}

	hRet=_keyboard->Acquire();
	if (hRet!=DI_OK)
	{
		ERROR_MSG("Couldn't acquire the keyboard!");
	}

	return true;
}

void Keyboard::poll()
{
	unsigned char buffer[256];

	if (_enable)
	{
		HRESULT hRet=_keyboard->Poll();

		hRet=_keyboard->GetDeviceState(sizeof(buffer),(void*)buffer); 
		if (hRet!=DI_OK)
		{
			switch (hRet)
			{
				case DIERR_INPUTLOST:
				case DIERR_NOTACQUIRED:
					hRet=_keyboard->Acquire();
					if (hRet!=DI_OK)
					{
						ERROR_MSG("Couldn't poll keyboard: Couldn't acquire the keyboard!");
						return;
					}
					else
					{
						hRet=_keyboard->GetDeviceState(sizeof(buffer),(void*)&buffer); 
						if (hRet!=DI_OK)
						{
							ERROR_MSG("Couldn't poll keyboard: Input lost!");
							return;
						}
					}
					break;
				case DIERR_INVALIDPARAM:
					ERROR_MSG("Couldn't poll keyboard: Invalid parameter!");
					return;
					break;
				case DIERR_NOTINITIALIZED:
					ERROR_MSG("Couldn't poll keyboard: Not initialized!");
					return;
					break;
				case E_PENDING:
					ERROR_MSG("Couldn't poll keyboard: Data is not yet available!");
					return;
					break;
			}
		}

		memcpy(_last_state,buffer,256);
	}
	else
	{
		memset(_last_state,0,256);
	}
}

int	 Keyboard::any_pressed()
{
	for (int i=0; i<256; i++)
	{
		if (_last_state[i] & 0x80) return 1;
	}
	return 0;
}

char Keyboard::get_pressed()
{
	for (int i=0; i<256; i++)
	{
		if ((i==DIK_LSHIFT) ||
			(i==DIK_RSHIFT) ||
			(i==DIK_LCONTROL) ||
			(i==DIK_RCONTROL)) continue;
		if (_last_state[i] & 128) 
		{
			if ((left_shift_pressed()) || (right_shift_pressed()))
				return _shift_key[i];
			else
				return _virt_key[i];
		}
	}

	return 0;
}

bool Keyboard::left_shift_pressed() const
{ 
	return (_last_state[DIK_LSHIFT] & 0x80)!=0;
}

bool Keyboard::right_shift_pressed()  const
{ 
	return (_last_state[DIK_RSHIFT] & 0x80)!=0;
}

