
#if defined(_WIN32) || defined(_WIN64)
#include <windows.h>
#include <CommCtrl.h>
#include "resource.h"
#endif

#include "Gfx.h"
#include "XML.h"

class Options
{
public:
	static	int					_adapter;
	static	int					_res_x;
	static	int					_res_y;
	static	int					_bitdepth;
	static	int					_refresh_rate;
	static	bool				_window_mode;
	static	Gfx::Resolutions	_available_res;
	static	bool				_show_launcher;
	static	bool				_sound_enable;
	static	float				_volume_music;
	static	float				_volume_fx;
};

int					Options::_adapter=0;
int					Options::_res_x=1280;
int					Options::_res_y=720;
int					Options::_bitdepth=32;
int					Options::_refresh_rate=60;
bool				Options::_window_mode=true;
Gfx::Resolutions	Options::_available_res;
bool				Options::_show_launcher=true;
bool				Options::_sound_enable=true;
float				Options::_volume_music=1.0f;
float				Options::_volume_fx=1.0f;

#if defined(_WIN32) || defined(_WIN64)

int	ComboBox_SelectExactString(HWND hwndCtl,int indexStart,LPCTSTR lpszFind)
{
	int f=ComboBox_FindStringExact(hwndCtl,indexStart,lpszFind);
	if (f!=LB_ERR)
	{
		ComboBox_SetCurSel(hwndCtl,f);
	}
	return f;
}

void ComboBox_AddUniqueString(HWND hwnd,const char* str)
{
	int n=ComboBox_FindStringExact(hwnd,0,str);
	if (n==LB_ERR)
	{
		ComboBox_AddString(hwnd,str);
	}
}

void fill_adapters_combo(HWND dlg_hwnd)
{
	HWND hwnd=GetDlgItem(dlg_hwnd,IDC_ADAPTERS);

	ComboBox_ResetContent(hwnd);
	for (size_t i=0; i<Options::_available_res.size(); i++)
	{
		ComboBox_AddUniqueString(hwnd,Options::_available_res[i]._adapter.c_str());
	}
}

void fill_resolution_combo(HWND dlg_hwnd,int adapter_ordinal)
{
	HWND hwnd=GetDlgItem(dlg_hwnd,IDC_RESOLUTIONS);

	ComboBox_ResetContent(hwnd);
	for (size_t i=0; i<Options::_available_res.size(); i++)
	{
		if (Options::_available_res[i]._adapter_ordinal==adapter_ordinal)
		{
			char buffer[512];
			sprintf_s((char*)&buffer,512,"%ix%i",Options::_available_res[i]._width,Options::_available_res[i]._height);
			ComboBox_AddUniqueString(hwnd,buffer);
		}
	}
}

void fill_bitdepth_combo(HWND dlg_hwnd,int adapter_ordinal,const int res_x,const int res_y)
{
	HWND hwnd=GetDlgItem(dlg_hwnd,IDC_BITDEPTHS);

	ComboBox_ResetContent(hwnd);
	for (size_t i=0; i<Options::_available_res.size(); i++)
	{
		if ((Options::_available_res[i]._adapter_ordinal==adapter_ordinal) &&
			(Options::_available_res[i]._width==res_x) &&
			(Options::_available_res[i]._height==res_y))
		{
			char buffer[512];
			sprintf_s((char*)&buffer,512,"%i",Options::_available_res[i]._bitdepth);
			ComboBox_AddUniqueString(hwnd,buffer);
		}
	}
}

void fill_refreshrate_combo(HWND dlg_hwnd,int adapter_ordinal,const int res_x,const int res_y,const int bitdepth)
{
	HWND hwnd=GetDlgItem(dlg_hwnd,IDC_REFRESHRATES);

	ComboBox_ResetContent(hwnd);
	for (size_t i=0; i<Options::_available_res.size(); i++)
	{
		if ((Options::_available_res[i]._adapter_ordinal==adapter_ordinal) &&
			(Options::_available_res[i]._width==res_x) &&
			(Options::_available_res[i]._height==res_y) &&
			(Options::_available_res[i]._bitdepth==bitdepth))
		{
			char buffer[512];
			sprintf_s((char*)&buffer,512,"%i Hz",Options::_available_res[i]._refresh_rate);
			ComboBox_AddUniqueString(hwnd,buffer);
		}
	}
}

void select_refreshrate(HWND dlg_hwnd,int ordinal,const int res_x,const int res_y,const int bitdepth,const int refresh_rate)
{
	char buffer[512];
	sprintf_s((char*)&buffer,512,"%i Hz",refresh_rate);
	std::string tmp=buffer;

	HWND hwnd=GetDlgItem(dlg_hwnd,IDC_REFRESHRATES);

	int	sel=ComboBox_FindStringExact(hwnd,0,buffer);
	if (sel==LB_ERR)
	{
		sel=ComboBox_GetCount(hwnd)-1;
	}

	ComboBox_SetCurSel(hwnd,sel);

	ComboBox_GetLBText(hwnd,sel,buffer);
	buffer[strlen(buffer)-3]=0;
	Options::_refresh_rate=atol(buffer);
}

void select_bitdepth(HWND dlg_hwnd,int ordinal,const int res_x,const int res_y,const int bitdepth)
{
	char buffer[512];
	sprintf_s((char*)&buffer,512,"%i",bitdepth);
	std::string tmp=buffer;

	HWND hwnd=GetDlgItem(dlg_hwnd,IDC_BITDEPTHS);

	int	sel=ComboBox_FindStringExact(hwnd,0,buffer);
	if (sel==LB_ERR)
	{
		sel=ComboBox_GetCount(hwnd)-1;
	}

	ComboBox_SetCurSel(hwnd,sel);

	ComboBox_GetLBText(hwnd,sel,buffer);

	Options::_bitdepth=atol(buffer);

	fill_refreshrate_combo(dlg_hwnd,ordinal,Options::_res_x,Options::_res_y,Options::_bitdepth);
	select_refreshrate(dlg_hwnd,ordinal,Options::_res_x,Options::_res_y,Options::_bitdepth,Options::_refresh_rate);
}

void select_resolution(HWND dlg_hwnd,int ordinal,const int res_x,const int res_y)
{
	char buffer[512];
	sprintf_s((char*)&buffer,512,"%ix%i",res_x,res_y);
	std::string tmp=buffer;

	HWND hwnd=GetDlgItem(dlg_hwnd,IDC_RESOLUTIONS);
	
	int	sel=ComboBox_FindStringExact(hwnd,0,buffer);
	if (sel==LB_ERR)
	{
		sel=ComboBox_GetCount(hwnd)-1;
	}

	ComboBox_SetCurSel(hwnd,sel);
	
	ComboBox_GetLBText(hwnd,sel,buffer);

				tmp=buffer;
	size_t		idx=tmp.find('x');

	Options::_res_x=atol(tmp.substr(0,idx).c_str());
	Options::_res_y=atol(tmp.substr(idx+1,std::string::npos).c_str());

	fill_bitdepth_combo(dlg_hwnd,ordinal,Options::_res_x,Options::_res_y);
	select_bitdepth(dlg_hwnd,ordinal,Options::_res_x,Options::_res_y,Options::_bitdepth);
}

void select_adapter(HWND dlg_hwnd,int ordinal)
{
	HWND hwnd=GetDlgItem(dlg_hwnd,IDC_ADAPTERS);

	std::string adapter_name="";
	for (size_t i=0; i<Options::_available_res.size(); i++)
	{
		if (Options::_available_res[i]._adapter_ordinal==ordinal)
		{
			adapter_name=Options::_available_res[i]._adapter;
			break;
		}
	}

	int new_ordinal=ordinal;
	if (adapter_name=="")
	{
		char buffer[512];

		ComboBox_SetCurSel(hwnd,0);
		ComboBox_GetLBText(hwnd,0,buffer);
		adapter_name=buffer;

		for (size_t i=0; i<Options::_available_res.size(); i++)
		{
			if (Options::_available_res[i]._adapter==adapter_name)
			{
				new_ordinal=Options::_available_res[i]._adapter_ordinal;
				break;
			}
		}
	}
	else
	{
		ComboBox_SelectExactString(hwnd,0,adapter_name.c_str());
	}

	Options::_adapter=new_ordinal;

	fill_resolution_combo(dlg_hwnd,new_ordinal);
	select_resolution(dlg_hwnd,new_ordinal,Options::_res_x,Options::_res_y);
}

void select_windowed(HWND dlg_hwnd,const bool window_mode)
{
	if (window_mode) Button_SetCheck(GetDlgItem(dlg_hwnd,IDC_WINDOWMODE),TRUE);
	else Button_SetCheck(GetDlgItem(dlg_hwnd,IDC_WINDOWMODE),FALSE);

	if (window_mode)
	{
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_REFRESHRATES),FALSE);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_BITDEPTHS),FALSE);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_REFRESHRATES_STATIC),FALSE);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_BITDEPTHS_STATIC),FALSE);
	}
	else
	{
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_REFRESHRATES),TRUE);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_BITDEPTHS),TRUE);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_REFRESHRATES_STATIC),TRUE);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_BITDEPTHS_STATIC),TRUE);
	}
}

void update_sound_controls(HWND dlg_hwnd)
{
	if (Options::_sound_enable) Button_SetCheck(GetDlgItem(dlg_hwnd,IDC_SOUNDENABLE),TRUE);
	else Button_SetCheck(GetDlgItem(dlg_hwnd,IDC_SOUNDENABLE),FALSE);

	if (Options::_sound_enable)
	{
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_VOLUME_MUSIC),SW_SHOW);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_VOLUME_FX),SW_SHOW);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_VOLUME_MUSIC_STATIC),SW_SHOW);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_VOLUME_FX_STATIC),SW_SHOW);
	}
	else
	{
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_VOLUME_MUSIC),SW_HIDE);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_VOLUME_FX),SW_HIDE);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_VOLUME_MUSIC_STATIC),SW_HIDE);
		EnableWindow(GetDlgItem(dlg_hwnd,IDC_VOLUME_FX_STATIC),SW_HIDE);
	}

	SendMessage(GetDlgItem(dlg_hwnd,IDC_VOLUME_MUSIC),TBM_SETPOS,TRUE,(int)(Options::_volume_music*65535.0f));
	SendMessage(GetDlgItem(dlg_hwnd,IDC_VOLUME_FX),TBM_SETPOS,TRUE,(int)(Options::_volume_fx*65535.0f));
}

INT_PTR CALLBACK options_callback(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
	switch (uMsg)
	{
		case WM_INITDIALOG:
			{
				// Initialize fields
				Options::_available_res.clear();
				Gfx::get_available_resolutions(&Options::_available_res);

				fill_adapters_combo(hwnd);
				select_adapter(hwnd,Options::_adapter);

				select_windowed(hwnd,Options::_window_mode);

				Button_SetCheck(GetDlgItem(hwnd,IDC_SHOWLAUNCHER),(Options::_show_launcher)?(TRUE):(FALSE));

				SendMessage(GetDlgItem(hwnd,IDC_VOLUME_MUSIC),TBM_SETRANGEMIN,MAKEWPARAM(TRUE,0),0);
				SendMessage(GetDlgItem(hwnd,IDC_VOLUME_MUSIC),TBM_SETRANGEMAX,MAKEWPARAM(TRUE,0),65535);
				SendMessage(GetDlgItem(hwnd,IDC_VOLUME_FX),TBM_SETRANGEMIN,MAKEWPARAM(TRUE,0),0);
				SendMessage(GetDlgItem(hwnd,IDC_VOLUME_FX),TBM_SETRANGEMAX,MAKEWPARAM(TRUE,0),65535);

				update_sound_controls(hwnd);

				return TRUE;
			}
			break;
		case WM_HSCROLL:
			{
				int			val=SendMessage((HWND)lParam,TBM_GETPOS,0,0);
				float		f=((float)val/65535.0f);

				if ((HWND)lParam==GetDlgItem(hwnd,IDC_VOLUME_MUSIC))
				{
					Options::_volume_music=f;
				}
				else if ((HWND)lParam==GetDlgItem(hwnd,IDC_VOLUME_FX))
				{
					Options::_volume_fx=f;
				}
			}
			break;
		case WM_COMMAND:
			if (wParam==IDOK)
			{
				EndDialog(hwnd,IDOK);
				return TRUE;
			}
			else if (wParam==IDCANCEL)
			{
				EndDialog(hwnd,IDCANCEL);
				return TRUE;
			}
			else
			{
				switch (HIWORD(wParam))
				{
					case BN_CLICKED:
						{ 
							int st=Button_GetCheck((HWND)lParam);
							if (LOWORD(wParam)==IDC_WINDOWMODE)
							{
								if (st==BST_CHECKED) Options::_window_mode=true;
								else Options::_window_mode=false;

								select_windowed(hwnd,Options::_window_mode);
							}
							if (LOWORD(wParam)==IDC_SHOWLAUNCHER)
							{
								if (st==BST_CHECKED) Options::_show_launcher=true;
								else Options::_show_launcher=false;

								Button_SetCheck(GetDlgItem(hwnd,IDC_SHOWLAUNCHER),(Options::_show_launcher)?(TRUE):(FALSE));
							}
							if (LOWORD(wParam)==IDC_SOUNDENABLE)
							{
								if (st==BST_CHECKED) Options::_sound_enable=true;
								else Options::_sound_enable=false;

								update_sound_controls(hwnd);
							}
						}
						break;
					case CBN_SELCHANGE:
						{
							if ((HWND)lParam==GetDlgItem(hwnd,IDC_ADAPTERS))
							{
								// Combo changed
								int index=ComboBox_GetCurSel((HWND)lParam);
								if (index!=LB_ERR)
								{
									char buffer[512];
									ComboBox_GetLBText((HWND)lParam,index,buffer);

									std::string adapter_name=buffer;
									for (size_t i=0; i<Options::_available_res.size(); i++)
									{
										if (Options::_available_res[i]._adapter==adapter_name)
										{
											Options::_adapter=Options::_available_res[i]._adapter_ordinal;
											break;
										}
									}

									fill_resolution_combo(hwnd,Options::_adapter);
									select_resolution(hwnd,Options::_adapter,Options::_res_x,Options::_res_y);
								}
							}
							else if ((HWND)lParam==GetDlgItem(hwnd,IDC_RESOLUTIONS))
							{
								// Combo changed
								int index=ComboBox_GetCurSel((HWND)lParam);
								if (index!=LB_ERR)
								{
									char buffer[512];
									ComboBox_GetLBText((HWND)lParam,index,buffer);

									std::string	tmp=buffer;
									size_t		idx=tmp.find('x');

									Options::_res_x=atol(tmp.substr(0,idx).c_str());
									Options::_res_y=atol(tmp.substr(idx+1,std::string::npos).c_str());

									fill_bitdepth_combo(hwnd,Options::_adapter,Options::_res_x,Options::_res_y);
									select_bitdepth(hwnd,Options::_adapter,Options::_res_x,Options::_res_y,Options::_bitdepth);								
								}
							}
							else if ((HWND)lParam==GetDlgItem(hwnd,IDC_BITDEPTHS))
							{
								// Combo changed
								int index=ComboBox_GetCurSel((HWND)lParam);
								if (index!=LB_ERR)
								{
									char buffer[512];
									ComboBox_GetLBText((HWND)lParam,index,buffer);

									std::string	tmp=buffer;
									size_t		idx=tmp.find('x');

									Options::_bitdepth=atol(buffer);

									fill_refreshrate_combo(hwnd,Options::_adapter,Options::_res_x,Options::_res_y,Options::_bitdepth);
									select_refreshrate(hwnd,Options::_adapter,Options::_res_x,Options::_res_y,Options::_bitdepth,Options::_refresh_rate);
								}
							}
							else if ((HWND)lParam==GetDlgItem(hwnd,IDC_REFRESHRATES))
							{
								// Combo changed
								int index=ComboBox_GetCurSel((HWND)lParam);
								if (index!=LB_ERR)
								{
									char buffer[512];
									ComboBox_GetLBText((HWND)lParam,index,buffer);

									std::string	tmp=buffer;
									size_t		idx=tmp.find('x');

									buffer[strlen(buffer)-3]=0;
									Options::_refresh_rate=atol(buffer);
								}
							}
						}
						break;
				}
			}
			break;
	}
	return FALSE;
}

bool launch_launcher()
{
	int ret=(int)DialogBox(GetModuleHandle(NULL),MAKEINTRESOURCE(IDD_RUN_OPTIONS),NULL,options_callback);

	if (ret==IDCANCEL) return false;

	return true;
}

#endif

int options_get_adapter()
{
	return Options::_adapter;
}

int	 options_get_res_x()
{
	return Options::_res_x;
}

int	options_get_res_y()
{
	return Options::_res_y;
}

bool options_get_window_mode()
{
	return Options::_window_mode;
}

bool options_show_launcher()
{
	return Options::_show_launcher;
}

int	options_get_bitdepth()
{
	return Options::_bitdepth;
}

int	options_get_refresh_rate()
{
	return Options::_refresh_rate;
}

bool options_get_sound()
{
	return Options::_sound_enable;
}

float options_get_music_volume()
{
	return Options::_volume_music;
}

float options_get_fx_volume()
{
	return Options::_volume_fx;
}

void options_load()
{
	XMLNode root_node=XMLNode::openFileHelper("options.xml");
	if (root_node.isEmpty()) return;

	XMLNode options_node=root_node.getChildNode("options",0);
	if (options_node.isEmpty()) return;

	Options::_show_launcher=options_node.getBoolAttribute("show_launcher",NULL,true);

	XMLNode gfx=options_node.getChildNode("gfx",0);
	if (!gfx.isEmpty())
	{
		Options::_adapter=gfx.getIntAttribute("adapter",NULL,0);
		Options::_res_x=gfx.getIntAttribute("res_x",NULL,1280);
		Options::_res_y=gfx.getIntAttribute("res_y",NULL,720);
		Options::_bitdepth=gfx.getIntAttribute("bitdepth",NULL,32);
		Options::_refresh_rate=gfx.getIntAttribute("refresh_rate",NULL,60);
		Options::_window_mode=gfx.getBoolAttribute("window_mode",NULL,true);
	}

	XMLNode snd=options_node.getChildNode("snd",0);
	if (!snd.isEmpty())
	{
		Options::_sound_enable=snd.getBoolAttribute("enable",NULL,true);
		Options::_volume_music=snd.getFloatAttribute("volume_music",NULL,1.0f);
		Options::_volume_fx=snd.getFloatAttribute("volume_fx",NULL,1.0f);
	}
}

void options_save()
{
	scXMLWriterNode* root=new scXMLWriterNode(NULL,"UTF-8");

	scXMLWriterNode* options=new scXMLWriterNode(root,"options");
	options->add_property("show_launcher",Options::_show_launcher);
	
	scXMLWriterNode* gfx=new scXMLWriterNode(options,"gfx");
	gfx->add_property("adapter",Options::_adapter);
	gfx->add_property("res_x",Options::_res_x);
	gfx->add_property("res_y",Options::_res_y);
	gfx->add_property("bitdepth",Options::_bitdepth);
	gfx->add_property("refresh_rate",Options::_refresh_rate);
	gfx->add_property("window_mode",Options::_window_mode);

	scXMLWriterNode* snd=new scXMLWriterNode(options,"snd");
	snd->add_property("enable",Options::_sound_enable);
	snd->add_property("volume_music",Options::_volume_music);
	snd->add_property("volume_fx",Options::_volume_fx);

	root->output("options.xml");

	delete root;
}

