#include "Light.h"
#include "Gfx.h"
#include "Utils.h"

Light::Light(const char* name) : Object3d("light3d",name)
{
	_type=LIGHT_DIR;
	_color.set(1.0f,1.0f,1.0f,1.0f);
	_range.set(0.0f,FLT_MAX);
	_cone.set(30.0f,45.0f);
#ifdef USE_D3D9
	_light.Falloff=1.0f;
	_light.Attenuation0=1.0f;
   	_light.Attenuation1=0.0f;
   	_light.Attenuation2=0.0f;
	_light.Theta=30.0f;
	_light.Phi=45.0f;
#endif
}

Light::~Light(void)
{
}

void Light::update()
{
	Vec3f dir;
	_ori.rotate_point(dir,Vec3f(0.0f,0.0f,1.0f));
#if defined(USE_D3D9 )&& !defined(USE_SHADERS)
	_light.Position=D3DXVECTOR3(_pos.x,_pos.y,_pos.z);
	_light.Direction=D3DXVECTOR3(dir.x,dir.y,dir.z);
	_light.Type=(D3DLIGHTTYPE)_type;
	_light.Ambient.r=_light.Ambient.g=_light.Ambient.b=_light.Ambient.a=0.0f; 
	_light.Diffuse.r=_color.r; _light.Diffuse.g=_color.g; _light.Diffuse.b=_color.b; _light.Diffuse.a=_color.a; 
	_light.Specular.r=_color.r; _light.Specular.g=_color.g; _light.Specular.b=_color.b; _light.Specular.a=_color.a; 
	_light.Range=sqrtf(_range.y);
	_light.Theta=deg_to_rad(_cone.x);
	_light.Phi=deg_to_rad(_cone.y);
	if (_type==LIGHT_SPOT)
	{
		_light.Attenuation0=0.0f;
		_light.Attenuation1=1.0f/_range.y;
	}
	else
	{
		_light.Attenuation0=1.0f;
		_light.Attenuation1=0.0f;
	}
#endif
}

void Light::draw()
{
	if (_link)
	{
		_pos=_link->_pos;
		_ori=_link->_ori;
	}

	GFX->set_world_transformation(0,_pos.x,_pos.y,_pos.z,_ori,_scale.x,_scale.y,_scale.z);
}
