#include "Material.h"
#include "TextureManager.h"
#include "Logger.h"
#include "Utils.h"

Material::Material(void)
{
	_cull_mode=D3DCULL_CW;
	_shader="basic";
	for (int i=0; i<MAX_TEX; i++) 
	{
		_textures[i]=-1;
		_textures_filter[i]=D3DTEXF_LINEAR;
	}
	_name="";
	_z_write=true;
	_point_sprite=false;
	_color_vertex=false;
	_color_vertex_e=false;
	_alpha_enable=false;
	_alpha_test=false;
	_src_blend=D3DBLEND_SRCALPHA;
	_dest_blend=D3DBLEND_INVSRCALPHA;
	set_material_color(1.0f,1.0f,1.0f,1.0f);
	_passes=0xFFFFFFFF;
}

Material::~Material(void)
{
}

void Material::set_shader(const char* shader)
{
	if (shader) _shader=shader;
	else _shader="";
}

void Material::set(const unsigned long fvf)
{
#ifdef USE_SHADERS
	_old_z_write=GFX->get_renderstate(D3DRS_ZWRITEENABLE);

	// Build render env
	RenderEnv new_env(GFX->get_render_env());
	new_env._props._psize=(fvf&D3DFVF_PSIZE)?(1):(0);
	new_env._props._normal=(fvf&D3DFVF_NORMAL)?(1):(0);
	new_env._props._color0=(fvf&D3DFVF_DIFFUSE)?(1):(0);
	new_env._props._color1=(fvf&D3DFVF_SPECULAR)?(1):(0);
	new_env._props._tex_coord_count=(fvf&D3DFVF_TEXCOUNT_MASK)>>D3DFVF_TEXCOUNT_SHIFT;

	if (_alpha_test) new_env._props._alpha_test_enabled=1;
	else new_env._props._alpha_test_enabled=0;

	new_env._props._map[0]=0;
	for (int i=0; i<MAX_TEX; i++) if (_textures[i]!=-1) new_env._props._map[0]+=1<<i;

	GFX->set_renderstate(D3DRS_ZWRITEENABLE,(_z_write)?(1):(0));
	if (_alpha_enable)
	{
		GFX->set_renderstate(D3DRS_ALPHABLENDENABLE,1);
		GFX->set_renderstate(D3DRS_SRCBLEND,_src_blend);
		GFX->set_renderstate(D3DRS_DESTBLEND,_dest_blend);
	}
	else
	{
		GFX->set_renderstate(D3DRS_ALPHABLENDENABLE,0);
	}

	if (_alpha_test)
	{
		GFX->set_renderstate(D3DRS_ALPHATESTENABLE,TRUE);
		GFX->set_renderstate(D3DRS_ALPHAFUNC,D3DCMP_GREATER);
		GFX->set_renderstate(D3DRS_ALPHAREF,_alpha_ref);
	}
	else
	{
		GFX->set_renderstate(D3DRS_ALPHATESTENABLE,FALSE);
	}

	if (_point_sprite)
	{
		GFX->set_renderstate(D3DRS_POINTSPRITEENABLE,TRUE);
		GFX->set_renderstate(D3DRS_POINTSCALEENABLE,TRUE);
		GFX->set_renderstate(D3DRS_POINTSIZE_MIN,FtoDW(0.00f));
		GFX->set_renderstate(D3DRS_POINTSCALE_A,FtoDW(0.0f));
		GFX->set_renderstate(D3DRS_POINTSCALE_B,FtoDW(0.0f));
		GFX->set_renderstate(D3DRS_POINTSCALE_C,FtoDW(1.0f));
	}
	else
	{
		GFX->set_renderstate(D3DRS_POINTSPRITEENABLE,FALSE);
	}
	GFX->set_renderstate(D3DRS_CULLMODE,_cull_mode);

	for (int i=0; i<MAX_TEX; i++)
	{
		if (_textures[i]!=-1)
		{
			TextureID tex=TEXTURES->get_texture(_textures[i]);
			GFX->set_texture(i,tex);

			GFX->set_sampler_state(i,D3DSAMP_MINFILTER,_textures_filter[i]);
			GFX->set_sampler_state(i,D3DSAMP_MAGFILTER,_textures_filter[i]);
			GFX->set_sampler_state(i,D3DSAMP_MIPFILTER,D3DTEXF_NONE);

			if (TEXTURES->get_wrap(_textures[i])) 
			{
				GFX->set_sampler_state(i,D3DSAMP_ADDRESSU,D3DTADDRESS_WRAP);
				GFX->set_sampler_state(i,D3DSAMP_ADDRESSV,D3DTADDRESS_WRAP);
			}
			else
			{
				GFX->set_sampler_state(i,D3DSAMP_ADDRESSU,D3DTADDRESS_CLAMP);
				GFX->set_sampler_state(i,D3DSAMP_ADDRESSV,D3DTADDRESS_CLAMP);
			}
		}
		else
		{
			GFX->set_texture(i,NULL);
		}
	}

	set_shader(_shader,&new_env,fvf);
#else
	GFX->set_fvf(fvf);

	// Get old state
	_old_lighting=GFX->get_renderstate(D3DRS_LIGHTING);
	_old_alpha_test=GFX->get_renderstate(D3DRS_ALPHATESTENABLE);
	_old_alpha_ref=GFX->get_renderstate(D3DRS_ALPHAREF);

	if (fvf&D3DFVF_NORMAL) GFX->set_renderstate(D3DRS_LIGHTING,1);
	else GFX->set_renderstate(D3DRS_LIGHTING,0);

	GFX->_device->SetMaterial(&_material);
	GFX->set_renderstate(D3DRS_CULLMODE,_cull_mode);
	if (_color_vertex)
	{
		GFX->set_renderstate(D3DRS_AMBIENTMATERIALSOURCE,D3DMCS_COLOR1);
		GFX->set_renderstate(D3DRS_DIFFUSEMATERIALSOURCE,D3DMCS_COLOR1);
		GFX->set_renderstate(D3DRS_SPECULARMATERIALSOURCE,D3DMCS_MATERIAL);
		if (_color_vertex_e) GFX->set_renderstate(D3DRS_EMISSIVEMATERIALSOURCE,D3DMCS_COLOR2);
		else 
		{
			if (_fx==FX_SOLID_EMISSIVE) GFX->set_renderstate(D3DRS_EMISSIVEMATERIALSOURCE,D3DMCS_COLOR1);
			else GFX->set_renderstate(D3DRS_EMISSIVEMATERIALSOURCE,D3DMCS_MATERIAL);
		}
		GFX->set_renderstate(D3DRS_COLORVERTEX,TRUE);
	}
	else
	{
		GFX->set_renderstate(D3DRS_AMBIENTMATERIALSOURCE,D3DMCS_MATERIAL);
		GFX->set_renderstate(D3DRS_DIFFUSEMATERIALSOURCE,D3DMCS_MATERIAL);
		GFX->set_renderstate(D3DRS_SPECULARMATERIALSOURCE,D3DMCS_MATERIAL);
		GFX->set_renderstate(D3DRS_EMISSIVEMATERIALSOURCE,D3DMCS_MATERIAL);
		GFX->set_renderstate(D3DRS_COLORVERTEX,FALSE);
	}
	if (_alpha_enable)
	{
		GFX->set_renderstate(D3DRS_ALPHABLENDENABLE,1);
		GFX->set_renderstate(D3DRS_SRCBLEND,_src_blend);
		GFX->set_renderstate(D3DRS_DESTBLEND,_dest_blend);
	}
	else
	{
		GFX->set_renderstate(D3DRS_ALPHABLENDENABLE,0);
	}

	if (_point_sprite)
	{
		GFX->set_renderstate(D3DRS_POINTSPRITEENABLE,TRUE);
		GFX->set_renderstate(D3DRS_POINTSCALEENABLE,TRUE);
		GFX->set_renderstate(D3DRS_POINTSIZE_MIN,FtoDW(0.00f));
		GFX->set_renderstate(D3DRS_POINTSCALE_A,FtoDW(0.0f));
		GFX->set_renderstate(D3DRS_POINTSCALE_B,FtoDW(0.0f));
		GFX->set_renderstate(D3DRS_POINTSCALE_C,FtoDW(1.0f));
	}
	else
	{
		GFX->set_renderstate(D3DRS_POINTSPRITEENABLE,FALSE);
	}

	if (_alpha_test)
	{
		GFX->set_renderstate(D3DRS_ALPHATESTENABLE,TRUE);
		GFX->set_renderstate(D3DRS_ALPHAFUNC,D3DCMP_GREATER);
		GFX->set_renderstate(D3DRS_ALPHAREF,_alpha_ref);
	}
	else
	{
		GFX->set_renderstate(D3DRS_ALPHATESTENABLE,FALSE);
	}

	if (_fx==FX_SOLID)
	{
		for (int i=0; i<MAX_TEX; i++)
		{
			GFX->set_texture(i,NULL);
		}
	}
	if (_fx==FX_TEXTURED)
	{
		long stage=0;
		for (int i=0; i<MAX_TEX; i++)
		{
			if (_textures[i]!=-1)
			{
				TextureID tex=TEXTURES->get_texture(_textures[i]);
				GFX->set_texture(stage,tex);

				GFX->set_sampler_state(stage,D3DSAMP_MINFILTER,_textures_filter[i]);
				GFX->set_sampler_state(stage,D3DSAMP_MAGFILTER,_textures_filter[i]);
				GFX->set_sampler_state(stage,D3DSAMP_MIPFILTER,D3DTEXF_LINEAR);

				if (TEXTURES->get_wrap(_textures[i])) 
				{
					GFX->set_sampler_state(stage,D3DSAMP_ADDRESSU,D3DTADDRESS_WRAP);
					GFX->set_sampler_state(stage,D3DSAMP_ADDRESSV,D3DTADDRESS_WRAP);
				}
				else
				{
					GFX->set_sampler_state(stage,D3DSAMP_ADDRESSU,D3DTADDRESS_CLAMP);
					GFX->set_sampler_state(stage,D3DSAMP_ADDRESSV,D3DTADDRESS_CLAMP);
				}

				stage++;
			}
		}
		GFX->set_texture_stage_state(0,D3DTSS_COLOROP,D3DTOP_MODULATE);
		GFX->set_texture_stage_state(0,D3DTSS_COLORARG1,D3DTA_TEXTURE);
		GFX->set_texture_stage_state(0,D3DTSS_COLORARG2,D3DTA_DIFFUSE);
		GFX->set_texture_stage_state(0,D3DTSS_ALPHAOP,D3DTOP_MODULATE);
		GFX->set_texture_stage_state(0,D3DTSS_ALPHAARG1,D3DTA_TEXTURE);
		GFX->set_texture_stage_state(0,D3DTSS_ALPHAARG2,D3DTA_DIFFUSE);
		GFX->set_texture_stage_state(0,D3DTSS_TEXCOORDINDEX,0);
	}

	GFX->set_renderstate(D3DRS_ZWRITEENABLE,(_z_write)?(1):(0));
#endif
}

void Material::unset()
{
	GFX->set_renderstate(D3DRS_ZWRITEENABLE,_old_z_write);
#ifdef USE_SHADERS
	unset_shader();
#else
	GFX->set_renderstate(D3DRS_LIGHTING,_old_lighting);
	GFX->set_renderstate(D3DRS_ALPHATESTENABLE,_old_alpha_test);
	GFX->set_renderstate(D3DRS_ALPHAREF,_old_alpha_ref);
#endif
}

void Material::set_material_color(float r,float g,float b,float a)
{
	_material.Ambient.r=r; _material.Ambient.g=g; _material.Ambient.b=b; _material.Ambient.a=a; 
	_material.Diffuse.r=r; _material.Diffuse.g=g; _material.Diffuse.b=b; _material.Diffuse.a=a; 
	_material.Specular.r=0.0f; _material.Specular.g=0.0f; _material.Specular.b=0.0f; _material.Specular.a=0.0f; 
	_material.Emissive.r=0.0f; _material.Emissive.g=0.0f; _material.Emissive.b=0.0f; _material.Emissive.a=0.0f; 
	_material.Power=0.0f;
}

Color Material::get_material_color()
{
	return Color(_material.Diffuse.r,_material.Diffuse.g,_material.Diffuse.b,_material.Diffuse.a);
}

void Material::set_emissive_color(float r,float g,float b,float a)
{
	_material.Emissive.r=r; _material.Emissive.g=g; _material.Emissive.b=b; _material.Emissive.a=a; 
}

void Material::set_alpha(const float a)
{
	_material.Ambient.a=_material.Diffuse.a=_material.Emissive.a=a;
}

float Material::get_alpha()
{
	return _material.Diffuse.a;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
// Material managment
////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

std::vector<Material*>	g_materials;
Material*				g_current_material_3ds=NULL;

Material* get_material(const std::string& name)
{
	for (size_t i=0; i<g_materials.size(); i++)
	{
		if (name==g_materials[i]->get_name()) return g_materials[i];
	}
	return NULL;
}

void load_material_3ds(FILE* file,const std::string& current_filename)
{
	unsigned short	chunk_id;
	unsigned long	chunk_len;
	unsigned long	start_pos=ftell(file);

	fread (&chunk_id,sizeof(unsigned short),1,file); //Read the chunk header
	if (feof(file)) return;
	fread (&chunk_len,sizeof(unsigned long),1,file); //Read the length of the chunk 
	if (feof(file)) return;

	switch (chunk_id)
	{
		// Ignore chunks
		case 0xAFFF:		// Material chunk (parent)
		case 0xA200:		// Texture Map 1 chunk (parent)
		case 0xA020:		// Diffuse Color chunk
			break;
		// Interesting chunks
		case 0xA000:		// Material Name chunk
			{
				g_current_material_3ds=new Material;
				g_current_material_3ds->set_material_color(1.0f,1.0f,1.0f,1.0f);
				g_current_material_3ds->set_shader("basic");
				g_current_material_3ds->set_cull_mode(D3DCULL_CW);
				g_materials.push_back(g_current_material_3ds);

				char buffer[512];
				int		i=0;
				char	c;
				do 
				{
					fread(&c,sizeof(unsigned char),1,file);
					buffer[i]=c;
					i++;
				} while ((c!='\0') && (i<20));
				buffer[i+1]=0;	
				g_current_material_3ds->set_name(strip_directory(strip_extension(current_filename))+"_"+buffer);
			}
			break;
		case 0xA300:		// Texture Name chunk
			{
				char	buffer[512];
				int		i=0;
				char	c;
				do 
				{
					fread(&c,sizeof(unsigned char),1,file);
					buffer[i]=c;
					i++;
				} while ((c!='\0') && (i<20));
				buffer[i+1]=0;		
				// Find texture
				std::string s=buffer;
				if (!file_exists(s.c_str()))
				{
					s=get_directory(current_filename)+s;
				}
				long tex_id=TEXTURES->load(buffer,s.c_str());
				if (tex_id!=-1)
				{
					TEXTURES->set_wrap(tex_id,true);

					g_current_material_3ds->set_shader("basic");
					g_current_material_3ds->set_texture(0,tex_id);
				}
				// Check if there is a emissive texture
/*				s=strip_extension(s)+"_emissive.dds";
				if (file_exists(s.c_str()))
				{
					std::string name=buffer;
					name=name+"_emissive";
					tex_id=TEXTURES->load(name.c_str(),s.c_str());
					if (tex_id!=-1)
					{
						TEXTURES->set_wrap(tex_id,true);

						g_current_material_3ds->set_emissive_color(1.0f,1.0f,1.0f,1.0f);
						g_current_material_3ds->set_texture(1,tex_id);
					}
				}*/
			}
			break;
		case 0x0011:		// RGB chunk
			{
				unsigned char r, g, b;
				fread(&r,sizeof(unsigned char),1,file);
				fread(&g,sizeof(unsigned char),1,file);
				fread(&b,sizeof(unsigned char),1,file);
				g_current_material_3ds->set_material_color((float)r/255.0f,(float)g/255.0f,(float)b/255.0f,1.0f);
			}
		break;
		// Jump unparsed
		default:
			fseek(file,start_pos+chunk_len,SEEK_SET);
			break;
	}

	unsigned long bytes_read=0;
	unsigned long current_pos=ftell(file);
	while (current_pos<(start_pos+chunk_len))
	{
		load_material_3ds(file,current_filename);
		current_pos=ftell(file);
	}
}

Material* Material::clone()
{
	Material* material=new Material;
	material->_name=_name;
	memcpy(&material->_material,&_material,sizeof(D3DMATERIAL9));
	material->_cull_mode=_cull_mode;
	for (int i=0; i<MAX_TEX; i++) 
	{
		material->_textures[i]=_textures[i];
		material->_textures_filter[i]=_textures_filter[i];
	}
	material->_alpha_enable=_alpha_enable;
	material->_color_vertex=_color_vertex;
	material->_color_vertex_e=_color_vertex_e;
	material->_z_write=_z_write;
	material->_point_sprite=_point_sprite;
	material->_alpha_test=_alpha_test;
	material->_alpha_ref=_alpha_ref;
	material->_src_blend=_src_blend;
	material->_dest_blend=_dest_blend;
	material->_passes=_passes;
	material->_shader=_shader;
	auto it=_custom_params.begin();
	while (it!=_custom_params.end())
	{
		material->_custom_params.insert(std::make_pair(it->first,it->second));
		++it;
	}
	material->_old_z_write=_old_z_write;
	material->_old_lighting=_old_lighting;
	material->_old_alpha_test=_old_alpha_test;
	material->_old_alpha_ref=_old_alpha_ref;

	return  material;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// SHADER MANAGEMENT
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef USE_SHADERS
Material::Shaders	Material::g_shaders;
Material::Decls		Material::g_decls;

void Material::set_shader(const std::string& shader_name,RenderEnv* env,const int fvf)
{
	auto it=g_shaders.find(shader_name);
	if (it==g_shaders.end())
	{
		Shader* tmp=new Shader;
		tmp->_name=shader_name;

		g_shaders.insert(std::make_pair(shader_name,tmp));
		it=g_shaders.find(shader_name);
	}

	auto jt=it->second->_variants.find(*env);
	if (jt==it->second->_variants.end())
	{
		create_shader_variant(it->second,env,fvf);
		jt=it->second->_variants.find(*env);
		if (jt==it->second->_variants.end())
		{
			ERROR_MSG("Can't create shader variant!");
			return;
		}
	}

	set_constants(true,&jt->second->_vertex_constant);
	set_constants(false,&jt->second->_pixel_constant);

#ifdef USE_D3D9
	GFX->_device->SetFVF(NULL);
	GFX->_device->SetVertexDeclaration(jt->second->_decl);
	GFX->_device->SetVertexShader(jt->second->_vertex);
	GFX->_device->SetPixelShader(jt->second->_pixel);
#endif
}

void Material::unset_shader()
{

}

void Material::clear_shaders()
{
	auto it=g_shaders.begin();
	while (it!=g_shaders.end())
	{
		auto jt=it->second->_variants.begin();
		while (jt!=it->second->_variants.end())
		{
			if (jt->second->_vertex) jt->second->_vertex->Release();
			if (jt->second->_pixel) jt->second->_pixel->Release();
			jt->second->_vertex_constant.clear();
			jt->second->_pixel_constant.clear();
			jt->second->_defines.clear();
			delete jt->second;
			++jt;
		}
		it->second->_variants.clear();
		delete it->second;

		++it;
	}
	g_shaders.clear();

	auto jt=g_decls.begin();
	while (jt!=g_decls.end())
	{
		jt->second->Release();
		++jt;
	}
	g_decls.clear();
}

void Material::create_shader_variant(Shader* shader,RenderEnv* env,const int fvf)
{
	ShaderVariant* variant=new ShaderVariant;

	variant->_decl=get_vertex_decl(fvf);

	ShaderDefine	d;

	d._name="LIGHTTYPE_POINT"; d._value="1"; variant->_defines.push_back(d);
	d._name="LIGHTTYPE_SPOT"; d._value="2"; variant->_defines.push_back(d);
	d._name="LIGHTTYPE_DIR"; d._value="3"; variant->_defines.push_back(d);
					
	d._name="fog_enable";
	if (env->_props._fog_linear) d._value="1";
	else d._value="0";
	variant->_defines.push_back(d);

	d._name="light_count";
	d._value=string_format("%i",env->_props._light_count);
	variant->_defines.push_back(d);

	for (int i=0; i<MAX_LIGHT; i++)
	{
		d._name=string_format("type_light%i",i);
		if (i<env->_props._light_count)	d._value=string_format("%i",env->_props._lights[i]._type);
		else d._value="0";
		variant->_defines.push_back(d);
	}

	for (int i=0; i<MAX_TEX; i++)
	{
		d._name=string_format("map%i_exists",i);
		if ((env->_props._map[0])&(1<<i)) d._value="1";
		else d._value="0";
		variant->_defines.push_back(d);
	}

	d._name="alpha_test";
	if (env->_props._alpha_test_enabled) d._value="1";
	else d._value="0";
	variant->_defines.push_back(d);

	d._name="component_normal";
	if (env->_props._normal) d._value="1";
	else d._value="0";
	variant->_defines.push_back(d);
	d._name="component_psize";
	if (env->_props._psize) d._value="1";
	else d._value="0";
	variant->_defines.push_back(d);
	d._name="component_color0";
	if (env->_props._color0) d._value="1";
	else d._value="0";
	variant->_defines.push_back(d);
	d._name="component_color1";
	if (env->_props._color1) d._value="1";
	else d._value="0";
	variant->_defines.push_back(d);
	for (int i=0; i<MAX_TEX; i++)
	{
		d._name=string_format("component_tex%i",i);
		if (i<env->_props._tex_coord_count) d._value="1";
		else d._value="0";
		variant->_defines.push_back(d);
	}
	d._name="pass";
	d._value=string_format("%i",env->_props._pass);
	variant->_defines.push_back(d);

#ifdef USE_D3D9
	D3DXMACRO	defines[4*sizeof(unsigned long)*8+16];

	for (size_t i=0; i<variant->_defines.size(); i++)
	{
		defines[i].Name=variant->_defines[i]._name.c_str();
		defines[i].Definition=variant->_defines[i]._value.c_str();
	}

	defines[variant->_defines.size()].Name=NULL;
	defines[variant->_defines.size()].Definition=NULL;


	// Compile the vertex shader
	LPD3DXBUFFER				tmp=NULL;
	LPD3DXBUFFER				errors=NULL;
	ID3DXConstantTable*			constants;

	HRESULT hRet=D3DXCompileShaderFromFile(string_format("shaders/d3d9/%s.vhl",shader->_name.c_str()).c_str(),
										  defines,
										  NULL,
										  string_format("entry_%s",shader->_name.c_str()).c_str(),
										  "vs_3_0",0,&tmp,&errors,&constants);
	if (hRet!=D3D_OK)
	{
		log("Can't compile vertex shader (entry_%s)!",shader->_name.c_str());
		if (errors)
		{
			char* bugs=(char*)errors->GetBufferPointer();
			log(bugs);
			errors->Release();
		}
		log("Macros defined:");
		for (size_t i=0; i<variant->_defines.size(); i++)
		{
			log("%s = %s",variant->_defines[i]._name.c_str(),variant->_defines[i]._value.c_str());
		}

		ERROR_MSG(string_format("Can't compile vertex shader (%s)!",shader->_name.c_str()).c_str());
		delete variant;
		return;
	}

	hRet=GFX->_device->CreateVertexShader((DWORD*)tmp->GetBufferPointer(),&variant->_vertex);

	tmp->Release();

	if (!variant->_vertex)
	{
		ERROR_MSG("Can't create vertex shader!");
		delete variant;
		return;
	}

	process_constants(constants,&variant->_vertex_constant);
	constants->Release();
	
	hRet=D3DXCompileShaderFromFile(string_format("shaders/d3d9/%s.shl",shader->_name.c_str()).c_str(),
										  defines,
										  NULL,
										  string_format("entry_%s",shader->_name.c_str()).c_str(),
										  "ps_3_0",0,&tmp,&errors,&constants);
	if (hRet!=D3D_OK)
	{
		log("Can't compile pixel shader (entry_%s)!",shader->_name.c_str());
		if (errors)
		{
			char* bugs=(char*)errors->GetBufferPointer();
			log(bugs);
			errors->Release();
		}
		log("Macros defined:");
		for (size_t i=0; i<variant->_defines.size(); i++)
		{
			log("%s = %s",variant->_defines[i]._name.c_str(),variant->_defines[i]._value.c_str());
		}

		ERROR_MSG("Can't compile pixel shader!");
		variant->_vertex->Release();
		delete variant;
		return;
	}

	process_constants(constants,&variant->_pixel_constant);
	constants->Release();

	hRet=GFX->_device->CreatePixelShader((DWORD*)tmp->GetBufferPointer(),&variant->_pixel);

	tmp->Release();

	if (!variant->_pixel)
	{
		ERROR_MSG("Can't create pixel shader!");
		variant->_vertex->Release();
		delete variant;
		return;
	}
#endif

	shader->_variants.insert(std::make_pair(*env,variant));
}

VertexDeclaration Material::get_vertex_decl(const int fvf)
{
	auto it=g_decls.find(fvf);
	if (it==g_decls.end())
	{
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Build vertex declaration
		D3DVERTEXELEMENT9	decl[128];
		int					idx=0;
		int					ptr=0;
		
		// Pos
		decl[idx].Stream=0; decl[idx].Offset=ptr; decl[idx].Type=D3DDECLTYPE_FLOAT3; decl[idx].Method=D3DDECLMETHOD_DEFAULT; decl[idx].Usage=D3DDECLUSAGE_POSITION; decl[idx].UsageIndex=0; idx++; ptr+=sizeof(float)*3;
		// PSize
		if (fvf&D3DFVF_PSIZE)
		{
			decl[idx].Stream=0; decl[idx].Offset=ptr; decl[idx].Type=D3DDECLTYPE_FLOAT1; decl[idx].Method=D3DDECLMETHOD_DEFAULT; decl[idx].Usage=D3DDECLUSAGE_PSIZE; decl[idx].UsageIndex=0; idx++; ptr+=sizeof(float);
		}
		// Normal
		if (fvf&D3DFVF_NORMAL)
		{
			decl[idx].Stream=0; decl[idx].Offset=ptr; decl[idx].Type=D3DDECLTYPE_FLOAT3; decl[idx].Method=D3DDECLMETHOD_DEFAULT; decl[idx].Usage=D3DDECLUSAGE_NORMAL; decl[idx].UsageIndex=0; idx++; ptr+=sizeof(float)*3;
		}
		// Color0
		if (fvf&D3DFVF_DIFFUSE)
		{
			decl[idx].Stream=0; decl[idx].Offset=ptr; decl[idx].Type=D3DDECLTYPE_D3DCOLOR; decl[idx].Method=D3DDECLMETHOD_DEFAULT; decl[idx].Usage=D3DDECLUSAGE_COLOR; decl[idx].UsageIndex=0; idx++; ptr+=sizeof(unsigned char)*4;
		}
		// Color1
		if (fvf&D3DFVF_SPECULAR)
		{
			decl[idx].Stream=0; decl[idx].Offset=ptr; decl[idx].Type=D3DDECLTYPE_D3DCOLOR; decl[idx].Method=D3DDECLMETHOD_DEFAULT; decl[idx].Usage=D3DDECLUSAGE_COLOR; decl[idx].UsageIndex=1; idx++; ptr+=sizeof(unsigned char)*4;
		}
		for (int i=0; i<((fvf&D3DFVF_TEXCOUNT_MASK)>>D3DFVF_TEXCOUNT_SHIFT); i++)
		{
			decl[idx].Stream=0; decl[idx].Offset=ptr; decl[idx].Type=D3DDECLTYPE_FLOAT2; decl[idx].Method=D3DDECLMETHOD_DEFAULT; decl[idx].Usage=D3DDECLUSAGE_TEXCOORD; decl[idx].UsageIndex=i; idx++; ptr+=sizeof(float)*2;
		}
		decl[idx].Stream=0x0FF; decl[idx].Offset=0; decl[idx].Type=D3DDECLTYPE_UNUSED; decl[idx].Method=0; decl[idx].Usage=0; decl[idx].UsageIndex=0;

		LPDIRECT3DVERTEXDECLARATION9	decl_ptr;
		if (GFX->_device->CreateVertexDeclaration(decl,&decl_ptr)!=D3D_OK)
		{
			ERROR_MSG("Can't create vertex declaration!");
			return DECL_INVALID;
		}

		g_decls.insert(std::make_pair(fvf,decl_ptr));
		return decl_ptr;
	}
	return it->second;
}

#ifdef USE_D3D9
void Material::process_constants(ID3DXConstantTable* constants,Constants* dest)
{
	// Add bindings for the constants
	D3DXCONSTANTTABLE_DESC	desc;
	D3DXCONSTANT_DESC		c[8];
	UINT					count=8;
	ConstantType			binding;
	HRESULT					hRet;

	hRet=constants->GetDesc(&desc);

	for (unsigned long i=0; i<desc.Constants; i++)
	{
		D3DXHANDLE handle=constants->GetConstant(NULL,i);
		count=8;
		hRet=constants->GetConstantDesc(handle,&c[0],&count);


		switch (c[0].Type)
		{
			case D3DXPT_SAMPLER:
			case D3DXPT_SAMPLER2D:
			case D3DXPT_SAMPLER3D:
			case D3DXPT_SAMPLERCUBE:
				break;
		default:
			{
				binding=get_binding_by_name((char*)(c[0].Name));
			
				Constant cnst;
				cnst._type=binding;
				cnst._pos=c[0].RegisterIndex;
				cnst._size=c[0].RegisterCount;
				if (binding==CONST_CUSTOM) cnst._extra=c[0].Name;
				dest->push_back(cnst);
			}
			break;
		}
	}
}
#endif

Material::ConstantType Material::get_binding_by_name(const std::string& name)
{
	if (name=="clip_matrix") return CONST_CLIPMATRIX;
	if (name=="material_ambient") return CONST_MATERIAL_AMBIENT;
	if (name=="material_diffuse") return CONST_MATERIAL_DIFFUSE;
	if (name=="material_specular") return CONST_MATERIAL_SPECULAR;
	if (name=="material_specular_params") return CONST_MATERIAL_SPECULARPARAMS;
	if (name=="material_emissive") return CONST_MATERIAL_EMISSIVE;
	if (name=="ambient_light") return CONST_AMBIENT_LIGHT;
	if (name=="lights") return CONST_LIGHTS;
	if (name=="fog_center") return CONST_FOG_CENTER;
	if (name=="fog_params") return CONST_FOG_PARAMS;
	if (name=="fog_color") return CONST_FOG_COLOR;
	if (name=="rt_res") return CONST_RTRES;
	if (name=="global_time") return CONST_GLOBALTIME;
	if (name.substr(0,12)=="world_matrix") return (ConstantType)(CONST_WORLDMATRIX0+(atol(name.c_str()+12)));

	return CONST_CUSTOM;
}

void Material::set_constants(const bool vertex_constants,Constants* constants)
{
#ifdef USE_D3D9
	auto device=GFX->_device;
#endif
	float*	data_f;
	Vec4f	v4;
	auto	it=constants->begin();
	while (it!=constants->end())
	{
		switch (it->_type)
		{
			case CONST_CLIPMATRIX:
				data_f=(float*)GFX->get_clip_matrix();
				break;
			case CONST_MATERIAL_AMBIENT:
				data_f=(float*)&(_material.Ambient);
				break;
			case CONST_MATERIAL_DIFFUSE:
				data_f=(float*)&(_material.Diffuse);
				break;
			case CONST_MATERIAL_SPECULAR:
				data_f=(float*)&(_material.Specular);
				break;
			case CONST_MATERIAL_SPECULARPARAMS:
				v4.set(_material.Power,_material.Power,_material.Power,_material.Power);
				data_f=(float*)&v4;
				break;
			case CONST_MATERIAL_EMISSIVE:
				data_f=(float*)&(_material.Emissive);
				break;
			case CONST_AMBIENT_LIGHT:
				data_f=(float*)&GFX->get_ambient_light();
				break;
			case CONST_LIGHTS:
				data_f=(float*)GFX->get_lights();
				break;
			case CONST_FOG_CENTER:
				v4=GFX->get_fog_center();
				data_f=(float*)&v4;
				break;
			case CONST_FOG_COLOR:
				data_f=(float*)&GFX->get_fog_color();
				break;
			case CONST_FOG_PARAMS:
				v4=GFX->get_fog_range();
				data_f=(float*)&v4;
				break;
			case CONST_RTRES:
				v4=GFX->get_real_resolution();
				data_f=(float*)&v4;
				break;
			case CONST_GLOBALTIME:
				v4.x=((float)timeGetTime())*0.001f;
				v4.y=v4.x*0.1f;
				v4.z=v4.y*0.1f;
				v4.w=v4.z*0.1f;
				data_f=(float*)&v4;
				break;
			default:
				if ((it->_type>=CONST_WORLDMATRIX0) && (it->_type<=CONST_WORLDMATRIX255))
				{
					data_f=(float*)GFX->get_world_matrix(it->_type-CONST_WORLDMATRIX0);
				}
				else
				{
					auto jt=_custom_params.find(it->_extra);
					if (jt==_custom_params.end()) { ++it; continue; }
					data_f=(float*)&(jt->second);
				}
				break;
		}

#ifdef USE_D3D9
		if (vertex_constants) device->SetVertexShaderConstantF(it->_pos,data_f,it->_size);
		else device->SetPixelShaderConstantF(it->_pos,data_f,it->_size);
#endif
		++it;
	}
}

#endif

