#include "Mesh.h"
#include "Gfx.h"
#include "Utils.h"
#include "TextureManager.h"
#include "Camera.h"
#include "Light.h"

///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
BaseMesh::BaseMesh(const bool dynamic)
{
	_vb=NULL;
	_ib=NULL;
	_dynamic=dynamic;
	_name="";
	_material=NULL;
	_current_iterator_i=NULL;
	_iterator_start_i=NULL;
	_bound_material=false;
	_primitive=D3DPT_TRIANGLELIST;
	_n_index=0;
	_render_verts=0;
	_pmin=Vec3f(FLT_MAX,FLT_MAX,FLT_MAX);
	_pmax=Vec3f(-FLT_MAX,-FLT_MAX,-FLT_MAX);
	_n_verts=0;
	if (dynamic) _keep_data=true;
	else _keep_data=false;
}

BaseMesh::~BaseMesh()
{
	if (_current_iterator_i) 
	{
		release_write_pointer_i();
	}
	if (_vb) 
	{
		delete _vb;
		_vb=NULL;
	}
	if (_ib) 
	{
		delete _ib;
		_ib=NULL;
	}
	if (_material)
	{
		if (_bound_material)
		{
			delete _material;
			_material=NULL;
		}
	}
}

bool BaseMesh::alloc_index(const UINT n)
{
	if (_dynamic)
	{
		if (_n_index>=n) return true;
	}

	if (_ib) delete _ib;

	_ib=new IndexBuffer;
	if (!_ib->create(n,_dynamic))
	{
		_ib=NULL;
		return false;
	}

	_n_index=n;

	return true;
}

WORD* BaseMesh::get_write_pointer_i()
{
	if (_current_iterator_i) return _current_iterator_i;

	_current_iterator_i=_ib->get_write_pointer();
	_iterator_start_i=_current_iterator_i;

	return _current_iterator_i;
}

void BaseMesh::release_write_pointer_i()
{
	if (!_current_iterator_i) return;

	if (_keep_data)
	{
		_index_array.resize(_n_index);
		memcpy(&_index_array[0],_iterator_start_i,_n_index*sizeof(WORD));
	}

	_ib->release_write_pointer();

	_current_iterator_i=NULL;
	_iterator_start_i=NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////

BaseMesh* build_cube_pn(const float sx,const float sy,const float sz)
{
	Mesh<VertexPN>*	mesh=new Mesh<VertexPN>(false);

	if (!mesh->alloc_vert(24)) 
	{
		delete mesh;
		return NULL;
	}

	float dx=sx*0.5f;
	float dy=sy*0.5f;
	float dz=sz*0.5f;

	// Build vertices
	VertexPN* ptr=mesh->get_write_pointer_v();

	// Bottom face
	ptr->pos.set(-dx,-dy,-dz); ptr->normal.set( 0.0f,-1.0f, 0.0f); ptr++;
	ptr->pos.set( dx,-dy,-dz); ptr->normal.set( 0.0f,-1.0f, 0.0f); ptr++;
	ptr->pos.set(-dx,-dy, dz); ptr->normal.set( 0.0f,-1.0f, 0.0f); ptr++;
	ptr->pos.set( dx,-dy, dz); ptr->normal.set( 0.0f,-1.0f, 0.0f); ptr++;
	// Top face
	ptr->pos.set(-dx, dy,-dz); ptr->normal.set( 0.0f, 1.0f, 0.0f); ptr++;
	ptr->pos.set( dx, dy,-dz); ptr->normal.set( 0.0f, 1.0f, 0.0f); ptr++;
	ptr->pos.set(-dx, dy, dz); ptr->normal.set( 0.0f, 1.0f, 0.0f); ptr++;
	ptr->pos.set( dx, dy, dz); ptr->normal.set( 0.0f, 1.0f, 0.0f); ptr++;
	// Right face
	ptr->pos.set( dx,-dy,-dz); ptr->normal.set( 1.0f, 0.0f, 0.0f); ptr++;
	ptr->pos.set( dx,-dy, dz); ptr->normal.set( 1.0f, 0.0f, 0.0f); ptr++;
	ptr->pos.set( dx, dy,-dz); ptr->normal.set( 1.0f, 0.0f, 0.0f); ptr++;
	ptr->pos.set( dx, dy, dz); ptr->normal.set( 1.0f, 0.0f, 0.0f); ptr++;
	// Left face
	ptr->pos.set(-dx,-dy,-dz); ptr->normal.set(-1.0f, 0.0f, 0.0f); ptr++;
	ptr->pos.set(-dx,-dy, dz); ptr->normal.set(-1.0f, 0.0f, 0.0f); ptr++;
	ptr->pos.set(-dx, dy,-dz); ptr->normal.set(-1.0f, 0.0f, 0.0f); ptr++;
	ptr->pos.set(-dx, dy, dz); ptr->normal.set(-1.0f, 0.0f, 0.0f); ptr++;
	// Front face
	ptr->pos.set(-dx,-dy,-dz); ptr->normal.set( 0.0f, 0.0f,-1.0f); ptr++;
	ptr->pos.set( dx,-dy,-dz); ptr->normal.set( 0.0f, 0.0f,-1.0f); ptr++;
	ptr->pos.set(-dx, dy,-dz); ptr->normal.set( 0.0f, 0.0f,-1.0f); ptr++;
	ptr->pos.set( dx, dy,-dz); ptr->normal.set( 0.0f, 0.0f,-1.0f); ptr++;
	// Back face
	ptr->pos.set(-dx,-dy, dz); ptr->normal.set( 0.0f, 0.0f, 1.0f); ptr++;
	ptr->pos.set( dx,-dy, dz); ptr->normal.set( 0.0f, 0.0f, 1.0f); ptr++;
	ptr->pos.set(-dx, dy, dz); ptr->normal.set( 0.0f, 0.0f, 1.0f); ptr++;
	ptr->pos.set( dx, dy, dz); ptr->normal.set( 0.0f, 0.0f, 1.0f); ptr++;

	mesh->release_write_pointer_v();

	// Build index
	if (!mesh->alloc_index(36))
	{
		delete mesh;
		return NULL;
	}

	WORD* iptr=mesh->get_write_pointer_i();

	*iptr= 0; iptr++; *iptr= 2; iptr++; *iptr= 1; iptr++;
	*iptr= 1; iptr++; *iptr= 2; iptr++; *iptr= 3; iptr++;
	*iptr= 4; iptr++; *iptr= 5; iptr++; *iptr= 6; iptr++;
	*iptr= 5; iptr++; *iptr= 7; iptr++; *iptr= 6; iptr++;
	*iptr= 8; iptr++; *iptr= 9; iptr++; *iptr=10; iptr++;
	*iptr= 9; iptr++; *iptr=11; iptr++; *iptr=10; iptr++;
	*iptr=12; iptr++; *iptr=14; iptr++; *iptr=13; iptr++;
	*iptr=13; iptr++; *iptr=14; iptr++; *iptr=15; iptr++;
	*iptr=16; iptr++; *iptr=17; iptr++; *iptr=18; iptr++;
	*iptr=17; iptr++; *iptr=19; iptr++; *iptr=18; iptr++;
	*iptr=20; iptr++; *iptr=22; iptr++; *iptr=21; iptr++;
	*iptr=21; iptr++; *iptr=22; iptr++; *iptr=23; iptr++;

	mesh->release_write_pointer_i();

	return mesh;
}

BaseMesh* build_cube_pnt(const float sx,const float sy,const float sz,long tex_id)
{
	Mesh<VertexPNT>*	mesh=new Mesh<VertexPNT>(false);

	if (!mesh->alloc_vert(24)) 
	{
		delete mesh;
		return NULL;
	}

	float dx=sx*0.5f;
	float dy=sy*0.5f;
	float dz=sz*0.5f;

	// Build vertices
	VertexPNT* ptr=mesh->get_write_pointer_v();

	// Bottom face
	ptr->pos.set(-dx,-dy,-dz); ptr->normal.set( 0.0f,-1.0f, 0.0f); ptr->uv.set(0.0f,0.0f); ptr++;
	ptr->pos.set( dx,-dy,-dz); ptr->normal.set( 0.0f,-1.0f, 0.0f); ptr->uv.set(1.0f,0.0f); ptr++;
	ptr->pos.set(-dx,-dy, dz); ptr->normal.set( 0.0f,-1.0f, 0.0f); ptr->uv.set(0.0f,1.0f); ptr++;
	ptr->pos.set( dx,-dy, dz); ptr->normal.set( 0.0f,-1.0f, 0.0f); ptr->uv.set(1.0f,1.0f); ptr++;
	// Top face
	ptr->pos.set(-dx, dy,-dz); ptr->normal.set( 0.0f, 1.0f, 0.0f); ptr->uv.set(0.0f,1.0f); ptr++;
	ptr->pos.set( dx, dy,-dz); ptr->normal.set( 0.0f, 1.0f, 0.0f); ptr->uv.set(1.0f,1.0f); ptr++;
	ptr->pos.set(-dx, dy, dz); ptr->normal.set( 0.0f, 1.0f, 0.0f); ptr->uv.set(0.0f,0.0f); ptr++;
	ptr->pos.set( dx, dy, dz); ptr->normal.set( 0.0f, 1.0f, 0.0f); ptr->uv.set(1.0f,0.0f); ptr++;
	// Right face
	ptr->pos.set( dx,-dy,-dz); ptr->normal.set( 1.0f, 0.0f, 0.0f); ptr->uv.set(0.0f,1.0f); ptr++;
	ptr->pos.set( dx,-dy, dz); ptr->normal.set( 1.0f, 0.0f, 0.0f); ptr->uv.set(1.0f,1.0f); ptr++;
	ptr->pos.set( dx, dy,-dz); ptr->normal.set( 1.0f, 0.0f, 0.0f); ptr->uv.set(0.0f,0.0f); ptr++;
	ptr->pos.set( dx, dy, dz); ptr->normal.set( 1.0f, 0.0f, 0.0f); ptr->uv.set(1.0f,0.0f); ptr++;
	// Left face
	ptr->pos.set(-dx,-dy,-dz); ptr->normal.set(-1.0f, 0.0f, 0.0f); ptr->uv.set(1.0f,1.0f); ptr++;
	ptr->pos.set(-dx,-dy, dz); ptr->normal.set(-1.0f, 0.0f, 0.0f); ptr->uv.set(0.0f,1.0f); ptr++;
	ptr->pos.set(-dx, dy,-dz); ptr->normal.set(-1.0f, 0.0f, 0.0f); ptr->uv.set(1.0f,0.0f); ptr++;
	ptr->pos.set(-dx, dy, dz); ptr->normal.set(-1.0f, 0.0f, 0.0f); ptr->uv.set(0.0f,0.0f); ptr++;
	// Front face
	ptr->pos.set(-dx,-dy,-dz); ptr->normal.set( 0.0f, 0.0f,-1.0f); ptr->uv.set(0.0f,1.0f); ptr++;
	ptr->pos.set( dx,-dy,-dz); ptr->normal.set( 0.0f, 0.0f,-1.0f); ptr->uv.set(1.0f,1.0f); ptr++;
	ptr->pos.set(-dx, dy,-dz); ptr->normal.set( 0.0f, 0.0f,-1.0f); ptr->uv.set(0.0f,0.0f); ptr++;
	ptr->pos.set( dx, dy,-dz); ptr->normal.set( 0.0f, 0.0f,-1.0f); ptr->uv.set(1.0f,0.0f); ptr++;
	// Back face
	ptr->pos.set(-dx,-dy, dz); ptr->normal.set( 0.0f, 0.0f, 1.0f); ptr->uv.set(1.0f,1.0f); ptr++;
	ptr->pos.set( dx,-dy, dz); ptr->normal.set( 0.0f, 0.0f, 1.0f); ptr->uv.set(0.0f,1.0f); ptr++;
	ptr->pos.set(-dx, dy, dz); ptr->normal.set( 0.0f, 0.0f, 1.0f); ptr->uv.set(1.0f,0.0f); ptr++;
	ptr->pos.set( dx, dy, dz); ptr->normal.set( 0.0f, 0.0f, 1.0f); ptr->uv.set(0.0f,0.0f); ptr++;

	mesh->release_write_pointer_v();

	// Build index
	if (!mesh->alloc_index(36))
	{
		delete mesh;
		return NULL;
	}

	WORD* iptr=mesh->get_write_pointer_i();

	*iptr= 0; iptr++; *iptr= 2; iptr++; *iptr= 1; iptr++;
	*iptr= 1; iptr++; *iptr= 2; iptr++; *iptr= 3; iptr++;
	*iptr= 4; iptr++; *iptr= 5; iptr++; *iptr= 6; iptr++;
	*iptr= 5; iptr++; *iptr= 7; iptr++; *iptr= 6; iptr++;
	*iptr= 8; iptr++; *iptr= 9; iptr++; *iptr=10; iptr++;
	*iptr= 9; iptr++; *iptr=11; iptr++; *iptr=10; iptr++;
	*iptr=12; iptr++; *iptr=14; iptr++; *iptr=13; iptr++;
	*iptr=13; iptr++; *iptr=14; iptr++; *iptr=15; iptr++;
	*iptr=16; iptr++; *iptr=17; iptr++; *iptr=18; iptr++;
	*iptr=17; iptr++; *iptr=19; iptr++; *iptr=18; iptr++;
	*iptr=20; iptr++; *iptr=22; iptr++; *iptr=21; iptr++;
	*iptr=21; iptr++; *iptr=22; iptr++; *iptr=23; iptr++;

	mesh->release_write_pointer_i();

	return mesh;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 3ds loader (very simple, just for meshes)
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void load_and_parse_chunk(FILE* file,std::vector<BaseMesh*> *meshes,std::vector<Camera*>* cameras,std::vector<Light*>* lights,const std::string& current_filename)
{
	static	std::string		current_name;
			unsigned short	chunk_id;
			unsigned long	chunk_len;
			unsigned long	start_pos=ftell(file);

	fread (&chunk_id,sizeof(unsigned short),1,file); //Read the chunk header
	if (feof(file)) return;
	fread (&chunk_len,sizeof(unsigned long),1,file); //Read the length of the chunk 
	if (feof(file)) return;

	Mesh<VertexPNT>*	mesh=NULL;
	if (meshes)
	{
		if (!meshes->empty())
		{
			mesh=(Mesh<VertexPNT>*)meshes->at(meshes->size()-1);
		}
	}
	Light*				lgt=NULL;
	if (lights)
	{
		if (!lights->empty())
		{
			lgt=lights->at(lights->size()-1);
		}
	}

	switch (chunk_id)
	{
		// Ignore chunks
		case 0x4D4D:		// Main chunk
		case 0x4100:		// Triangular Mesh chunk (parent)
		case 0xA200:		// Texture Map 1 chunk (parent)
		case 0x3D3D:		// Object Block chunk
			break;
		// Interesting chunks
		case 0xAFFF:		// Material chunk (parent)
			{
				fseek(file,-6,SEEK_CUR);
				load_material_3ds(file,current_filename);
			}
			break;
		case 0x4000:		// Object Block chunk
			{
				char buffer[512];
				int		i=0;
				char	c;
				do 
				{
					fread(&c,sizeof(unsigned char),1,file);
					buffer[i]=c;
					i++;
				} while ((c!='\0') && (i<20));
				buffer[i+1]=0;
				current_name=buffer;
			}
			break;
		case 0x4110:		// Vertices List chunk
			{
				mesh=new Mesh<VertexPNT>(false);
				meshes->push_back(mesh);
				mesh->set_name(current_name);

				float max_y=0.0f;

				unsigned short n_vert;
				fread(&n_vert,sizeof(unsigned short),1,file);
				mesh->alloc_vert(n_vert);
				VertexPNT* dest=mesh->get_write_pointer_v();
				for (unsigned short i=0; i<n_vert; i++)
				{
					fread(&(dest->pos.x),sizeof(float),3,file);
					// Scale vertex (this is probably not very correct, but it serves my purposes for now) - Scale factor=0.035 (trial and error)
					// Swap Y and Z also in the process
					dest->pos.set(dest->pos.x*0.035f,dest->pos.z*0.035f,dest->pos.y*0.035f);

					if (dest->pos.y>max_y) max_y=dest->pos.y;
					dest++;
				}
			}
			break;
		case 0x4120:		// Faces Description chunk
			{
				unsigned short n_idx;
				unsigned short face_flags;
				fread(&n_idx,sizeof(unsigned short),1,file);
				mesh->alloc_index(n_idx*3);
				WORD* dest=mesh->get_write_pointer_i();
				for (unsigned short i=0; i<n_idx; i++)
				{
					fread(dest,sizeof(unsigned short),1,file); dest++;
					fread(dest,sizeof(unsigned short),1,file); dest++;
					fread(dest,sizeof(unsigned short),1,file); dest++;
					fread(&face_flags,sizeof(unsigned short),1,file);
				}
			}
			break;
		case 0x4130:		// Faces Material List chunk
			{
				char buffer[512];
				int		i=0;
				char	c;
				do 
				{
					fread(&c,sizeof(unsigned char),1,file);
					buffer[i]=c;
					i++;
				} while ((c!='\0') && (i<20));
				buffer[i+1]=0;
				
				Material* material=get_material(strip_directory(strip_extension(current_filename))+"_"+buffer);
				mesh->set_material(material,false);

				WORD n_entries;
				fread(&n_entries,sizeof(WORD),1,file);
				fseek(file,n_entries*sizeof(WORD),SEEK_CUR);
			}
			break;
		case 0x4140:		// Mapping Coordinates List chunk
			{
				unsigned short n_vert;
				fread(&n_vert,sizeof(unsigned short),1,file);
				VertexPNT* dest=mesh->get_write_pointer_v();
				for (unsigned short i=0; i<n_vert; i++)
				{
					fread(&(dest->uv.x),sizeof(float),2,file);
					dest->uv.y=1.0f-dest->uv.y;
					dest++;
				}
				mesh->release_write_pointer_v();
			}
			break;
		case 0x4600:
			{
				Vec3f	pos;
				fread(&(pos.x),sizeof(float),1,file);
				fread(&(pos.z),sizeof(float),1,file);
				fread(&(pos.y),sizeof(float),1,file);
		
				lgt=new Light(current_name.c_str());
				lgt->_pos=pos;
				lgt->set_type(LIGHT_POINT);

				lights->push_back(lgt);
			}
			break;
		case 0x4700:
			{
				Vec3f	pos;
				Vec3f	target;
				float	bank;
				float	lens;

				fread(&(pos.x),sizeof(float),1,file);
				fread(&(pos.z),sizeof(float),1,file);
				fread(&(pos.y),sizeof(float),1,file);
				fread(&(target.x),sizeof(float),1,file);
				fread(&(target.z),sizeof(float),1,file);
				fread(&(target.y),sizeof(float),1,file);
				fread(&bank,sizeof(float),1,file);
				fread(&lens,sizeof(float),1,file);

				pos=pos*0.035f;
				target=target*0.035f;

				Camera* cam=new Camera(current_name.c_str());
				cam->set_lookat(pos,target);
				cam->set_fov(lens);
				cameras->push_back(cam);
			}
			break;
		// Jump unparsed
		default:
			fseek(file,start_pos+chunk_len,SEEK_SET);
			break;
	}

	unsigned long bytes_read=0;
	unsigned long current_pos=ftell(file);
	while (current_pos<(start_pos+chunk_len))
	{
		load_and_parse_chunk(file,meshes,cameras,lights,current_filename);
		current_pos=ftell(file);
	}
}

bool load_3ds(const std::string& filename,std::vector<BaseMesh*>* meshes,std::vector<Camera*>* cameras,std::vector<Light*>* lights)
{
	FILE* file=NULL;
	fopen_s(&file,filename.c_str(),"rb");
	if (!file) return false;

	while (!feof(file))
	{
		load_and_parse_chunk(file,meshes,cameras,lights,filename);
	}

	for (size_t i=0; i<meshes->size(); i++)
	{
		build_normals_pnt((Mesh<VertexPNT>*)meshes->at(i));
	}

	fclose(file);

	return true;
}

void build_normals_pnt(Mesh<VertexPNT>* mesh)
{
	VertexPNT* v=mesh->get_write_pointer_v();

	VertexPNT* dest=v;
	for (unsigned long i=0; i<mesh->get_vertex_count(); i++)
	{
		dest->normal.set(0.0f,0.0f,0.0f);
		dest++;
	}
	
	WORD*	i_ptr=mesh->get_write_pointer_i();
	Vec3f	n;
	Vec3f	a,b,c;
	Vec3f	v1,v2;
	for (unsigned long i=0; i<mesh->get_index_count()/3; i++)
	{
		// Get triangle normal
		a=v[i_ptr[0]].pos;
		b=v[i_ptr[1]].pos;
		c=v[i_ptr[2]].pos;
		v1=c-a;
		v2=b-a;
		n=cross_product(v1,v2);
		v[i_ptr[0]].normal=v[i_ptr[0]].normal+n;
		v[i_ptr[1]].normal=v[i_ptr[1]].normal+n;
		v[i_ptr[2]].normal=v[i_ptr[2]].normal+n;

		i_ptr+=3;
	}

	dest=v;
	for (unsigned long i=0; i<mesh->get_vertex_count(); i++)
	{
		dest->normal.normalize();
		dest++;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
GeomMesh* geom_add_quad(GeomMesh* src_mesh,const Vec3f& center,const Vec3f& dir0,const Vec3f& dir1,const unsigned long color,const Rect& uv_rect)
{
	GeomMesh* dest=(src_mesh)?(src_mesh):(new GeomMesh(false));

	Vec3f		normal=cross_product(dir1,dir0); normal.normalize();
	VertexPNCT	vertex;
	vertex.normal=normal;
	vertex.color=color;

	size_t base_idx=dest->get_current_vertex_count();
	vertex.pos=center-dir0*0.5f+dir1*0.5f; vertex.uv.set(uv_rect.x1,uv_rect.y1); dest->push_vertex(vertex);
	vertex.pos=center+dir0*0.5f+dir1*0.5f; vertex.uv.set(uv_rect.x2,uv_rect.y1); dest->push_vertex(vertex);
	vertex.pos=center+dir0*0.5f-dir1*0.5f; vertex.uv.set(uv_rect.x2,uv_rect.y2); dest->push_vertex(vertex);
	vertex.pos=center-dir0*0.5f-dir1*0.5f; vertex.uv.set(uv_rect.x1,uv_rect.y2); dest->push_vertex(vertex);
	
	dest->push_index(base_idx); dest->push_index(base_idx+2); dest->push_index(base_idx+1);
	dest->push_index(base_idx); dest->push_index(base_idx+3); dest->push_index(base_idx+2);

	return dest;
}

GeomMesh* geom_add_quad(VertexPNCT*& v_ptr,WORD*& i_ptr,int& vertex_count,int& index_count,const Vec3f& center,const Vec3f& dir0,const Vec3f& dir1,const unsigned long color,const Rect& uv_rect)
{
	Vec3f		normal=cross_product(dir1,dir0); normal.normalize();

	v_ptr->pos=center-dir0*0.5f+dir1*0.5f; v_ptr->uv.set(uv_rect.x1,uv_rect.y1); v_ptr->color=color; v_ptr->normal=normal; v_ptr++; 
	v_ptr->pos=center+dir0*0.5f+dir1*0.5f; v_ptr->uv.set(uv_rect.x2,uv_rect.y1); v_ptr->color=color; v_ptr->normal=normal; v_ptr++; 
	v_ptr->pos=center+dir0*0.5f-dir1*0.5f; v_ptr->uv.set(uv_rect.x2,uv_rect.y2); v_ptr->color=color; v_ptr->normal=normal; v_ptr++; 
	v_ptr->pos=center-dir0*0.5f-dir1*0.5f; v_ptr->uv.set(uv_rect.x1,uv_rect.y2); v_ptr->color=color; v_ptr->normal=normal; v_ptr++; 

	*i_ptr=vertex_count; i_ptr++;
	*i_ptr=vertex_count+2; i_ptr++;
	*i_ptr=vertex_count+1; i_ptr++;
	*i_ptr=vertex_count; i_ptr++;
	*i_ptr=vertex_count+3; i_ptr++;
	*i_ptr=vertex_count+2; i_ptr++;

	vertex_count+=4;
	index_count+=6;

	return NULL;
}

GeomMesh* geom_add_quad(VertexPNCCT*& v_ptr,WORD*& i_ptr,int& vertex_count,int& index_count,const Vec3f& center,const Vec3f& dir0,const Vec3f& dir1,const unsigned long color,const unsigned long color2,const Rect& uv_rect)
{
	Vec3f		normal=cross_product(dir1,dir0); normal.normalize();

	v_ptr->pos=center-dir0*0.5f+dir1*0.5f; v_ptr->uv.set(uv_rect.x1,uv_rect.y1); v_ptr->color1=color; v_ptr->color2=color2; v_ptr->normal=normal; v_ptr++; 
	v_ptr->pos=center+dir0*0.5f+dir1*0.5f; v_ptr->uv.set(uv_rect.x2,uv_rect.y1); v_ptr->color1=color; v_ptr->color2=color2; v_ptr->normal=normal; v_ptr++; 
	v_ptr->pos=center+dir0*0.5f-dir1*0.5f; v_ptr->uv.set(uv_rect.x2,uv_rect.y2); v_ptr->color1=color; v_ptr->color2=color2; v_ptr->normal=normal; v_ptr++; 
	v_ptr->pos=center-dir0*0.5f-dir1*0.5f; v_ptr->uv.set(uv_rect.x1,uv_rect.y2); v_ptr->color1=color; v_ptr->color2=color2; v_ptr->normal=normal; v_ptr++; 

	*i_ptr=vertex_count; i_ptr++;
	*i_ptr=vertex_count+2; i_ptr++;
	*i_ptr=vertex_count+1; i_ptr++;
	*i_ptr=vertex_count; i_ptr++;
	*i_ptr=vertex_count+3; i_ptr++;
	*i_ptr=vertex_count+2; i_ptr++;

	vertex_count+=4;
	index_count+=6;

	return NULL;
}
GeomMesh* geom_add_cube(GeomMesh* src_mesh,
						const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
						const unsigned long color,const unsigned char faces,Rect* uv)
{
	GeomMesh* dest=(src_mesh)?(src_mesh):(new GeomMesh(false));

	if (faces&FACE_MZ) geom_add_quad(dest,center-dir_z*0.5f,dir_x,dir_y,color,uv[0]);
	if (faces&FACE_PZ) geom_add_quad(dest,center+dir_z*0.5f,-dir_x,dir_y,color,uv[1]);
	if (faces&FACE_MX) geom_add_quad(dest,center-dir_x*0.5f,-dir_z,dir_y,color,uv[2]);
	if (faces&FACE_PX) geom_add_quad(dest,center+dir_x*0.5f,dir_z,dir_y,color,uv[3]);
	if (faces&FACE_MY) geom_add_quad(dest,center-dir_y*0.5f,dir_x,-dir_z,color,uv[4]);
	if (faces&FACE_PY) geom_add_quad(dest,center+dir_y*0.5f,dir_x,dir_z,color,uv[5]);

	return dest;
}


GeomMesh* geom_add_cube(VertexPNCT*& v_ptr,WORD*& i_ptr,int& vertex_count,int& index_count,
						const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
						const unsigned long color,const unsigned char faces,Rect* uv)
{
	if (faces&FACE_MZ) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center-dir_z*0.5f,dir_x,dir_y,color,uv[0]);
	if (faces&FACE_PZ) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center+dir_z*0.5f,-dir_x,dir_y,color,uv[1]);
	if (faces&FACE_MX) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center-dir_x*0.5f,-dir_z,dir_y,color,uv[2]);
	if (faces&FACE_PX) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center+dir_x*0.5f,dir_z,dir_y,color,uv[3]);
	if (faces&FACE_MY) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center-dir_y*0.5f,dir_x,-dir_z,color,uv[4]);
	if (faces&FACE_PY) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center+dir_y*0.5f,dir_x,dir_z,color,uv[5]);

	return NULL;
}

GeomMesh* geom_add_cube(VertexPNCCT*& v_ptr,WORD*& i_ptr,int& vertex_count,int& index_count,
						const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
						const unsigned long color,const unsigned long color2,const unsigned char faces,Rect* uv)
{
	if (faces&FACE_MZ) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center-dir_z*0.5f,dir_x,dir_y,color,color2,uv[0]);
	if (faces&FACE_PZ) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center+dir_z*0.5f,-dir_x,dir_y,color,color2,uv[1]);
	if (faces&FACE_MX) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center-dir_x*0.5f,-dir_z,dir_y,color,color2,uv[2]);
	if (faces&FACE_PX) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center+dir_x*0.5f,dir_z,dir_y,color,color2,uv[3]);
	if (faces&FACE_MY) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center-dir_y*0.5f,dir_x,-dir_z,color,color2,uv[4]);
	if (faces&FACE_PY) geom_add_quad(v_ptr,i_ptr,vertex_count,index_count,center+dir_y*0.5f,dir_x,dir_z,color,color2,uv[5]);

	return NULL;
}

GeomMesh* geom_add_spheroid(GeomMesh* src_mesh,
							const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
							const float min_lat,const float max_lat,const float lat_inc,
							const float min_lon,const float max_lon,const float lon_inc,
							const unsigned long color,const Rect& rect_uv)
{
	GeomMesh* dest=(src_mesh)?(src_mesh):(new GeomMesh(false));

	VertexPNCT	v;
	v.color=color;

	float	lat=min_lat;
	int		lat_count=0;
	int		lon_count=0;

	size_t	base_idx=dest->get_current_vertex_count();

	while (lat<=max_lat)
	{
		lat_count++;

		float lon=min_lon;
		
		lon_count=0;
		
		float a=-cosf(deg_to_rad(lat));
		float ia=sinf(deg_to_rad(lat));

		v.uv.y=(lat-min_lat)/(max_lat-min_lat);
		v.uv.y=v.uv.y*rect_uv.y2+(1.0f-v.uv.y)*rect_uv.y1;

		while (lon<=max_lon)
		{
			lon_count++;

			float s=sinf(deg_to_rad(lon))*ia;
			float c=cosf(deg_to_rad(lon))*ia;

			v.pos=dir_x*c*0.5f+dir_y*a*0.5f+dir_z*s*0.5f;
			v.normal=v.pos; v.normal.normalize();
			v.pos=v.pos+center;
			v.uv.x=(lon-min_lon)/(max_lon-min_lon);
			v.uv.x=v.uv.x*rect_uv.x2+(1.0f-v.uv.x)*rect_uv.x1;
			dest->push_vertex(v);

			lon+=lon_inc;
		}

		lat+=lat_inc;
	}

	int ly=lat_count;
	int lx=lon_count;

	for (int yy=0; yy<ly-1; yy++)
	{
		for (int xx=0; xx<lx-1; xx++)
		{
			dest->push_index(base_idx+yy*lx+xx); dest->push_index(base_idx+yy*lx+xx+1); dest->push_index(base_idx+(yy+1)*lx+xx+1);
			dest->push_index(base_idx+yy*lx+xx); dest->push_index(base_idx+(yy+1)*lx+xx+1); dest->push_index(base_idx+(yy+1)*lx+xx);
		}
	}

	return dest;
}

GeomMesh* geom_add_spheroid(GeomMesh* src_mesh,
							const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
							const float min_lat,const float max_lat,const float lat_inc,
							const float min_lon,const float max_lon,const float lon_inc,
							Interpolator<Color>* color,const Rect& rect_uv)
{
	GeomMesh* dest=(src_mesh)?(src_mesh):(new GeomMesh(false));

	VertexPNCT	v;

	float	lat=min_lat;
	int		lat_count=0;
	int		lon_count=0;

	size_t	base_idx=dest->get_current_vertex_count();

	while (lat<=max_lat)
	{
		lat_count++;

		float lon=min_lon;

		lon_count=0;

		float a=-cosf(deg_to_rad(lat));
		float ia=sinf(deg_to_rad(lat));
		float t_lat=(lat-min_lat)/(max_lat-min_lat);
		v.uv.y=t_lat;
		v.color=float_to_colorref(color->get(t_lat));

		while (lon<=max_lon)
		{
			lon_count++;

			float s=sinf(deg_to_rad(lon))*ia;
			float c=cosf(deg_to_rad(lon))*ia;

			v.pos=dir_x*c*0.5f+dir_y*a*0.5f+dir_z*s*0.5f;
			v.normal=v.pos; v.normal.normalize();
			v.pos=v.pos+center;
			v.uv.x=(lon-min_lon)/(max_lon-min_lon);
			dest->push_vertex(v);

			lon+=lon_inc;
		}

		lat+=lat_inc;
	}

	int ly=lat_count;
	int lx=lon_count;

	for (int yy=0; yy<ly-1; yy++)
	{
		for (int xx=0; xx<lx-1; xx++)
		{
			dest->push_index(base_idx+yy*lx+xx); dest->push_index(base_idx+yy*lx+xx+1); dest->push_index(base_idx+(yy+1)*lx+xx+1);
			dest->push_index(base_idx+yy*lx+xx); dest->push_index(base_idx+(yy+1)*lx+xx+1); dest->push_index(base_idx+(yy+1)*lx+xx);
		}
	}

	return dest;
}

GeomMesh* geom_add_cap(GeomMesh* src_mesh,
					   const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
					   const float min_lon,const float max_lon,const float lon_inc,
					   const unsigned long color,const Rect& rect_uv)
{
	GeomMesh* dest=(src_mesh)?(src_mesh):(new GeomMesh(false));

	size_t base_index=dest->get_current_vertex_count();
	// Apex vertex
	VertexPNCT	appex;
	appex.pos=center+dir_y;
	appex.normal=cross_product(dir_z,dir_x); appex.normal.normalize();
	appex.color=color;
	appex.uv.set(rect_uv.x1+(rect_uv.x2-rect_uv.x1)*0.5f,
				 rect_uv.y1+(rect_uv.y2-rect_uv.y1)*0.5f);
	dest->push_vertex(appex);

	// Outlying area
	int			lon_count=0;
	float		lon=min_lon;
	float		su=(rect_uv.x2-rect_uv.x1)*0.5f;
	float		sv=(rect_uv.x2-rect_uv.x1)*0.5f;
	VertexPNCT	v;

	while (lon<=max_lon)
	{
		lon_count++;

		float s=sinf(deg_to_rad(lon));
		float c=cosf(deg_to_rad(lon));

		v.pos=center+dir_x*c*0.5f+dir_z*s*0.5f;
		v.normal=appex.normal;
		v.color=color;
		v.uv.set(s*su,c*sv);
		dest->push_vertex(v);

		lon+=lon_inc;
	}

	int lx=lon_count;

	for (int xx=0; xx<lx-1; xx++)
	{
		dest->push_index(base_index); dest->push_index(base_index+xx+1); dest->push_index(base_index+xx+2); 
	}

	return dest;
}

GeomMesh* geom_add_cylinder(GeomMesh* src_mesh,
							const Vec3f& bottom_center,const Vec3f& axis_x,const Vec3f& dir_y,const Vec3f& axis_z,
							const float bottom_radius,const float top_radius,
							const float y_div,
							const float min_lon,const float max_lon,const float lon_inc,
							const unsigned long color,const Rect& rect_uv)
{
	GeomMesh* dest=(src_mesh)?(src_mesh):(new GeomMesh(false));

	VertexPNCT	v;
	v.color=color;

	float	c1=(bottom_radius-top_radius);
	float	c2=dir_y.norm_squared();
	float	h=sqrt(c1*c1+c2);
	float	ny=c1/h;
	float	sny=sqrtf(c2)/h;

	float	lat=0.0f;
	int		lat_count=0;
	int		lon_count=0;
	Vec3f	inc_y=dir_y/y_div;
	size_t	base_idx=dest->get_current_vertex_count();

	while (lat<=y_div)
	{
		lat_count++;

		float lon=min_lon;

		lon_count=0;

		float t=(lat/y_div);
		float radius=(1.0f-t)*bottom_radius+t*top_radius;
			
		v.uv.y=rect_uv.y1+(rect_uv.y2-rect_uv.y1)*t;

		while (lon<=max_lon)
		{
			lon_count++;

			float s=sinf(deg_to_rad(lon));
			float c=cosf(deg_to_rad(lon));

			v.pos=radius*c*axis_x+radius*s*axis_z;
			v.normal=v.pos;
			v.normal.normalize();
			v.normal.set(v.normal.x*sny,ny,v.normal.z*sny);
			

			v.pos=v.pos+bottom_center+dir_y*t;
			v.uv.x=rect_uv.x1+(rect_uv.x2-rect_uv.x1)*((lon-min_lon)/(max_lon-min_lon));
			dest->push_vertex(v);

			lon+=lon_inc;
		}

		lat+=1.0f;
	}

	int ly=lat_count;
	int lx=lon_count;

	for (int yy=0; yy<ly-1; yy++)
	{
		for (int xx=0; xx<lx-1; xx++)
		{
			dest->push_index(base_idx+yy*lx+xx); dest->push_index(base_idx+yy*lx+xx+1); dest->push_index(base_idx+(yy+1)*lx+xx+1);
			dest->push_index(base_idx+yy*lx+xx); dest->push_index(base_idx+(yy+1)*lx+xx+1); dest->push_index(base_idx+(yy+1)*lx+xx);
		}
	}

	return dest;
}

size_t BaseMesh::get_triangle_count()
{
	switch (_primitive)
	{
		case D3DPT_POINTLIST:
		case D3DPT_LINELIST:
		case D3DPT_LINESTRIP:
			return 0;
			break;
		case D3DPT_TRIANGLELIST:
			if (_n_index>0) return _n_index/3;
			else return _n_verts/3;
			break;
		case D3DPT_TRIANGLESTRIP:
		case D3DPT_TRIANGLEFAN:
			if (_n_index>0) return _n_index-2;
			else return _n_verts-2;
			break;
	}
	return 0;
}

