#pragma once

#include "Defs.h"
#include "Gfx.h"
#include "Material.h"
#include <vector>
#include <string>

class Camera;
class Light;

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// Vertex defs
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
struct VertexPN
{
	Vec3f	pos;
	Vec3f	normal;

	static DWORD get_fvf() { return D3DFVF_XYZ|D3DFVF_NORMAL; }
};

struct VertexPC
{
	Vec3f			pos;
	unsigned long	color;

	static DWORD get_fvf() { return D3DFVF_XYZ|D3DFVF_DIFFUSE; }
};

struct VertexPSC
{
	Vec3f			pos;
	float			s;
	unsigned long	color;

	static DWORD get_fvf() { return D3DFVF_XYZ|D3DFVF_PSIZE|D3DFVF_DIFFUSE; }
};

struct VertexPNT
{
	Vec3f	pos;
	Vec3f	normal;
	Vec2f	uv;

	static DWORD get_fvf() { return D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_TEX1; }
};

struct VertexPNCT
{
	Vec3f			pos;
	Vec3f			normal;
	unsigned long	color;
	Vec2f			uv;

	static DWORD get_fvf() { return D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_DIFFUSE|D3DFVF_TEX1; }
};

struct VertexPNCCT
{
	Vec3f			pos;
	Vec3f			normal;
	unsigned long	color1;
	unsigned long	color2;
	Vec2f			uv;

	static DWORD get_fvf() { return D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_DIFFUSE|D3DFVF_SPECULAR|D3DFVF_TEX1; }
};

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
class VertexBuffer
{
public:
	VertexBuffer()
	{
#ifdef USE_D3D9
		_vb=NULL;
		_dynamic=false;
		_current_iterator=NULL;
#endif
	}
	~VertexBuffer()
	{
#ifdef USE_D3D9
		if (_current_iterator)
		{
			_vb->Unlock();
			_current_iterator=NULL;
		}
		if (_vb)
		{
			_vb->Release();
			_vb=NULL;
		}
#endif
	}
/*
	inline	void	set(const UINT n)
	{
#ifdef USE_D3D9
		GFX->get_device()->SetStreamSource(n,_vb,0,_vertex_size);
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
	}

	inline	void	unset(const UINT n)
	{
#ifdef USE_D3D9
		GFX->get_device()->SetStreamSource(n,NULL,0,0);
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
	}
	*/
	inline	bool	create(const UINT n_vertex,const int vertex_size,const DWORD fvf,const bool dynamic)
	{
#ifdef USE_D3D9
		DWORD flags=D3DUSAGE_WRITEONLY;
		if (dynamic) flags|=D3DUSAGE_DYNAMIC;
		if (GFX->_device->CreateVertexBuffer(n_vertex*vertex_size,flags,fvf,D3DPOOL_DEFAULT,&_vb,NULL)!=D3D_OK)
		{
			_vb=NULL;
			return false;
		}
		_dynamic=dynamic;
		_vertex_size=vertex_size;
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
		return true;
	}

	inline	void*	get_write_pointer()
	{
#ifdef USE_D3D9
		if (_current_iterator) return _current_iterator;

		DWORD flags=0;
		if (_dynamic) flags|=D3DLOCK_DISCARD;

		void* ptr;
		if (_vb->Lock(0,0,(void**)&ptr,flags)!=D3D_OK)
		{
			return NULL;
		}
		_current_iterator=ptr;
		return _current_iterator;
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
		return NULL;
	}

	inline	void	release_write_pointer()
	{
#ifdef USE_D3D9
		_vb->Unlock();
		_current_iterator=NULL;
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
	}

#ifdef USE_D3D9
	LPDIRECT3DVERTEXBUFFER9	_vb;
	int						_vertex_size;
	void*					_current_iterator;
	bool					_dynamic;
#endif
};

//////////////////////////////////////////////////////////////////////////////////////////

class IndexBuffer
{
public:
	IndexBuffer()
	{
		_dynamic=false;
		_current_iterator=NULL;
		_ib=NULL;
	}
	~IndexBuffer()
	{
		if (_current_iterator)
		{
			_ib->Unlock();
			_current_iterator=NULL;
		}
		if (_ib)
		{
			_ib->Release();
			_ib=NULL;
		}
	}
/*
	inline	void	set()
	{
#ifdef USE_D3D9
		GFX->get_device()->SetIndices(_ib);
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
	}

	inline	void	unset()
	{
#ifdef USE_D3D9
		GFX->get_device()->SetIndices(NULL);
#elif defined(USE_OGL31)
		__asm int 3
#else
#error "Undefined platform!"
#endif
	}*/

	inline	bool	create(const UINT n,const bool dynamic)
	{
		DWORD flags=D3DUSAGE_WRITEONLY;
		if (dynamic) flags|=D3DUSAGE_DYNAMIC;

		if (GFX->_device->CreateIndexBuffer(n*sizeof(WORD),flags,D3DFMT_INDEX16,D3DPOOL_DEFAULT,&_ib,NULL)!=D3D_OK)
		{
			return false;
		}
		_dynamic=dynamic;
		return true;
	}

	inline	WORD*	get_write_pointer()
	{
		if (_current_iterator) return _current_iterator;

		DWORD flags=0;
		if (_dynamic) flags|=D3DLOCK_DISCARD;

		WORD* ptr;
		if (_ib->Lock(0,0,(void**)&ptr,flags)!=D3D_OK)
		{
			return NULL;
		}
		_current_iterator=ptr;

		return _current_iterator;
	}

	inline	void	release_write_pointer()
	{
		_ib->Unlock();
		_current_iterator=NULL;
	}

#ifdef USE_D3D9
	LPDIRECT3DINDEXBUFFER9	_ib;
	bool					_dynamic;
	WORD*					_current_iterator;
#endif

};

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// Base mesh (material, draw functions)
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
class BaseMesh
{
public:
	BaseMesh(const bool dynamic);
	virtual ~BaseMesh();

	virtual void		draw(Material* override_material=NULL) = 0;

			bool		alloc_index(const UINT n);
			WORD*		get_write_pointer_i();
			void		release_write_pointer_i();

	inline	void		set_material(Material* material,const bool bound) { _material=material; _bound_material=bound; }
	inline	Material*	get_material() { return _material; }

	inline	std::string	get_name() { return _name; }
	inline	void		set_name(const std::string& name) { _name=name; }
	inline	UINT		get_vertex_count() { return _n_verts; }
	inline	UINT		get_index_count() { return _n_index; }
	inline	void		push_index(WORD idx) { _index_array.push_back(idx); }

	inline	void		set_primitive(D3DPRIMITIVETYPE prim) { _primitive=prim; }
	inline	void		set_render_verts(UINT n) { _render_verts=n; }
	inline	UINT		get_render_verts() { return _render_verts; }

	inline	void		get_aabb(Vec3f& pmin,Vec3f& pmax) { pmin=_pmin; pmax=_pmax; }
			
			size_t		get_triangle_count();
	virtual void		get_triangle(const int index,Vec3f* pos) = 0;

protected:
	std::string				_name;
	bool					_dynamic;
	VertexBuffer*			_vb;
	IndexBuffer*			_ib;
	UINT					_n_verts,_n_index;
	UINT					_render_verts;
	WORD*					_current_iterator_i;
	WORD*					_iterator_start_i;
	Material*				_material;
	bool					_bound_material;
	std::vector<WORD>		_index_array;
	D3DPRIMITIVETYPE		_primitive;
	Vec3f					_pmin,_pmax;
	bool					_keep_data;
};

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// Mesh (accessor methods, etc)
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
template<class T>
class Mesh : public BaseMesh
{
public:
	Mesh(const bool dynamic);
	virtual ~Mesh(void);

	virtual	void	draw(Material* override_material=NULL);

			bool	alloc_vert(const UINT n);
			T*		get_write_pointer_v();
			void	release_write_pointer_v();

			void	load(std::vector<T>* vertex,std::vector<WORD>* index)
			{
				alloc_vert((int)vertex->size());
				T* ptr_v=get_write_pointer_v();
				memcpy(ptr_v,&(vertex->at(0)),sizeof(T)*vertex->size());

				alloc_index((int)index->size());
				WORD* ptr_i=get_write_pointer_i();
				memcpy(ptr_i,&(index->at(0)),index->size()*sizeof(WORD));
			}

			size_t	get_current_vertex_count()
			{
				return _vertex_array.size();
			}

			void	push_vertex(const T& v)
			{
				_vertex_array.push_back(v);
			}

			void close(const bool clear_data)
			{
				if ((_vertex_array.size()>0) && (_index_array.size()>0))
				{
					load(&_vertex_array,&_index_array);
					if (clear_data)
					{
						_vertex_array.clear();
						_index_array.clear();
					}
				}
			}
			void	compute_aabb();
	virtual	void	get_triangle(const int index,Vec3f* pos);

protected:
	T*				_current_iterator_v;
	T*				_iterator_start_v;
	std::vector<T>	_vertex_array;
};

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// Helper functions
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
BaseMesh*	build_cube_pn(const float sx,const float sy,const float sz);
BaseMesh*	build_cube_pnt(const float sx,const float sy,const float sz,long tex_id);
bool		load_3ds(const std::string& filename,
					 std::vector<BaseMesh*>* meshes,
					 std::vector<Camera*>* cameras,
					 std::vector<Light*>* lights);

void		build_normals_pnt(Mesh<VertexPNT>* mesh);

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// Templated functions
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
template<class T>
Mesh<T>::Mesh(const bool dynamic) : BaseMesh(dynamic)
{
	_current_iterator_v=NULL;
	_iterator_start_v=NULL;
}

template<class T>
Mesh<T>::~Mesh(void)
{
	if (_current_iterator_v) release_write_pointer_v();
	_iterator_start_v=NULL;
}

template<class T>
bool Mesh<T>::alloc_vert(const UINT n)
{
	if (_dynamic)
	{
		if (n<=_n_verts) return true;
	}

	if (_vb) delete _vb;
	_n_verts=0;
	_render_verts=0;

	long fvf=T::get_fvf();

	_vb=new VertexBuffer;
	if (!_vb->create(n,sizeof(T),fvf,_dynamic))
	{
		delete _vb;
		_vb=NULL;
		return false;
	}

	_n_verts=n;
	_render_verts=n;

	return true;
}

template<class T>
T* Mesh<T>::get_write_pointer_v()
{
	if (_current_iterator_v) return _current_iterator_v;

	_current_iterator_v=(T*)_vb->get_write_pointer();
	_iterator_start_v=_current_iterator_v;
	return _current_iterator_v;	
}

template<class T>
void Mesh<T>::release_write_pointer_v()
{
	if (!_current_iterator_v) return;

	compute_aabb();

	if (_keep_data)
	{
		_vertex_array.resize(_n_verts);
		memcpy(&_vertex_array[0],_iterator_start_v,_n_verts*sizeof(T));
	}

	_vb->release_write_pointer();

	_current_iterator_v=NULL;
	_iterator_start_v=NULL;
}

template<class T>
void Mesh<T>::draw(Material* override_material)
{
	if (_current_iterator_v) release_write_pointer_v();
	if (_current_iterator_i) release_write_pointer_i();

	if (_ib)
	{
		if (_n_index==0) return;
	}
	else
	{
		if (_render_verts==0) return;
	}

	// Setup material
	long fvf=T::get_fvf();

	Material* mat=(override_material)?(override_material):(_material);
	if (mat) 
	{
		if (!mat->pass_enabled(GFX->get_pass())) return;
		mat->set(fvf);
	}

	GFX->render(_primitive,_vb,_render_verts,_ib,_n_index);

	if (mat) mat->unset();
}

template<class T>
void Mesh<T>::compute_aabb()
{
	_pmin.set(FLT_MAX,FLT_MAX,FLT_MAX);
	_pmax.set(-FLT_MAX,-FLT_MAX,-FLT_MAX);

	T* ptr=_iterator_start_v;
	for (UINT i=0; i<_n_verts; i++)
	{
		_pmin.x=Min(_pmin.x,((Vec3f*)ptr)->x);
		_pmin.y=Min(_pmin.y,((Vec3f*)ptr)->y);
		_pmin.z=Min(_pmin.z,((Vec3f*)ptr)->z);
		_pmax.x=Max(_pmax.x,((Vec3f*)ptr)->x);
		_pmax.y=Max(_pmax.y,((Vec3f*)ptr)->y);
		_pmax.z=Max(_pmax.z,((Vec3f*)ptr)->z);
		ptr++;
	}
}

template<class T>
void Mesh<T>::get_triangle(const int index,Vec3f* pos)
{
	int i[3];

	switch (_primitive)
	{
		case D3DPT_POINTLIST:
		case D3DPT_LINELIST:
		case D3DPT_LINESTRIP:
			return;
			break;
		case D3DPT_TRIANGLELIST:
			if (_current_iterator_i)
			{
				if (_n_index>0)
				{
					i[0]=_iterator_start_i[index*3];
					i[1]=_iterator_start_i[index*3+1];
					i[2]=_iterator_start_i[index*3+2];
				}
				else
				{
					i[0]=index*3;
					i[1]=index*3+1;
					i[2]=index*3+2;
				}
			}
			else
			{
				if (_n_index>0)
				{
					i[0]=_index_array[index*3];
					i[1]=_index_array[index*3+1];
					i[2]=_index_array[index*3+2];
				}
				else
				{
					i[0]=index*3;
					i[1]=index*3+1;
					i[2]=index*3+2;
				}
			}
			break;
		case D3DPT_TRIANGLESTRIP:
		case D3DPT_TRIANGLEFAN:
			if (_current_iterator_i)
			{
				if (_n_index>0)
				{
					i[0]=_iterator_start_i[index];
					i[1]=_iterator_start_i[index+1];
					i[2]=_iterator_start_i[index+2];
				}
				else
				{
					i[0]=index;
					i[1]=index+1;
					i[2]=index+2;
				}
			}
			else
			{
				if (_n_index>0)
				{
					i[0]=_index_array[index];
					i[1]=_index_array[index+1];
					i[2]=_index_array[index+2];
				}
				else
				{
					i[0]=index;
					i[1]=index+1;
					i[2]=index+2;
				}
			}
			break;
	}

	if (_current_iterator_v)
	{
		pos[0]=_iterator_start_v[i[0]].pos;
		pos[1]=_iterator_start_v[i[1]].pos;
		pos[2]=_iterator_start_v[i[2]].pos;
	}
	else
	{
		pos[0]=_vertex_array[i[0]].pos;
		pos[1]=_vertex_array[i[1]].pos;
		pos[2]=_vertex_array[i[2]].pos;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// Geometry building "library"
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef Mesh<VertexPNCT> GeomMesh;
typedef Mesh<VertexPNCCT> GeomMeshE;

const	int	FACE_PX=1<<1;
const	int	FACE_MX=1<<2;
const	int	FACE_PY=1<<3;
const	int	FACE_MY=1<<4;
const	int	FACE_PZ=1<<5;
const	int	FACE_MZ=1<<6;

GeomMesh* geom_add_quad(GeomMesh* src_mesh,
						const Vec3f& center,const Vec3f& dir0,const Vec3f& dir1,
						const unsigned long color,const Rect& uv_rect);
GeomMesh* geom_add_quad(VertexPNCT*& v_ptr,WORD*& i_ptr,int& vertex_count,int& index_count,
						const Vec3f& center,const Vec3f& dir0,const Vec3f& dir1,
						const unsigned long color,const Rect& uv_rect);
GeomMesh* geom_add_quad(VertexPNCCT*& v_ptr,WORD*& i_ptr,int& vertex_count,int& index_count,
						const Vec3f& center,const Vec3f& dir0,const Vec3f& dir1,
						const unsigned long color,const unsigned long color2,const Rect& uv_rect);
GeomMesh* geom_add_cube(GeomMesh* src_mesh,
						const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
						const unsigned long color,const unsigned char faces,Rect* uv);
GeomMesh* geom_add_cube(VertexPNCT*& v_ptr,WORD*& i_ptr,int& vertex_count,int& index_count,
						const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
						const unsigned long color,const unsigned char faces,Rect* uv);
GeomMesh* geom_add_cube(VertexPNCCT*& v_ptr,WORD*& i_ptr,int& vertex_count,int& index_count,
						const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
						const unsigned long color,const unsigned long color2,const unsigned char faces,Rect* uv);
GeomMesh* geom_add_spheroid(GeomMesh* src_mesh,
							const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
							const float min_lat,const float max_lat,const float lat_inc,
							const float min_lon,const float max_lon,const float lon_inc,
							const unsigned long color,const Rect& rect_uv);
GeomMesh* geom_add_spheroid(GeomMesh* src_mesh,
							const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
							const float min_lat,const float max_lat,const float lat_inc,
							const float min_lon,const float max_lon,const float lon_inc,
							Interpolator<Color>* color,const Rect& rect_uv);
GeomMesh* geom_add_cap(GeomMesh* src_mesh,
					   const Vec3f& center,const Vec3f& dir_x,const Vec3f& dir_y,const Vec3f& dir_z,
					   const float min_lon,const float max_lon,const float lon_inc,
					   const unsigned long color,const Rect& rect_uv);
GeomMesh* geom_add_cylinder(GeomMesh* src_mesh,
							const Vec3f& bottom_center,const Vec3f& axis_x,const Vec3f& dir_y,const Vec3f& axis_z,
							const float bottom_radius,const float top_radius,
							const float y_div,
							const float min_lon,const float max_lon,const float lon_inc,
							const unsigned long color,const Rect& rect_uv);
