// ModelManager.h: interface for the TextureManager class.
//
//////////////////////////////////////////////////////////////////////

#ifndef __MODELMANAGER_H
#define __MODELMANAGER_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "MathLib.h"
#include "Mesh.h"

class Image_A8R8G8B8;
class VoxelWorld;
class Camera;

class ModelManager  
{
public:
	typedef enum { EXTRUSION_DIR_PX=0, EXTRUSION_DIR_PY=1, EXTRUSION_DIR_PZ=2,
				   EXTRUSION_DIR_NX=3, EXTRUSION_DIR_NY=4, EXTRUSION_DIR_NZ=5 } ExtrusionDir;
	typedef enum { GEOMOP_CUBE=0 } GeomOpType;
protected:
	struct VoxelSlice
	{
		long		h1;
		long		h2;
		long		ox,oy;
		std::string	color_image;
		std::string	height_image;
	};
	typedef std::vector<VoxelSlice>	VoxelSlices;
	struct VoxelBuild
	{
		VoxelBuild() { _size.set(0.0f,0.0f,0.0f); _dir=EXTRUSION_DIR_PY; _texture_id=-1; _ix=_iy=0; for (int i=0; i<256; i++) _material_scale[i].set(1.0f,1.0f,1.0f); }
		Vec3f			_size;
		ExtrusionDir	_dir;
		long			_texture_id;
		long			_ix,_iy;
		Vec3f			_material_scale[256];
		VoxelSlices		_slices;
	};
	struct GeomOp
	{
		GeomOp(GeomOpType t) { _type=t; _size.set(0.0f,0.0f,0.0f); }
		GeomOpType		_type;
		Vec3f			_size;
	};
	struct GeomBuild
	{
		GeomBuild() { _texture_id=-1; }
		long				_texture_id;
		std::vector<GeomOp>	_ops;
	};
	struct Model
	{
		Model() { _voxel_model=NULL; _voxel_build=NULL; _geom_model=NULL; _geom_build=NULL; _mesh_model=NULL; }
		char*		_name;
		VoxelWorld*	_voxel_model;
		VoxelBuild*	_voxel_build;
		GeomMesh*	_geom_model;
		GeomBuild*	_geom_build;
		BaseMesh*	_mesh_model;
		bool		_set_aabb;
		Vec3f		_pmax,_pmin;
	};
	typedef std::vector<Model*>	Models;
public:
	ModelManager();
	virtual ~ModelManager();

	void				clear(const char* name);
	void				clear(long id);
	void				clear_all();
	long				new_voxel_model(const char *name);
	long				new_build_geom_model(const char *name);
	long				load_3ds(const char *name,const char* filename);
	long				add_model(const char* name,BaseMesh* mesh);
	long				find_by_name(const char *name);

	void				draw(const long model_id,const Vec3f& pos,const Quaternion& ori,const Vec3f& scale,Camera* camera,Material* override_material=NULL);
	void				model_build(const long model_id);

	void				voxel_size(const long voxel_model_id,const float sx,const float sy,const float sz);
	void				voxel_slice(const long voxel_model_id,const int offset_x,const int offset_y,const long h1,const long h2,const char* color_image,const char* heightmap);
	void				voxel_materials(const long voxel_model_id,const long texture_id,const long ix,const long iy);
	void				voxel_materials_scale(const long voxel_model_id,const long material_id,const float sx,const float sy,const float sz);
	void				voxel_extrusion_dir(const long voxel_model_id,const ExtrusionDir dir);

	long				geom_push_op(const long geom_model_id,const GeomOpType op_type);
	void				geom_set_size(const long geom_model_id,const long op_id,const float sx,const float sy,const float sz);
	void				geom_set_texture(const long geom_model_id,const long texture_id);

	bool				get_aabb(const long model_id,Vec3f& pmin,Vec3f& pmax);
	void				set_aabb(const long model_id,const Vec3f& pmin,const Vec3f& pmax);

	BaseMesh*			get_mesh(const long model_id);

protected:
	Model*				get_model(const long model_id);
protected:
	Models				_models;
};

extern ModelManager* MODEL_MANAGER;

#endif // !defined(AFX_TEXTUREMANAGER_H__F26661F1_5CBF_43D8_AD3B_4333CD88A491__INCLUDED_)
