#pragma once

#include "Defs.h"
#include <recast.h>
#include <DetourNavMesh.h>
#include <DetourNavMeshBuilder.h>
#include <DetourNavMeshQuery.h>

#ifdef _DEBUG
#pragma comment(lib, "Detour_d.lib")
#pragma comment(lib, "Recast_d.lib")
#else
#pragma comment(lib, "Detour.lib")
#pragma comment(lib, "Recast.lib")
#endif

class Tilemap3d;
class ModelObj3d;
class NavPath;

class NavMesh
{
public:
	NavMesh();
	virtual ~NavMesh(void);

	void	add_tilemap3d(Tilemap3d* tilemap);
	void	add_modelobj3d(ModelObj3d* model_obj);

	void	draw();
	void	init();

	inline	void	set_max_vertex_per_poly(const int v) { _max_vertex_per_poly=v; }
	inline	void	set_agent_radius(const float v) { _agent_radius=v; }
	inline	void	set_agent_height(const float v) { _agent_height=v; }
	inline	void	set_agent_step(const float v) { _agent_step=v; }
	inline	void	set_cellsize(const float v) { _cellsize=v; }
	inline	void	set_cellheight(const float v) { _cellheight=v; }

			NavPath*	get_path(const Vec3f& pos1,const Vec3f& pos2,NavPath* input_path,dtQueryFilter* input_query_filter=NULL);

protected:
	void	build_navmesh();
	void	build_detour_data();

protected:
	bool				_init;
	int					_max_vertex_per_poly;
	float				_agent_radius;
	float				_agent_height;
	float				_agent_step;
	float				_cellsize;
	float				_cellheight;
	Vec3f				_pmin,_pmax;
	std::vector<Vec3f>	_verts;
	std::vector<int>	_tris;
	rcPolyMesh*			_pmesh;
	rcPolyMeshDetail*	_dmesh;
	dtNavMesh*			_nav_mesh;
	dtNavMeshQuery*		_query_object;
	dtQueryFilter*		_default_query_filter;
};
