
#include "RndGenerator.h"

RndGenerator	g_rnd(76545678);

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RndGenerator::RndGenerator(const long seed)
{
	set_seed(seed);
}

RndGenerator::~RndGenerator()
{
}

void RndGenerator::set_seed(const long seed)
{
	// Re-seed generator
	_mt[0]=seed;
	for (_mti=1; _mti<MERS_N; _mti++) 
	{
		_mt[_mti]=(1812433253UL*(_mt[_mti-1] ^ (_mt[_mti-1] >> 30)) + _mti);
	}
}

unsigned long RndGenerator::get_random()
{  
	unsigned long	y;

	if (_mti>=MERS_N) 
	{
		// generate MERS_N words at one time
		const unsigned long LOWER_MASK = (1LU << MERS_R)-1; // lower MERS_R bits
		const unsigned long UPPER_MASK = -1L  << MERS_R;      // upper (32 - MERS_R) bits
		static const unsigned long mag01[2] = {0, MERS_A};

		int kk;
		for (kk=0; kk<MERS_N-MERS_M; kk++) 
		{    
			y=(_mt[kk] & UPPER_MASK) | (_mt[kk+1] & LOWER_MASK);
			_mt[kk]=_mt[kk+MERS_M] ^ (y >> 1) ^ mag01[y & 1];
		}

		for (; kk<MERS_N-1; kk++) 
		{    
			y=(_mt[kk] & UPPER_MASK) | (_mt[kk+1] & LOWER_MASK);
			_mt[kk]=_mt[kk+(MERS_M-MERS_N)] ^ (y >> 1) ^ mag01[y & 1];
		}      

		y=(_mt[MERS_N-1] & UPPER_MASK) | (_mt[0] & LOWER_MASK);
		_mt[MERS_N-1]=_mt[MERS_M-1] ^ (y >> 1) ^ mag01[y & 1];
		_mti=0;
	}

	y=_mt[_mti++];

	return y;
}

long RndGenerator::get_random_i(const long min,const long max)
{
	if (min==max) return min;
	long n=long(get_random()%(max-min+1))+min;

	return n;
}

float RndGenerator::get_random_f(const float min,const float max)
{
	float f=(float(get_random())/float(0x00FFFFFFFF))*(max-min)+min;
	
	return f;
}
