// Screen.h: interface for the Screen class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SCREEN_H__676B95DB_390F_4F6B_9405_CB63985E87C3__INCLUDED_)
#define AFX_SCREEN_H__676B95DB_390F_4F6B_9405_CB63985E87C3__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Drawable.h"

class Font;

class PrioritySort
{
public:
	bool operator()(Drawable* obj1,Drawable* obj2)
	{
		if (!obj1) return false;
		if (!obj2) return true;
		if (obj1->get_priority()>obj2->get_priority())
		{
			return true;
		}
		return false;
	}
};

class Screen  
{
protected:
	struct RemoveElem
	{
		Drawable*	drawable;
		bool		should_delete;
	};
	typedef std::vector<RemoveElem>	Removes;
	typedef std::vector<Drawable*>	DrawableVector;
public:
	Screen();
	virtual ~Screen();

	virtual	long	run();
			void	run(float t);
	virtual void	render() = 0;
	virtual void	anim(const float t) = 0;
			
	inline	void	set_exit(bool b,long exit_code) { _exit=b; _exit_code=exit_code; }

	inline	void	add_drawable(Drawable* drawable) { _drawables.push_back(drawable); }
			void	remove_all_drawables(const bool should_delete);
	inline	void	remove_drawable(Drawable* drawable,const bool should_delete=false) 
					{ 
						RemoveElem elem;
						elem.drawable=drawable;
						elem.should_delete=should_delete;
						_removes.push_back(elem);
					}
			void	remove_drawables();

			bool	create_font();
			void	delete_font();

	virtual	void	mouse_callback(unsigned long msg,unsigned long wparam,unsigned long lparam);
	inline	bool	is_anim_enable() { return _anim_enable; }

public:
	static	void	load_assets(std::string filename);
	inline	Screen*	get_previous_screen() { return _prev_screen; }

protected:
	bool			_exit;
	DrawableVector	_drawables;
	Removes			_removes;
	long			_exit_code;
	UINT			_font_size;
	bool			_anim_enable;
	Vec2f			_desktop_res;
	Screen*			_prev_screen;
#ifdef USE_D3D9
	Font*			_font;
#endif
};

extern Screen*	SCREEN;

#endif // !defined(AFX_SCREEN_H__676B95DB_390F_4F6B_9405_CB63985E87C3__INCLUDED_)
