#include "Sparks.h"
#include "BlitManager.h"
#include "Utils.h"

Sparks::Sparks() : Object3d("sparks",NULL)
{
	Material* mat=new Material;
	mat->set_color_vertex(true,false);
	mat->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);
	mat->set_shader("basic");

	_mesh=new Mesh<VertexPC>(true);
	_mesh->alloc_vert(10000);
	_mesh->set_primitive(D3DPT_LINELIST);
	_mesh->set_render_verts(0);
	_mesh->set_material(mat,true);
}

Sparks::~Sparks(void)
{
	if (_mesh)
	{
		delete _mesh;
		_mesh=NULL;
	}
}

void Sparks::draw()
{
	if (!_draw) return;
	if (!_mesh) return;
	if (_particles.empty()) return;
	
	if (GFX->get_pass()==PASS_SOLID) return;

	GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0.0f,0.0f,0.0f,1.0f));

	_mesh->draw();
}

void Sparks::anim(const float t)
{
	if (_particles.empty()) return;

	VertexPC*	vp=_mesh->get_write_pointer_v();
	auto		it=_particles.begin();
	UINT		n=0;

	while (it!=_particles.end())
	{
		it->life-=t;
		if (it->life<0.0f)
		{
			it=_particles.erase(it);
			continue;
		}
		n++;

		float l=(it->life/it->max_life);

		unsigned long c=((unsigned long)floorf(l*255))<<24;

		// Move particles
		it->old_pos=it->pos;
		it->pos=it->pos+it->speed*t;

		// Update the mesh
		vp->pos=it->pos;
		vp->color=(it->color1&0x00FFFFFF)+c;

		vp++;

		vp->pos=it->old_pos;
		vp->color=(it->color2&0x00FFFFFF)+c;

		vp++;

		++it;
	}
	_mesh->set_render_verts(n*2);
	_mesh->release_write_pointer_v();
}

void Sparks::new_explosion(const Vec3f& pos,const float speed,const long n_particles,const float life,const unsigned long color1,const unsigned long color2)
{
	for (int i=0; i<n_particles; i++)
	{
		Particle p;
		p.old_pos=p.pos=Vec3f(pos.x,pos.y,pos.z-0.2f);
		p.life=life;
		p.max_life=life;
		p.color1=color1;
		p.color2=color2;
		p.speed.set(((float)(rand()%200)-100.0f)/100.0f,
					((float)(rand()%200)-100.0f)/100.0f,
					((float)(rand()%200)-100.0f)/100.0f);
		p.speed.normalize(); p.speed=p.speed*((float)(rand()%100)+100.0f)*speed;

		_particles.push_back(p);
	}
}
