#include "Sprite3d.h"
#include "Gfx.h"
#include "BlitManager3d.h"

Sprite3d::Sprite3d(const char* type,const char* name) : Object3d(type,name)
{
	_color.set(1.0f,1.0f,1.0f,1.0f);
	_size.set(10.0f,10.0f);
	_current_image=-1;
	_frame=-1;
	_frame_uv.set(0.0f,0.0f,1.0f,1.0f);
	_ground=false;
	_perp=false;
	_hotspot.set(0.0f,0.0f,0.0f);
	_current_anim=NULL;
	_next_anim="";
	_blit_mode=BLIT_NORMAL;
	_billboard=true;
}

Sprite3d::~Sprite3d(void)
{
	if (_current_anim)
	{
		delete _current_anim;
		_current_anim=NULL;
	}
}

void Sprite3d::draw()
{
	if (!_draw) return;
	if ((_size.x==0.0f) || (_size.y==0.0f)) return;
	if (_current_image==-1) return;

	Vec3f	color=GFX->get_color_modulation();
	color.x*=_color.x;
	color.y*=_color.y;
	color.z*=_color.z;
	Color	c(color.x,color.y,color.z,_color.a);

	if ((GFX->get_pass()==PASS_SOLID) && (_color.a<1.0f)) return;
	if ((GFX->get_pass()==PASS_ALPHA) && (_color.a==1.0f)) return;

	if (!_ground)
	{
		if (!_perp)
		{
			if (_billboard) BLIT3D->add(_current_image,_pos+_hotspot,_size,c,_frame_uv,_blit_mode);
			else BLIT3D->add(_current_image,_pos+_hotspot,Vec3f(0.0f,_size.y,0.0f),Vec3f(_size.x,0.0f,0.0f),Vec3f(0.0f,0.0f,-1.0f),c,_frame_uv,_blit_mode);
		}
		else
		{
			BLIT3D->add_perp(_current_image,_pos+_hotspot,_size,c,_frame_uv,_blit_mode);
		}
	}
	else
		BLIT3D->add_ground(_current_image,_pos+_hotspot,_size,c,_frame_uv,_blit_mode);
}

void Sprite3d::set_frame(const long frame)
{
	_frame=frame;
	if (_current_image!=-1) _frame_uv=TEXTURES->get_uv(_current_image,_frame);
}

void Sprite3d::set_image(long img) 
{ 
	_current_image=img; 
	if (_frame!=-1) set_frame(_frame);
}

void Sprite3d::anim(const float t)
{
	if (_current_anim)
	{
		_current_anim->anim(t);
		if (_current_anim->is_ended())
		{
			delete _current_anim;
			_current_anim=NULL;
			if (_next_anim!="")
			{
				play_animation(_next_anim.c_str(),NULL);
			}
		}
	}
}

void Sprite3d::play_animation(const char* anim_name,const bool force_start,const char* next_anim)
{
	if (_current_image==-1) return;
	
	TextureManager::Anim* anim=TEXTURES->fetch_anim(_current_image,TEXTURES->get_animation_id(_current_image,anim_name));
	if (anim)
	{
		if (_current_anim)
		{
			if (!force_start)
			{
				if (!_stricmp(_current_anim->get_name(),anim_name)) return;
			}

			delete _current_anim;
		}

		_current_anim=new SpriteAnim3d(this,anim);

		if (next_anim) _next_anim=next_anim;
		else _next_anim="";
	}
	else
	{
		if (next_anim)
		{
			play_animation(next_anim,force_start);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

SpriteAnim3d::SpriteAnim3d(Sprite3d* spr,TextureManager::Anim* anim)
{
	_sprite=spr;
	_anim=anim;
	_frame_time=_anim->time;
	_frame_index=0;

	_sprite->set_frame(_anim->frames[_frame_index]);
}

SpriteAnim3d::~SpriteAnim3d()
{

}

void SpriteAnim3d::anim(const float t)
{
	if (_frame_index<0) return;

	_frame_time-=t;
	if (_frame_time<=0.0f)
	{
		_frame_time+=_anim->time;

		_frame_index++;
		if (_frame_index>=(long)_anim->frames.size())
		{
			if (_anim->loop) _frame_index=0;
			else
			{
				_frame_index=-1;
				return;
			}
		}
		_sprite->set_frame(_anim->frames[_frame_index]);
	}
}

