// TextureManager.h: interface for the TextureManager class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TEXTUREMANAGER_H__F26661F1_5CBF_43D8_AD3B_4333CD88A491__INCLUDED_)
#define AFX_TEXTUREMANAGER_H__F26661F1_5CBF_43D8_AD3B_4333CD88A491__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <vector>
#include "MathLib.h"
#include "Defs.h"

class Image_A8R8G8B8;

class TextureManager  
{
public:
	struct Frame
	{
		char*	name;
		Rect	uv;
	};
	typedef std::vector<Frame>	Frames;
	struct Anim
	{
		char*				name;
		float				time;
		bool				loop;
		std::vector<long>	frames;
	};
	typedef std::vector<Anim>	Anims;
	struct Texture
	{
		TextureID	texture;
		char*		filename;
		char*		name;
		int			tx,ty;
		int			sx,sy;
		bool		wrap;
		Frames		frames;
		Anims		anims;
	};
	typedef std::vector<Texture*>	Textures;
public:
	TextureManager();
	virtual ~TextureManager();

	void				clear(const char* name);
	void				clear(long id);
	void				clear_all();
	void				remove(long id);
	long				load(const char *name,const char *filename);
	long				load(const char *name,Image_A8R8G8B8* img);
	long				add(const char *name,LPDIRECT3DTEXTURE9 texture);
	long				find_by_name(const char *name);
	TextureID			get_texture(long id);
	int					get_size_x(long id);
	int					get_size_y(long id);
	int					get_texture_size_x(long id);
	int					get_texture_size_y(long id);
	std::string			get_texture_name(long id);
	bool				get_wrap(const long id);
	void				set_wrap(const long id,const bool b);

	void				add_frame(const long texture_id,const char* name,const float u1,const float v1,const float u2,const float v2);
	Rect				get_uv(const long texture_id,const char* name);
	Rect				get_uv(const long texture_id,const long frame_id);
	long				get_frame_id(const long texture_id,const char* name);

	long				add_animation(const long texture_id,const char* name);
	long				get_animation_id(const long texture_id,const char* name);
	void				set_animation_time(const long texture_id,const long anim_id,const float time);
	void				set_animation_loop(const long texture_id,const long anim_id,const bool loop);
	void				add_animation_frame(const long texture_id,const long anim_id,const long frame_id);
	void				add_animation_frame(const long texture_id,const long anim_id,const char* frame_name);

	Anim*				fetch_anim(const long texture_id,const long anim_id);

protected:
	Texture*			fetch_texture(const long texture_id);
	void*				lock_texture(TextureID texture,int& pitch);
	void				unlock_texture(TextureID texture);
protected:
	Textures			_textures;
};

extern TextureManager* TEXTURES;

#endif // !defined(AFX_TEXTUREMANAGER_H__F26661F1_5CBF_43D8_AD3B_4333CD88A491__INCLUDED_)
