// StarmapScreen.cpp: implementation of the StarmapScreen class.
//
//////////////////////////////////////////////////////////////////////

#include "UIButton.h"
#include "TextureManager.h"
#include "BlitManager.h"
#include "Launcher.h"
#include "Sound.h"
#include "MouseCursor.h"

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

UIButton::UIButton(const char* img_name,const Vec3f& pos,const std::string& tooltip,const std::string& callback_id,const std::string& snd_hover) : Drawable("uibutton",NULL)
{
	_img=TEXTURES->find_by_name(img_name);
	_pos=pos;
	_size.set((float)TEXTURES->get_size_x(_img),(float)TEXTURES->get_size_y(_img));
	_color.set(1.0f,1.0f,1.0f,0.8f);
	_callback_id=callback_id;
	_tooltip=tooltip;
	_available=true;
	_was_inside=false;
	_snd_hover=snd_hover;
	_background_color.set(0.0f,0.0f,0.0f,0.0f);
	_foreground_color.set(0.0f,0.0f,0.0f,0.0f);
}

UIButton::~UIButton()
{

}

void UIButton::draw()
{
	if (!_draw) return;

	if (_background_color.a>0.0f)
	{
		BLIT->add(_pos.x,_pos.y,_pos.z+0.02f,_size.x,_size.y,_background_color);
	}
	BLIT->add(_img,
			  _pos.x,_pos.y,_pos.z,
			  _size.x,_size.y,
			  _color.r,_color.g,_color.b,_color.a,
			  0.0f,0.0f,1.0f,1.0f,
			  D3DTEXF_POINT);
	if (_foreground_color.a>0.0f)
	{
		BLIT->add(_pos.x,_pos.y,_pos.z-0.02f,_size.x,_size.y,_foreground_color);
	}
}

bool UIButton::is_inside()
{
	if (_color.a==1.0f) return true;

	return false;
}

bool UIButton::is_inside(const Vec3f& p)
{
	if ((p.x>_pos.x) && (p.x<(_pos.x+_size.x)) &&
		(p.y>_pos.y) && (p.y<(_pos.y+_size.y)))
	{
		return true;
	}
	return false;
}

void UIButton::anim(const float t)
{
	if (!_draw) return;

	if (_available) _color.a=0.8f;
	else _color.a=0.3f;

	if (CURSOR)
	{
		Vec3f p=CURSOR->_pos;
		if (is_inside(p))
		{
			if (_available) _color.a=1.0f;
			else _color.a=0.4f;

			if (!_was_inside)
			{
				SOUND->play(_snd_hover.c_str(),options_get_fx_volume());
			}
			_was_inside=true;
		}
		else _was_inside=false;
	}
}
