
#pragma once

#include "Mesh.h"

class VoxelWorld
{
protected:
	typedef unsigned char	Voxel;
	struct VoxelCluster
	{
		int			_x,_y,_z;
		GeomMeshE*	_mesh;
	};
	struct FaceCache
	{
		int				_x,_y,_z;
		int				_type;
		unsigned char	_faces;
		unsigned long	_diffuse_color;
	};
public:
	VoxelWorld(void);
	virtual ~VoxelWorld(void);

			void			create(const int sx,const int sy,const int sz,const int cluster_size_x,const int cluster_size_y,const int cluster_size_z,const float voxel_size_x,const float voxel_size_y,const float voxel_size_z);
	inline	void			set_uv_scale(const int i,const Vec3f& uv_scale) { _uv_scale[i].set(1.0f/uv_scale.x,1.0f/uv_scale.y,1.0f/uv_scale.z); }
	inline	void			set_color(const int i,const Color& color) { _color[i]=color; }
	inline	void			set_emissive_color(const int i,const Color& color) { _emissive_color[i]=float_to_colorref(color); }
			void			set_color_mode(const bool color_mode);
	inline	void			set_offset(const Vec3f& offset) { _offset=offset; }
			void			rebuild();
			void			render(Camera* camera);
			void			clear();
			void			set(const int x,const int y,const int z,const Voxel voxel);
			void			set_c(const int x,const int y,const int z,const Voxel voxel,const unsigned long color);
			void			set(const Vec3f& pmin,const Vec3f& pmax,const Voxel voxel);
			unsigned char	get(const int x,const int y,const int z);
			void			set_material_texture(const long texture_id,const long ix,const long iy);
	inline	void			set_material_color(const Color& c) { _material_color=c; }
			void			set_material_alpha_enable(const bool b);
			bool			raycast(const Vec3f& origin,const Vec3f& dir,int& ret_x,int& ret_y,int& ret_z,int& ret_face);
			bool			raycast(const Vec3f& origin,const Vec3f& dir,float& t,const float t_max);

protected:
			void			rebuild_cluster(VoxelCluster* cluster);
			bool			raycast(VoxelCluster* cluster,const Vec3f& origin,const Vec3f& dir,const float t_min,int& ret_x,int& ret_y,int& ret_z,int& ret_face,float& ret_t);
			Rect			get_uv(const unsigned char material,const int face,const Rect& uv_base,const Vec3f& center,const Vec3f& dirx,const Vec3f& diry,const Vec3f& dirz);
			void			get_aabb(VoxelCluster* cluster,Vec3f& pmin,Vec3f& pmax);

	static	bool			aabb_ray_intersect(const Vec3f& pmin,const Vec3f& pmax,const Vec3f& ray_origin,const Vec3f& ray_dir,float& t,int& plane);
public:
	int				_sx,_sy,_sz;
protected:
	int				_cluster_size_x,_cluster_size_y,_cluster_size_z;
	int				_cx,_cy,_cz;
	float			_voxel_size_x;
	float			_voxel_size_y;
	float			_voxel_size_z;
	Material*		_material;
	bool			_color_mode;
	Voxel*			_voxels;
	unsigned long*	_voxels_color;
	VoxelCluster*	_clusters;
	Vec3f			_uv_scale[256];
	Rect			_uv_rect[256];
	Color			_color[256];
	unsigned long	_emissive_color[256];
	Color			_material_color;
	FaceCache*		_face_cache;
	Vec3f			_offset;
public:
	Vec3f			_pos;
	Quaternion		_ori;
};

