#include "LUAMain.h"

// Lua alloc function
 void* sc_lua_alloc(void *ud, void *ptr, size_t osize, size_t nsize) 
{
	(void)ud;     /* not used */
	(void)osize;  /* not used */

	if (nsize==0) 
	{	
		// new size == 0? free memory
		free(ptr);  /* ANSI requires that free(NULL) has no effect */
		return NULL;
	}
	else
	{
		// (re)alloc memory
		/* ANSI requires that realloc(NULL, size) == malloc(size) */
		if (ptr) return realloc(ptr,nsize);
		else return malloc(nsize);
	}
}

LUA* SCLUA=NULL;

LUA::LUA(void)
{
	if (!SCLUA) SCLUA=this;
	_base_env=NULL;
}

LUA::~LUA(void)
{
	if (SCLUA==this) SCLUA=NULL;
}

bool LUA::init()
{
	shutdown();

	_base_env=new LUAEnv;
	
	return true;
}

void LUA::shutdown()
{
	if (_base_env)
	{
		delete(_base_env); _base_env=NULL;
	}
}

LUAEnv*	LUA::get_environment_from_state(lua_State* L)
{
	LUAEnvs::iterator it;
	for(it = _environments.begin(); it != _environments.end(); it++)
	{
		if((*it)->get_env() == L)
			return (*it);
	}

	return NULL;
}

void		LUA::remove_environment(LUAEnv* env)
{
	LUAEnvs::iterator it;
	for(it = _environments.begin(); it != _environments.end(); it++)
	{
		if((*it) == env)
		{
			_environments.erase(it);
			break;
		}
	}
}

void LUA::register_library(const std::string& lib_name,LUA_LIBRARY_FUNCTION func)
{
	_libraries.insert(std::make_pair(lib_name,func));
}

LUA_LIBRARY_FUNCTION LUA::get_library_function(const std::string& lib_name)
{
	if (_libraries.count(lib_name)>0)
	{
		return _libraries[lib_name];
	}
	return NULL;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

std::vector<LUAReference*>			LUAReference::LUAReferences;
std::map<std::string,LUAReference*>	LUAReference::LUANameMap;
std::vector<unsigned long>			LUAReference::LUAFreeIDs;
unsigned long						LUAReference::LUACounter=0;

LUAReference::LUAReference(const char* type,const char* name)
{
	if (name)
	{
		auto it=LUANameMap.find(name);
		if (it==LUANameMap.end()) _lua_name=name;
		else
		{
			char buffer[512];
			sprintf_s((char*)&buffer,512,"%s-%i",name,LUACounter++);
			_lua_name=buffer;
		}
	}
	else
	{
		// Build unique ID
		char buffer[512];
		sprintf_s((char*)&buffer,512,"%s-%08i-%08i-%i",type,timeGetTime(),rand(),LUACounter++);
		_lua_name=buffer;
	}

	if (LUAFreeIDs.empty())
	{
		_lua_id=LUAReferences.size();
		LUAReferences.push_back(this);
	}
	else
	{
		_lua_id=LUAFreeIDs.back();
		LUAFreeIDs.pop_back();
		LUAReferences[_lua_id]=this;
	}

	LUANameMap.insert(std::make_pair(_lua_name,this));
}

LUAReference::~LUAReference()
{
	auto it=LUANameMap.find(_lua_name);
	if (it!=LUANameMap.end()) LUANameMap.erase(it);

	LUAReferences[_lua_id]=NULL;

	LUAFreeIDs.push_back(_lua_id);
}
