
#include "Defs.h"

#include "Gfx.h"
#include "TextureManager.h"
#include "BlitManager.h"
#include "BlitManager3d.h"
#include "Keyboard.h"
#include "Timers.h"
#include "Sound.h"
#include "Utils.h"
#include "Screen.h"
#include "ParticleSystem.h"
#include "Mesh.h"
#include "Camera.h"
#include "Light.h"
#include "Sprite3d.h"
#include "Sprite.h"
#include "Tilemap3d.h"
#include "lua/LUAMain.h"
#include "XML.h"

#include "lua/LUAEnv.h"

bool init_gfx(HINSTANCE hInst)
{
	new Gfx;

	GFX->set_window_mode(true);
	if (!GFX->init(hInst)) return false;
	
	new TextureManager();
	new BlitManager();
	new BlitManager3d();

	return true;
}

bool init_input(HINSTANCE hInst)
{
	new Keyboard;
	
	if (!KEY->init(hInst)) return false;

	return true;
}

bool init_sound()
{
	new Sound;

	if (!SOUND->init()) 
	{
		ERROR_MSG("Couldn't start sound... Swarm will be silent...");
	}

	return true;
}

bool init(HINSTANCE hInst)
{
	if (!init_gfx(hInst)) return false;
	if (!init_input(hInst)) return false;
	if (!init_sound()) return false;

	new Timers;

	return true;
}

void shutdown()
{
	delete SOUND;
	delete TIMERS;
	delete KEY;
	delete BLIT;
	delete TEXTURES;
	delete GFX;
}


class TestScreen : public Screen
{
public:
	TestScreen()
	{
		create_font();

		_camera=new Camera(NULL);
		_camera->_pos=Vec3f(0.0f,50.0f,-50.0f);
		_camera->_ori.rotation_quaternion(Vec3f(0.0f,1.0f,0.0f),0.0f);

		_ang=0.0f;

		_tilemap=new Tilemap3d(NULL);
		
		_tilemap->set_material_color_diffuse(0,0xFFFFFFFF);
		_tilemap->set_material_texture(0,"wall1.dds");
		_tilemap->set_material_texture(0,FACE_TOP,"top1.dds");
		_tilemap->set_material_color_diffuse(1,0xFFFFFFFF);
		_tilemap->set_material_texture(1,"wall2.dds");
		_tilemap->set_material_texture(1,FACE_TOP,"top2.dds");

		_tilemap->create(50,50);
		for (int x=0; x<50; x++) 
		{
			for (int y=0; y<50; y++)
			{
//				_tilemap->set(x,y,0.0f,0);
				_tilemap->set(x,y,(float)(rand()%5),rand()%2);
//				_tilemap->set(x,y,25.0f-((float)x)/2.0f,0);
//				_tilemap->set(x,y,((float)y)/2.0f,0);
//				_tilemap->set(x,y,25.0f-((float)y)/2.0f,0);
			}
		}
/*		_tilemap->set(20,15,5.0f,0);
		_tilemap->set(20,16,-4.0f,0);
		_tilemap->set(20,17,-3.0f,0);
		_tilemap->set(20,18,-2.0f,0);
		_tilemap->set(20,19,-1.0f,0);*/

		_tilemap->set_tile_size(4.0f,4.0f);
		_tilemap->build_geom();

		_tilemap->_pos.set(-100.0f,0.0f,-100.0f);

		_light=new Light(NULL);
		_light->set_type(LIGHT_POINT);
		_light->_pos.set(0.0f,20.0f,0.0f);
		_light->set_range(0.0f,FLT_MAX);
		_light->_ori.rotation_quaternion(Vec3f(1.0f,0.0f,0.0f),30.0f);
		_light->set_color(1.0f,1.0f,1.0f,1.0f);
	}
	virtual ~TestScreen()
	{
		TEXTURES->clear_all();
	}
	virtual void anim(const float t)
	{
		if (KEY->is_pressed(DIK_ESCAPE)) _exit=true;
	}

	virtual void render()
	{ 
		GFX->begin_scene();

		// 3d tests
		GFX->clear(D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,1.0f,0.0f,1.0f,1.0f,1.0f,0);
		GFX->set_ambient_light(Color(0.3f,0.3f,0.3f,1.0f));

		_ang+=0.005f; if (_ang>360.0f) _ang-=360.0f;
		
		_camera->_pos.x=150*sinf(_ang);
		_camera->_pos.z=150*cosf(_ang);
		_camera->set_lookat(_camera->_pos,Vec3f(0.0f,0.0f,0.0f));

		_camera->set();
		GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0,0,0,1));
		GFX->clear_lights();
		GFX->add_light(_light);

		GFX->set_pass(PASS_SOLID);
		_tilemap->draw();

		BLIT3D->flush();

		GFX->clear_lights();

		GFX->end_scene();
	}

public:
	Camera*					_camera;
	float					_ang;
	Tilemap3d*				_tilemap;
	Light*					_light;
};


INT WINAPI WinMain(HINSTANCE hInst,HINSTANCE,LPSTR,INT)
{
	if (!init(hInst)) return -1;

	Screen* tmp=new TestScreen;
	tmp->run();

	delete tmp;

	shutdown();

    return 0;
}
