
#include "Defs.h"

#include "Gfx.h"
#include "TextureManager.h"
#include "BlitManager.h"
#include "BlitManager3d.h"
#include "Keyboard.h"
#include "Timers.h"
#include "Sound.h"
#include "Utils.h"
#include "Screen.h"
#include "ParticleSystem.h"
#include "Mesh.h"
#include "Camera.h"
#include "Light.h"
#include "Sprite3d.h"
#include "Sprite.h"
#include "Tilemap3d.h"
#include "ModelObj3d.h"
#include "lua/LUAMain.h"
#include "XML.h"

#include "lua/LUAEnv.h"

bool init_gfx(HINSTANCE hInst)
{
	new Gfx;

	GFX->set_window_mode(true);
	if (!GFX->init(hInst)) return false;
	
	new TextureManager();
	new BlitManager();
	new BlitManager3d();

	return true;
}

bool init_input(HINSTANCE hInst)
{
	new Keyboard;
	
	if (!KEY->init(hInst)) return false;

	return true;
}

bool init_sound()
{
	new Sound;

	if (!SOUND->init()) 
	{
		ERROR_MSG("Couldn't start sound... Swarm will be silent...");
	}

	return true;
}

bool init(HINSTANCE hInst)
{
	if (!init_gfx(hInst)) return false;
	if (!init_input(hInst)) return false;
	if (!init_sound()) return false;

	new Timers;

	return true;
}

void shutdown()
{
	delete SOUND;
	delete TIMERS;
	delete KEY;
	delete BLIT;
	delete TEXTURES;
	delete GFX;
}


class TestScreen : public Screen
{
public:
	TestScreen()
	{
		create_font();

		_camera=new Camera(NULL);
		_camera->_pos=Vec3f(0.0f,50.0f,-50.0f);
		_camera->_ori.rotation_quaternion(Vec3f(0.0f,1.0f,0.0f),0.0f);

		_ang=0.0f;

		int wood_texture=TEXTURES->load("wood","wall2.dds");

		_tilemap=new Tilemap3d(NULL);
		
		_tilemap->set_material_color_diffuse(0,0xFFFFFFFF);
		_tilemap->set_material_texture(0,"wall1.dds");
		_tilemap->set_material_texture(0,FACE_TOP,"top1.dds");
		_tilemap->set_material_color_diffuse(1,0xFFFFFFFF);
		_tilemap->set_material_texture(1,"wall2.dds");
		_tilemap->set_material_texture(1,FACE_TOP,"top2.dds");

		_tilemap->create(50,50);
		for (int x=0; x<50; x++)
		{
			for (int y=0; y<50; y++)
			{
				_tilemap->set(x,y,((float)(rand()%100))/100.0f,rand()%2);
			}
		}
		_tilemap->set_tile_size(4.0f,4.0f);
		_tilemap->build_geom();

		_tilemap->_pos.set(-100.0f,0.0f,-100.0f);

		_light=new Light(NULL);
		_light->set_type(LIGHT_POINT);
		_light->_pos.set(0.0f,20.0f,20.0f);
		_light->set_range(0.0f,FLT_MAX);
		_light->_ori.rotation_quaternion(Vec3f(1.0f,0.0f,0.0f),30.0f);
		_light->set_color(1.0f,1.0f,1.0f,1.0f);

		// Build geometry pieces
		Material* material=new Material;

		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Test 1: 4 stacked transparent quads
		/*_mesh=geom_add_quad(NULL ,Vec3f(0.0f,5.0f,0.0f),Vec3f(5.0f,0.0f,0.0f),Vec3f(0.0f,0.0f,5.0f),0x055FF0000,Rect(0.0f,0.0f,1.0f,1.0f));
		_mesh=geom_add_quad(_mesh,Vec3f(0.0f,6.0f,0.0f),Vec3f(5.0f,0.0f,0.0f),Vec3f(0.0f,0.0f,5.0f),0x055FF0000,Rect(0.0f,0.0f,1.0f,1.0f));
		_mesh=geom_add_quad(_mesh,Vec3f(0.0f,7.0f,0.0f),Vec3f(5.0f,0.0f,0.0f),Vec3f(0.0f,0.0f,5.0f),0x055FF0000,Rect(0.0f,0.0f,1.0f,1.0f));
		_mesh=geom_add_quad(_mesh,Vec3f(0.0f,8.0f,0.0f),Vec3f(5.0f,0.0f,0.0f),Vec3f(0.0f,0.0f,5.0f),0x055FF0000,Rect(0.0f,0.0f,1.0f,1.0f));
		material->set_material_color(1.0f,1.0f,1.0f,1.0f);
		material->set_color_vertex(true,false);
		material->set_fx(FX_SOLID);
		material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);//*/
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Test 2: A cube without the backface, textured
		/*_mesh=geom_add_cube(NULL ,Vec3f(0.0f,5.0f,0.0f),Vec3f(5.0f,0.0f,0.0f),Vec3f(0.0f,5.0f,0.0f),Vec3f(0.0f,0.0f,5.0f),0x0FFFFFFFF,
							true,Rect(0.0f,0.0f,1.0f,1.0f),
							false,Rect(0.0f,0.0f,1.0f,1.0f),
							true,Rect(0.0f,0.0f,1.0f,1.0f),
							true,Rect(0.0f,0.0f,1.0f,1.0f),
							true,Rect(0.0f,0.0f,1.0f,1.0f),
							true,Rect(0.0f,0.0f,1.0f,1.0f));
		material->set_material_color(1.0f,1.0f,1.0f,1.0f);
		material->set_color_vertex(true,false);
		material->set_fx(FX_TEXTURED);
		material->set_cull_mode(D3DCULL_NONE);
		material->set_texture(0,wood_texture);
		material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);//*/
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Test 3: A spheroid, solid
		/*_mesh=geom_add_spheroid(NULL ,Vec3f(0.0f,5.0f,0.0f),Vec3f(5.0f,0.0f,0.0f),Vec3f(0.0f,10.0f,0.0f),Vec3f(0.0f,0.0f,5.0f),
								0.0f,180.0f,5.0f,0.0f,360.0f,10.0f,
								0x0FFFFFFFF,Rect(0.0f,0.0f,1.0f,1.0f));
		material->set_material_color(1.0f,1.0f,1.0f,1.0f);
		material->set_color_vertex(true,false);
		material->set_fx(FX_SOLID);
		material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);//*/
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Test 4: A spheroid, solid, missing a slice
		/*_mesh=geom_add_spheroid(NULL ,Vec3f(0.0f,5.0f,0.0f),Vec3f(5.0f,0.0f,0.0f),Vec3f(0.0f,2.5f,0.0f),Vec3f(0.0f,0.0f,5.0f),
								0.0f,180.0f,5.0f,0.0f,240.0f,10.0f,
								0x0FFFFFFFF,Rect(0.0f,0.0f,1.0f,1.0f));
		material->set_material_color(1.0f,1.0f,1.0f,1.0f);
		material->set_color_vertex(true,false);
		material->set_cull_mode(D3DCULL_NONE);
		material->set_fx(FX_SOLID);
		material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);//*/
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Test 5: A half-spheroid, solid yellow, red-capped
		/*_mesh=geom_add_spheroid(NULL ,Vec3f(0.0f,5.0f,0.0f),Vec3f(5.0f,0.0f,0.0f),Vec3f(0.0f,5.0f,0.0f),Vec3f(0.0f,0.0f,5.0f),
								0.0f,90.0f,5.0f,0.0f,360.0f,10.0f,
								0x0FFFFFF00,Rect(0.0f,0.0f,1.0f,1.0f));
		_mesh=geom_add_cap(_mesh,Vec3f(0.0f,5.0f,0.0f),Vec3f(5.0f,0.0f,0.0f),Vec3f(0.0f,0.0f,0.0f),Vec3f(0.0f,0.0f,5.0f),
						   0.0f,360.0f,10.0f,
						   0x0FFFF0000,Rect(0.0f,0.0f,1.0f,1.0f));
		material->set_material_color(1.0f,1.0f,1.0f,1.0f);
		material->set_color_vertex(true,false);
		material->set_cull_mode(D3DCULL_NONE);
		material->set_fx(FX_SOLID);
		material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);//*/
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Test 6: An inverted cone, cyan, with a dark green cap
		/*_mesh=geom_add_cap(NULL,Vec3f(0.0f,5.0f,0.0f),Vec3f(5.0f,0.0f,0.0f),Vec3f(0.0f,-5.0f,0.0f),Vec3f(0.0f,0.0f,5.0f),
						   0.0f,360.0f,10.0f,
						   0x0FF00FFFF,Rect(0.0f,0.0f,1.0f,1.0f));
		_mesh=geom_add_cap(_mesh,Vec3f(0.0f,5.0f,0.0f),Vec3f(5.0f,0.0f,0.0f),Vec3f(0.0f,0.0f,0.0f),Vec3f(0.0f,0.0f,5.0f),
						   0.0f,360.0f,10.0f,
						   0x0FF006600,Rect(0.0f,0.0f,1.0f,1.0f));
		material->set_material_color(1.0f,1.0f,1.0f,1.0f);
		material->set_color_vertex(true,false);
		material->set_cull_mode(D3DCULL_NONE);
		material->set_fx(FX_SOLID);
		material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);//*/
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Test 6: A orange cylinder, larger radius on the bottom -> Note that the center is now the bottom center
		_mesh=geom_add_cylinder(NULL,Vec3f(0.0f,5.0f,0.0f),Vec3f(1.0f,0.0f,0.0f),Vec3f(0.0f,6.0f,0.0f),Vec3f(0.0f,0.0f,0.5f),
								6.0f,1.0f,10.0f,
								0.0f,360.0f,10.0f,
								0x0FFFF7700,Rect(0.0f,0.0f,1.0f,1.0f));
		material->set_material_color(1.0f,1.0f,1.0f,1.0f);
		material->set_color_vertex(true,false);
		material->set_cull_mode(D3DCULL_NONE);
		material->set_alpha(true,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA);

		_mesh->set_material(material,true);

		_mesh->close(true);

		_object3d=new ModelObj3d("model",NULL);
		_object3d->set_mesh(_mesh,false);
	}
	virtual ~TestScreen()
	{
		TEXTURES->clear_all();
	}
	virtual void anim(const float t)
	{
		if (KEY->is_pressed(DIK_ESCAPE)) _exit=true;
	}

	virtual void render()
	{
		GFX->begin_scene();

		// 3d tests
		GFX->clear(D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,1.0f,0.0f,1.0f,1.0f,1.0f,0);
		GFX->set_ambient_light(Color(0.3f,0.3f,0.3f,1.0f));

		_ang+=0.005f; if (_ang>360.0f) _ang-=360.0f;
		
		float dist=30.0f;

		_camera->_pos.x=dist*sinf(_ang);
		_camera->_pos.y=dist;
		_camera->_pos.z=dist*cosf(_ang);
		_camera->set_lookat(_camera->_pos,Vec3f(0.0f,0.0f,0.0f));

		_camera->set();
		GFX->set_world_transformation(0,0.0f,0.0f,0.0f,Quaternion(0,0,0,1));
		GFX->clear_lights();
		GFX->add_light(_light);
		  
		GFX->set_pass(PASS_SOLID);
		_tilemap->draw();
		_object3d->draw();

		BLIT3D->flush();

		GFX->clear_lights();

		GFX->end_scene();
	}

public:
	Camera*					_camera;
	float					_ang;
	Tilemap3d*				_tilemap;
	Light*					_light;
	GeomMesh*				_mesh;
	ModelObj3d*				_object3d;
};


INT WINAPI WinMain(HINSTANCE hInst,HINSTANCE,LPSTR,INT)
{
	if (!init(hInst)) return -1;

	Screen* tmp=new TestScreen;
	tmp->run();

	delete tmp;

	shutdown();

    return 0;
}
